"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.makeFieldBatchCodec = exports.FieldBatchEncoder = void 0;
const core_utils_1 = require("@fluidframework/core-utils");
const treeCompressionUtils_1 = require("../../treeCompressionUtils");
const versioned_1 = require("../../versioned");
const format_1 = require("./format");
const chunkDecoding_1 = require("./chunkDecoding");
const schemaBasedEncoding_1 = require("./schemaBasedEncoding");
const uncompressedEncode_1 = require("./uncompressedEncode");
/**
 * Helper for processing multiple fields one at a time.
 */
class FieldBatchEncoder {
    constructor() {
        this.fields = [];
    }
    add(field) {
        (0, core_utils_1.assert)(field.mode === 1 /* CursorLocationType.Fields */, "Cursor for batch must be in fields mode.");
        this.fields.push(field);
        return this.fields.length - 1;
    }
    encode(codec) {
        return codec.encode(this.fields);
    }
}
exports.FieldBatchEncoder = FieldBatchEncoder;
function makeFieldBatchCodec(options) {
    // TODO: every time context changes, withSchemaValidation and makeVersionedCodec recompile their json validators.
    // Those should be reused. Making more code context aware could fix that.
    return (context) => (0, versioned_1.makeVersionedValidatedCodec)(options, format_1.validVersions, format_1.EncodedFieldBatch, {
        encode: (data) => {
            for (const cursor of data) {
                (0, core_utils_1.assert)(cursor.mode === 1 /* CursorLocationType.Fields */, "FieldBatch expects fields cursors");
            }
            let encoded;
            switch (context.encodeType) {
                case treeCompressionUtils_1.TreeCompressionStrategy.Uncompressed:
                    encoded = (0, uncompressedEncode_1.uncompressedEncode)(data);
                    break;
                case treeCompressionUtils_1.TreeCompressionStrategy.Compressed:
                    // eslint-disable-next-line unicorn/prefer-ternary
                    if (context.schema !== undefined) {
                        encoded = (0, schemaBasedEncoding_1.schemaCompressedEncode)(context.schema.schema, context.schema.policy, data);
                    }
                    else {
                        // TODO: consider enabling a somewhat compressed but not schema accelerated encode.
                        encoded = (0, uncompressedEncode_1.uncompressedEncode)(data);
                    }
                    break;
                default:
                    (0, core_utils_1.unreachableCase)(context.encodeType);
            }
            // TODO: consider checking input data was in schema.
            return encoded;
        },
        decode: (data) => {
            // TODO: consider checking data is in schema.
            return (0, chunkDecoding_1.decode)(data).map((chunk) => chunk.cursor());
        },
    });
}
exports.makeFieldBatchCodec = makeFieldBatchCodec;
//# sourceMappingURL=codecs.js.map