"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.ForestSummarizer = void 0;
const client_utils_1 = require("@fluid-internal/client-utils");
const shared_object_base_1 = require("@fluidframework/shared-object-base");
const core_utils_1 = require("@fluidframework/core-utils");
const core_1 = require("../../core");
const util_1 = require("../../util");
const codec_1 = require("../../codec");
// eslint-disable-next-line import/no-internal-modules
const chunkTree_1 = require("../chunked-forest/chunkTree");
const codec_2 = require("./codec");
/**
 * The storage key for the blob in the summary containing tree data
 */
const treeBlobKey = "ForestTree";
/**
 * Provides methods for summarizing and loading a forest.
 */
class ForestSummarizer {
    get codec() {
        this.lazyCodec ?? (this.lazyCodec = (0, codec_2.makeForestSummarizerCodec)(this.options, this.fieldBatchCodec)({
            schema: { schema: this.schema, policy: this.policy },
            encodeType: this.encodeType,
        }));
        return this.lazyCodec;
    }
    constructor(forest, schema, policy, encodeType, fieldBatchCodec, options = { jsonValidator: codec_1.noopValidator }) {
        this.forest = forest;
        this.schema = schema;
        this.policy = policy;
        this.encodeType = encodeType;
        this.fieldBatchCodec = fieldBatchCodec;
        this.options = options;
        this.key = "Forest";
        this.schema.on("beforeSchemaChange", () => {
            this.lazyCodec = undefined;
        });
    }
    /**
     * Synchronous monolithic summarization of tree content.
     *
     * TODO: when perf matters, this should be replaced with a chunked async version using a binary format.
     *
     * @returns a snapshot of the forest's tree as a string.
     */
    getTreeString(stringify) {
        const rootCursor = this.forest.getCursorAboveDetachedFields();
        const fieldMap = new Map();
        // TODO: Encode all detached fields in one operation for better performance and compression
        (0, core_1.forEachField)(rootCursor, (cursor) => {
            const key = cursor.getFieldKey();
            const innerCursor = this.forest.allocateCursor();
            (0, core_utils_1.assert)(this.forest.tryMoveCursorToField({ fieldKey: key, parent: undefined }, innerCursor) === 1 /* TreeNavigationResult.Ok */, "failed to navigate to field");
            fieldMap.set(key, innerCursor);
        });
        const encoded = this.codec.encode(fieldMap);
        fieldMap.forEach((value) => value.free());
        return stringify(encoded);
    }
    getAttachSummary(stringify, fullTree, trackState, telemetryContext) {
        return (0, shared_object_base_1.createSingleBlobSummary)(treeBlobKey, this.getTreeString(stringify));
    }
    async summarize(stringify, fullTree, trackState, telemetryContext) {
        return (0, shared_object_base_1.createSingleBlobSummary)(treeBlobKey, this.getTreeString(stringify));
    }
    getGCData(fullGC) {
        // TODO: Properly implement garbage collection. Right now, garbage collection is performed automatically
        // by the code in SharedObject (from which SharedTreeCore extends). The `runtime.uploadBlob` API delegates
        // to the `BlobManager`, which automatically populates the summary with ISummaryAttachment entries for each
        // blob.
        return {
            gcNodes: {},
        };
    }
    async load(services, parse) {
        if (await services.contains(treeBlobKey)) {
            const treeBuffer = await services.readBlob(treeBlobKey);
            const treeBufferString = (0, client_utils_1.bufferToString)(treeBuffer, "utf8");
            // TODO: this code is parsing data without an optional validator, this should be defined in a typebox schema as part of the
            // forest summary format.
            const fields = this.codec.decode(parse(treeBufferString));
            const allocator = (0, util_1.idAllocatorFromMaxId)();
            const fieldChanges = [];
            for (const [fieldKey, field] of fields) {
                const chunked = (0, chunkTree_1.chunkField)(field, chunkTree_1.defaultChunkPolicy);
                const nodeCursors = chunked.flatMap((chunk) => (0, core_1.mapCursorField)(chunk.cursor(), (cursor) => cursor.fork()));
                const buildId = { minor: allocator.allocate(nodeCursors.length) };
                fieldChanges.push([
                    fieldKey,
                    {
                        build: [
                            {
                                id: buildId,
                                trees: nodeCursors,
                            },
                        ],
                        local: [{ count: nodeCursors.length, attach: buildId }],
                    },
                ]);
            }
            (0, core_utils_1.assert)(this.forest.isEmpty, 0x797 /* forest must be empty */);
            (0, core_1.applyDelta)({ fields: new Map(fieldChanges) }, this.forest, (0, core_1.makeDetachedFieldIndex)("init"));
        }
    }
}
exports.ForestSummarizer = ForestSummarizer;
//# sourceMappingURL=forestSummarizer.js.map