"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.makeV0Codec = void 0;
const core_utils_1 = require("@fluidframework/core-utils");
const util_1 = require("../../util");
const chunked_forest_1 = require("../chunked-forest");
const treeCompressionUtils_1 = require("../treeCompressionUtils");
const genericFieldKind_1 = require("./genericFieldKind");
const modularChangeFormat_1 = require("./modularChangeFormat");
function makeV0Codec(fieldKinds, revisionTagCodec, fieldsCodecWithoutContext, { jsonValidator: validator }) {
    const nodeChangesetCodec = {
        encode: encodeNodeChangesForJson,
        decode: decodeNodeChangesetFromJson,
        encodedSchema: modularChangeFormat_1.EncodedNodeChangeset,
    };
    const getMapEntry = (field) => {
        const codec = field.changeHandler
            .codecsFactory(nodeChangesetCodec, revisionTagCodec)
            .resolve(0);
        return {
            codec,
            compiledSchema: codec.json.encodedSchema
                ? validator.compile(codec.json.encodedSchema)
                : undefined,
        };
    };
    const fieldChangesetCodecs = new Map([[genericFieldKind_1.genericFieldKind.identifier, getMapEntry(genericFieldKind_1.genericFieldKind)]]);
    fieldKinds.forEach((fieldKind, identifier) => {
        fieldChangesetCodecs.set(identifier, getMapEntry(fieldKind));
    });
    const getFieldChangesetCodec = (fieldKind) => {
        const entry = fieldChangesetCodecs.get(fieldKind);
        (0, core_utils_1.assert)(entry !== undefined, 0x5ea /* Tried to encode unsupported fieldKind */);
        return entry;
    };
    function encodeFieldChangesForJson(change) {
        const encodedFields = [];
        for (const [field, fieldChange] of change) {
            const { codec, compiledSchema } = getFieldChangesetCodec(fieldChange.fieldKind);
            const encodedChange = codec.json.encode(fieldChange.change);
            if (compiledSchema !== undefined && !compiledSchema.check(encodedChange)) {
                (0, util_1.fail)("Encoded change didn't pass schema validation.");
            }
            const fieldKey = field;
            const encodedField = {
                fieldKey,
                fieldKind: fieldChange.fieldKind,
                change: encodedChange,
            };
            encodedFields.push(encodedField);
        }
        return encodedFields;
    }
    function encodeNodeChangesForJson(change) {
        const encodedChange = {};
        const { fieldChanges, nodeExistsConstraint } = change;
        if (fieldChanges !== undefined) {
            encodedChange.fieldChanges = encodeFieldChangesForJson(fieldChanges);
        }
        if (nodeExistsConstraint !== undefined) {
            encodedChange.nodeExistsConstraint = nodeExistsConstraint;
        }
        return encodedChange;
    }
    function decodeFieldChangesFromJson(encodedChange) {
        const decodedFields = new Map();
        for (const field of encodedChange) {
            const { codec, compiledSchema } = getFieldChangesetCodec(field.fieldKind);
            if (compiledSchema !== undefined && !compiledSchema.check(field.change)) {
                (0, util_1.fail)("Encoded change didn't pass schema validation.");
            }
            const fieldChangeset = codec.json.decode(field.change);
            const fieldKey = (0, util_1.brand)(field.fieldKey);
            decodedFields.set(fieldKey, {
                fieldKind: field.fieldKind,
                change: (0, util_1.brand)(fieldChangeset),
            });
        }
        return decodedFields;
    }
    function decodeNodeChangesetFromJson(encodedChange) {
        const decodedChange = {};
        const { fieldChanges, nodeExistsConstraint } = encodedChange;
        if (fieldChanges !== undefined) {
            decodedChange.fieldChanges = decodeFieldChangesFromJson(fieldChanges);
        }
        if (nodeExistsConstraint !== undefined) {
            decodedChange.nodeExistsConstraint = nodeExistsConstraint;
        }
        return decodedChange;
    }
    function encodeBuilds(builds, fieldsCodec) {
        if (builds === undefined) {
            return undefined;
        }
        const treeBatcher = new chunked_forest_1.FieldBatchEncoder();
        const buildsArray = (0, util_1.nestedMapToFlatList)(builds).map(([r, i, t]) => 
        // `undefined` does not round-trip through JSON strings, so it needs special handling.
        // Most entries will have an undefined revision due to the revision information being inherited from the `ModularChangeset`.
        // We therefore optimize for the common case by omitting the revision when it is undefined.
        r !== undefined
            ? [revisionTagCodec.encode(r), i, treeBatcher.add(t.cursor())]
            : [i, treeBatcher.add(t.cursor())]);
        return buildsArray.length === 0
            ? undefined
            : { builds: buildsArray, trees: treeBatcher.encode(fieldsCodec) };
    }
    function decodeBuilds(encoded, fieldsCodec) {
        if (encoded === undefined || encoded.builds.length === 0) {
            return undefined;
        }
        const chunks = fieldsCodec.decode(encoded.trees);
        const getChunk = (index) => {
            (0, core_utils_1.assert)(index < chunks.length, "out of bounds index for build chunk");
            return (0, chunked_forest_1.chunkFieldSingle)(chunks[index], chunked_forest_1.defaultChunkPolicy);
        };
        const list = encoded.builds.map((tuple) => tuple.length === 3
            ? [revisionTagCodec.decode(tuple[0]), tuple[1], getChunk(tuple[2])]
            : [undefined, tuple[0], getChunk(tuple[1])]);
        return (0, util_1.nestedMapFromFlatList)(list);
    }
    function encodeRevisionInfos(revisions) {
        const encodedRevisions = [];
        for (const revision of revisions) {
            const encodedRevision = {
                revision: revisionTagCodec.encode(revision.revision),
            };
            if (revision.rollbackOf !== undefined) {
                encodedRevision.rollbackOf = revisionTagCodec.encode(revision.rollbackOf);
            }
            encodedRevisions.push(encodedRevision);
        }
        return encodedRevisions;
    }
    function decodeRevisionInfos(revisions) {
        const decodedRevisions = [];
        for (const revision of revisions) {
            const decodedRevision = {
                revision: revisionTagCodec.decode(revision.revision),
            };
            if (revision.rollbackOf !== undefined) {
                decodedRevision.rollbackOf = revisionTagCodec.decode(revision.rollbackOf);
            }
            decodedRevisions.push(decodedRevision);
        }
        return decodedRevisions;
    }
    // TODO: provide schema here to enable schema based compression.
    const fieldsCodecSchemaless = fieldsCodecWithoutContext({
        encodeType: treeCompressionUtils_1.TreeCompressionStrategy.Compressed,
    });
    // TODO: use withSchemaValidation here to validate data against format.
    return {
        encode: (change) => {
            // Destroys only exist in rollback changesets, which are never sent.
            (0, core_utils_1.assert)(change.destroys === undefined, "Unexpected changeset with destroys");
            return {
                maxId: change.maxId,
                revisions: change.revisions === undefined
                    ? change.revisions
                    : encodeRevisionInfos(change.revisions),
                changes: encodeFieldChangesForJson(change.fieldChanges),
                builds: encodeBuilds(change.builds, fieldsCodecSchemaless),
            };
        },
        decode: (change) => {
            const encodedChange = change;
            const decoded = {
                fieldChanges: decodeFieldChangesFromJson(encodedChange.changes),
            };
            if (encodedChange.builds !== undefined) {
                decoded.builds = decodeBuilds(encodedChange.builds, fieldsCodecSchemaless);
            }
            if (encodedChange.revisions !== undefined) {
                decoded.revisions = decodeRevisionInfos(encodedChange.revisions);
            }
            if (encodedChange.maxId !== undefined) {
                decoded.maxId = encodedChange.maxId;
            }
            return decoded;
        },
        encodedSchema: modularChangeFormat_1.EncodedModularChangeset,
    };
}
exports.makeV0Codec = makeV0Codec;
//# sourceMappingURL=modularChangeCodecs.js.map