/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { assert, unreachableCase } from "@fluidframework/core-utils";
import { TreeCompressionStrategy } from "../../treeCompressionUtils";
import { makeVersionedValidatedCodec } from "../../versioned";
import { EncodedFieldBatch, validVersions } from "./format";
import { decode } from "./chunkDecoding";
import { schemaCompressedEncode } from "./schemaBasedEncoding";
import { uncompressedEncode } from "./uncompressedEncode";
/**
 * Helper for processing multiple fields one at a time.
 */
export class FieldBatchEncoder {
    constructor() {
        this.fields = [];
    }
    add(field) {
        assert(field.mode === 1 /* CursorLocationType.Fields */, "Cursor for batch must be in fields mode.");
        this.fields.push(field);
        return this.fields.length - 1;
    }
    encode(codec) {
        return codec.encode(this.fields);
    }
}
export function makeFieldBatchCodec(options) {
    // TODO: every time context changes, withSchemaValidation and makeVersionedCodec recompile their json validators.
    // Those should be reused. Making more code context aware could fix that.
    return (context) => makeVersionedValidatedCodec(options, validVersions, EncodedFieldBatch, {
        encode: (data) => {
            for (const cursor of data) {
                assert(cursor.mode === 1 /* CursorLocationType.Fields */, "FieldBatch expects fields cursors");
            }
            let encoded;
            switch (context.encodeType) {
                case TreeCompressionStrategy.Uncompressed:
                    encoded = uncompressedEncode(data);
                    break;
                case TreeCompressionStrategy.Compressed:
                    // eslint-disable-next-line unicorn/prefer-ternary
                    if (context.schema !== undefined) {
                        encoded = schemaCompressedEncode(context.schema.schema, context.schema.policy, data);
                    }
                    else {
                        // TODO: consider enabling a somewhat compressed but not schema accelerated encode.
                        encoded = uncompressedEncode(data);
                    }
                    break;
                default:
                    unreachableCase(context.encodeType);
            }
            // TODO: consider checking input data was in schema.
            return encoded;
        },
        decode: (data) => {
            // TODO: consider checking data is in schema.
            return decode(data).map((chunk) => chunk.cursor());
        },
    });
}
//# sourceMappingURL=codecs.js.map