/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { Type } from "@sinclair/typebox";
import { schemaFormat, RevisionTagSchema } from "../../core";
import { brandedNumberType, JsonCompatibleReadOnlySchema, } from "../../util";
import { EncodedFieldBatch } from "../chunked-forest";
const noAdditionalProps = { additionalProperties: false };
export const ChangesetLocalIdSchema = brandedNumberType();
export const EncodedChangeAtomId = Type.Object({
    /**
     * Uniquely identifies the changeset within which the change was made.
     */
    revision: Type.Optional(RevisionTagSchema),
    /**
     * Uniquely identifies, in the scope of the changeset, the change made to the field.
     */
    localId: ChangesetLocalIdSchema,
}, noAdditionalProps);
const EncodedValueChange = Type.Object({
    revision: Type.Optional(RevisionTagSchema),
    value: Type.Optional(JsonCompatibleReadOnlySchema),
}, noAdditionalProps);
const EncodedValueConstraint = Type.Object({
    value: Type.Optional(JsonCompatibleReadOnlySchema),
    violated: Type.Boolean(),
}, noAdditionalProps);
const EncodedFieldChange = Type.Object({
    fieldKey: schemaFormat.FieldKeySchema,
    fieldKind: schemaFormat.FieldKindIdentifierSchema,
    // Implementation note: node and field change encoding is mutually recursive.
    // This field marks a boundary in that recursion to avoid constructing excessively complex
    // recursive types. Encoded changes are validated at this boundary at runtime--see modularChangeCodecs.ts.
    change: JsonCompatibleReadOnlySchema,
}, noAdditionalProps);
const EncodedFieldChangeMap = Type.Array(EncodedFieldChange);
const EncodedNodeExistsConstraint = Type.Object({
    violated: Type.Boolean(),
});
export const EncodedNodeChangeset = Type.Object({
    valueChange: Type.Optional(EncodedValueChange),
    fieldChanges: Type.Optional(EncodedFieldChangeMap),
    valueConstraint: Type.Optional(EncodedValueConstraint),
    nodeExistsConstraint: Type.Optional(EncodedNodeExistsConstraint),
}, noAdditionalProps);
export const EncodedRevisionInfo = Type.Object({
    revision: Type.Readonly(RevisionTagSchema),
    rollbackOf: Type.ReadonlyOptional(RevisionTagSchema),
}, noAdditionalProps);
/**
 * Index of field in an EncodedFieldBatch.
 */
export const EncodedTreeIndex = Type.Number({ multipleOf: 1, minimum: 0 });
// TODO:YA6307 adopt more efficient encoding, likely based on contiguous runs of IDs
export const EncodedBuildsArray = Type.Array(Type.Union([
    Type.Tuple([ChangesetLocalIdSchema, EncodedTreeIndex]),
    Type.Tuple([RevisionTagSchema, ChangesetLocalIdSchema, EncodedTreeIndex]),
]));
export const EncodedBuilds = Type.Object({
    builds: EncodedBuildsArray,
    /**
     * Fields indexed by the EncodedTreeIndexes above.
     * TODO: Strongly typing this here may result in redundant schema validation of this data.
     */
    trees: EncodedFieldBatch,
});
export const EncodedModularChangeset = Type.Object({
    maxId: Type.Optional(ChangesetLocalIdSchema),
    changes: EncodedFieldChangeMap,
    revisions: Type.ReadonlyOptional(Type.Array(EncodedRevisionInfo)),
    builds: Type.Optional(EncodedBuilds),
}, noAdditionalProps);
//# sourceMappingURL=modularChangeFormat.js.map