/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { assert } from "@fluidframework/core-utils";
import { UsageError } from "@fluidframework/telemetry-utils";
import { noopValidator } from "../codec";
import { Compatibility, TreeStoredSchemaRepository, makeDetachedFieldIndex, moveToDetachedField, rootFieldKey, schemaDataIsEmpty, } from "../core";
import { SharedTreeCore } from "../shared-tree-core";
import { defaultSchemaPolicy, ForestSummarizer, SchemaSummarizer, buildForest, buildChunkedForest, makeTreeChunker, DetachedFieldIndexSummarizer, createNodeKeyManager, nodeKeyFieldKey as defailtNodeKeyFieldKey, jsonableTreeFromFieldCursor, TreeCompressionStrategy, ViewSchema, FieldKinds, normalizeNewFieldContent, makeMitigatedChangeFamily, makeFieldBatchCodec, } from "../feature-libraries";
import { createEmitter } from "../events";
import { brand, disposeSymbol, fail } from "../util";
import { WrapperTreeView as ClassWrapperTreeView, toFlexConfig, } from "../class-tree";
import { afterSchemaChanges, initializeContent, schematize, } from "./schematizedTree";
import { createTreeCheckout } from "./treeCheckout";
import { CheckoutFlexTreeView } from "./treeView";
import { SharedTreeChangeFamily } from "./sharedTreeChangeFamily";
/**
 * Shared tree, configured with a good set of indexes and field kinds which will maintain compatibility over time.
 *
 * TODO: detail compatibility requirements.
 */
export class SharedTree extends SharedTreeCore {
    get storedSchema() {
        return this.view.storedSchema;
    }
    constructor(id, runtime, attributes, optionsParam, telemetryContextPrefix) {
        const options = { ...defaultSharedTreeOptions, ...optionsParam };
        const schema = new TreeStoredSchemaRepository();
        const forest = options.forest === ForestType.Optimized
            ? buildChunkedForest(makeTreeChunker(schema, defaultSchemaPolicy))
            : buildForest();
        const removedRoots = makeDetachedFieldIndex("repair", options);
        const schemaSummarizer = new SchemaSummarizer(runtime, schema, options, {
            getCurrentSeq: () => this.runtime.deltaManager.lastSequenceNumber,
        });
        const fieldBatchCodec = makeFieldBatchCodec(options);
        const forestSummarizer = new ForestSummarizer(forest, schema, defaultSchemaPolicy, options.summaryEncodeType, fieldBatchCodec, options);
        const removedRootsSummarizer = new DetachedFieldIndexSummarizer(removedRoots);
        const innerChangeFamily = new SharedTreeChangeFamily(options);
        const changeFamily = makeMitigatedChangeFamily(innerChangeFamily, SharedTreeChangeFamily.emptyChange, (error) => {
            // TODO:6344 Add telemetry for these errors.
            // Rethrowing the error has a different effect depending on the context in which the
            // ChangeFamily was invoked:
            // - If the ChangeFamily was invoked as part of incoming op processing, rethrowing the error
            // will cause the runtime to disconnect the client, log a severe error, and not reconnect.
            // This will not cause the host application to crash because it is not on the stack at that time.
            // TODO: let the host application know that the client is now disconnected.
            // - If the ChangeFamily was invoked as part of dealing with a local change, rethrowing the
            // error will cause the host application to crash. This is not ideal, but is better than
            // letting the application either send an invalid change to the server or allowing the
            // application to continue working when its local branches contain edits that cannot be
            // reflected in its views.
            // The best course of action for a host application in such a state is to restart.
            // TODO: let the host application know about this situation and provide a way to
            // programmatically reload the SharedTree container.
            throw error;
        });
        super([schemaSummarizer, forestSummarizer, removedRootsSummarizer], changeFamily, options, id, runtime, attributes, telemetryContextPrefix);
        /**
         * Creating multiple editable tree contexts for the same branch, and thus with the same underlying AnchorSet does not work due to how TreeNode caching works.
         * This flag is used to detect if one already exists for the main branch and error if creating a second.
         * THis should catch most accidental violations of this restriction but there are still ways to create two conflicting contexts (for example calling constructing one manually).
         *
         * TODO:
         * 1. API docs need to reflect this limitation or the limitation has to be removed.
         */
        this.hasView2 = false;
        this._events = createEmitter();
        const localBranch = this.getLocalBranch();
        this.view = createTreeCheckout({
            branch: localBranch,
            changeFamily,
            schema,
            forest,
            events: this._events,
            removedRoots,
        });
    }
    requireSchema(schema, onSchemaIncompatible, nodeKeyManager, nodeKeyFieldKey) {
        assert(this.hasView2 === false, 0x7f1 /* Cannot create second view from tree. */);
        const viewSchema = new ViewSchema(defaultSchemaPolicy, {}, schema);
        const compatibility = viewSchema.checkCompatibility(this.storedSchema);
        if (compatibility.write !== Compatibility.Compatible ||
            compatibility.read !== Compatibility.Compatible) {
            return undefined;
        }
        this.hasView2 = true;
        const view2 = new CheckoutFlexTreeView(this.view, schema, nodeKeyManager ?? createNodeKeyManager(this.runtime.idCompressor), nodeKeyFieldKey ?? brand(defailtNodeKeyFieldKey), () => {
            assert(this.hasView2, 0x7f2 /* unexpected dispose */);
            this.hasView2 = false;
        });
        const onSchemaChange = () => {
            const compatibilityInner = viewSchema.checkCompatibility(this.storedSchema);
            if (compatibilityInner.write !== Compatibility.Compatible ||
                compatibilityInner.read !== Compatibility.Compatible) {
                view2[disposeSymbol]();
                onSchemaIncompatible();
                return false;
            }
            else {
                return true;
            }
        };
        afterSchemaChanges(this._events, this.view, onSchemaChange);
        return view2;
    }
    contentSnapshot() {
        const cursor = this.view.forest.allocateCursor();
        try {
            moveToDetachedField(this.view.forest, cursor);
            return {
                schema: this.storedSchema.clone(),
                tree: jsonableTreeFromFieldCursor(cursor),
                removed: this.view.getRemovedRoots(),
            };
        }
        finally {
            cursor.free();
        }
    }
    schematizeInternal(config, nodeKeyManager, nodeKeyFieldKey) {
        if (this.hasView2 === true) {
            throw new UsageError("Only one view can be constructed from a given tree at a time. Dispose of the first before creating a second.");
        }
        // TODO:
        // When this becomes a more proper out of schema adapter, editing should be made lazy.
        // This will improve support for readonly documents, cross version collaboration and attribution.
        // Check for empty.
        if (this.view.forest.isEmpty && schemaDataIsEmpty(this.storedSchema)) {
            this.view.transaction.start();
            initializeContent(this.view, config.schema, () => {
                const field = { field: rootFieldKey, parent: undefined };
                const content = normalizeNewFieldContent({ schema: config.schema }, config.schema.rootFieldSchema, config.initialTree);
                switch (this.storedSchema.rootFieldSchema.kind.identifier) {
                    case FieldKinds.optional.identifier: {
                        const fieldEditor = this.editor.optionalField(field);
                        assert(content.length <= 1, 0x7f4 /* optional field content should normalize at most one item */);
                        fieldEditor.set(content.length === 0 ? undefined : content[0], true);
                        break;
                    }
                    case FieldKinds.sequence.identifier: {
                        const fieldEditor = this.editor.sequenceField(field);
                        // TODO: should do an idempotent edit here.
                        fieldEditor.insert(0, content);
                        break;
                    }
                    default: {
                        fail("unexpected root field kind during initialize");
                    }
                }
            });
            this.view.transaction.commit();
        }
        schematize(this.view.events, this.view, config);
        return (this.requireSchema(config.schema, () => fail("schema incompatible"), nodeKeyManager, nodeKeyFieldKey) ?? fail("Schematize failed"));
    }
    schematize(config) {
        const flexConfig = toFlexConfig(config);
        const view = this.schematizeInternal(flexConfig);
        return new ClassWrapperTreeView(view);
    }
    async loadCore(services) {
        await super.loadCore(services);
        this._events.emit("afterBatch");
    }
}
/**
 * Used to distinguish between different forest types.
 * @internal
 */
export var ForestType;
(function (ForestType) {
    /**
     * The "ObjectForest" forest type.
     */
    ForestType[ForestType["Reference"] = 0] = "Reference";
    /**
     * The "ChunkedForest" forest type.
     */
    ForestType[ForestType["Optimized"] = 1] = "Optimized";
})(ForestType || (ForestType = {}));
// TODO: The default summaryEncodeType is set to Uncompressed as there are many out of schema tests that break when using Compressed.
// This should eventually be changed to use Compressed as the default tree compression strategy so production gets the compressed format.
export const defaultSharedTreeOptions = {
    jsonValidator: noopValidator,
    forest: ForestType.Reference,
    summaryEncodeType: TreeCompressionStrategy.Uncompressed,
};
/**
 * A channel factory that creates {@link ISharedTree}s.
 * @internal
 */
export class SharedTreeFactory {
    constructor(options = {}) {
        this.options = options;
        this.type = "https://graph.microsoft.com/types/tree";
        this.attributes = {
            type: this.type,
            snapshotFormatVersion: "0.0.0",
            packageVersion: "0.0.0",
        };
    }
    async load(runtime, id, services, channelAttributes) {
        const tree = new SharedTree(id, runtime, channelAttributes, this.options, "SharedTree");
        await tree.load(services);
        return tree;
    }
    create(runtime, id) {
        const tree = new SharedTree(id, runtime, this.attributes, this.options, "SharedTree");
        tree.initializeLocal();
        return tree;
    }
}
//# sourceMappingURL=sharedTree.js.map