import { FluidObject } from '@fluidframework/core-interfaces';
import { IChannel } from '@fluidframework/datastore-definitions';
import { IChannelAttributes } from '@fluidframework/datastore-definitions';
import { IChannelFactory } from '@fluidframework/datastore-definitions';
import { IChannelServices } from '@fluidframework/datastore-definitions';
import { IExperimentalIncrementalSummaryContext } from '@fluidframework/runtime-definitions';
import { IFluidDataStoreRuntime } from '@fluidframework/datastore-definitions';
import { IFluidHandle } from '@fluidframework/core-interfaces';
import { IFluidLoadable } from '@fluidframework/core-interfaces';
import { IGarbageCollectionData } from '@fluidframework/runtime-definitions';
import { ISharedObject } from '@fluidframework/shared-object-base';
import { ISummaryTreeWithStats } from '@fluidframework/runtime-definitions';
import { ITelemetryContext } from '@fluidframework/runtime-definitions';
import { SessionSpaceCompressedId } from '@fluidframework/runtime-definitions';
import { StableId } from '@fluidframework/runtime-definitions';
import type { Static } from '@sinclair/typebox';
import type { TSchema } from '@sinclair/typebox';

/**
 * Converts an enum into a collection of schema which can be used in a union.
 * @remarks
 * Currently only supports `string` enums.
 * @example
 * ```typescript
 * enum Mode {
 * 	a = "A",
 * 	b = "B",
 * }
 * const ModeNodes = adaptEnum(schema, Mode);
 * type ModeNodes = NodeFromSchema<(typeof ModeNodes)[keyof typeof ModeNodes]>;
 * const nodeFromString: ModeNodes = ModeNodes(Mode.a);
 * const nodeFromSchema: ModeNodes = new ModeNodes.a();
 * const nameFromNode: Mode = nodeFromSchema.value;
 * class Parent extends schemaFactory.object("Parent", {
 * 	mode: typedObjectValues(ModeNodes),
 * }) {}
 * ```
 * @privateRemarks
 * TODO:
 * Extends this to support numeric enums.
 * Maybe require an explicit nested scope to group them under, or at least a warning about collisions.
 * Maybe just provide `SchemaFactory.nested` to east creating nested scopes?
 * @internal
 */
export declare function adaptEnum<TScope extends string, const TEnum extends Record<string, string>>(factory: SchemaFactory<TScope>, members: TEnum): (<TValue extends TEnum[keyof TEnum]>(value: TValue) => {
    readonly value: TValue;
}) & { readonly [Property in keyof TEnum]: {
        new (data?: EmptyObject | undefined): {
            readonly value: TEnum[Property];
        };
        readonly identifier: `${TScope}.${TEnum[Property]}`;
        readonly kind: NodeKind.Object;
        readonly info: unknown;
        readonly implicitlyConstructable: true;
    }; };

/**
 * Minimal selection of adapters (nothing for general out of schema, field level adjustments etc.).
 * Would be used with schematize and have actual conversion/update functionality.
 *
 * TODO: Support more kinds of adapters
 * TODO: support efficient lookup of adapters
 * @internal
 */
export declare interface Adapters {
    readonly tree?: readonly TreeAdapter[];
}

/**
 * Types for use in fields.
 *
 * "Any" is boxed in an array to allow use as variadic parameter.
 * @internal
 */
export declare type AllowedTypes = readonly [Any] | readonly LazyItem<FlexTreeNodeSchema>[];

/**
 * Types for use in fields.
 * @beta
 */
declare type AllowedTypes_2 = readonly LazyItem<TreeNodeSchema>[];

/**
 * Types for use in fields.
 * This representation is optimized for runtime use in view-schema.
 *
 * @remarks
 * See {@link TreeTypeSet} for a stored-schema compatible version using the {@link TreeNodeSchemaIdentifier}.
 * See {@link AllowedTypes} for a compile time optimized version.
 * @internal
 */
export declare type AllowedTypeSet = Any | ReadonlySet<FlexTreeNodeSchema>;

/**
 * Takes in `AllowedTypes` and returns a TypedTree union.
 * @internal
 */
export declare type AllowedTypesToFlexInsertableTree<T extends AllowedTypes> = [
T extends readonly LazyItem<FlexTreeNodeSchema>[] ? InsertableFlexNode<Assume<FlexListToUnion<T>, FlexTreeNodeSchema>> : ContextuallyTypedNodeData
][_InlineTrick];

/**
 * What kinds of updates to stored schema to permit.
 *
 * TODO:
 * Currently this does not account for lazy schema updates, and/or use of adapters.
 * @internal
 */
export declare enum AllowedUpdateType {
    /**
     * Do not update the stored schema to match view schema.
     */
    None = 0,
    /**
     * Update the stored schema to match the view schema if the current document contents are compatible with the view schema.
     * TODO: support this option.
     */
    /**
     * Update the stored schema to match view schema if all possible documents based on the current stored schema would be compatible with the view schema.
     */
    SchemaCompatible = 1
}

/**
 * Converts properties of an object which permit undefined into optional properties.
 * Removes fields which only allow undefined.
 * @internal
 */
export declare type AllowOptional<T> = [FlattenKeys<RequiredFields<T> & OptionalFields<T>>][_InlineTrick];

/**
 * Converts properties of an object which permit undefined into optional properties.
 * Removes fields which only allow undefined.
 *
 * @remarks
 * This version does not flatten the resulting type.
 * This version exists because some cases recursive types need to avoid this
 * flattening since it causes complication issues.
 *
 * See also `AllowOptional`.
 * @internal
 */
export declare type AllowOptionalNotFlattened<T> = [RequiredFields<T> & OptionalFields<T>][_InlineTrick];

/**
 * A way to refer to a particular tree location within an {@link AnchorSet}.
 * Associated with a ref count on the underlying {@link AnchorNode}.
 * @internal
 */
export declare type Anchor = Brand<number, "rebaser.Anchor">;

/**
 * Events for {@link AnchorNode}.
 * These events are triggered while the internal data structures are being updated.
 * Thus these events must not trigger reading of the anchorSet or forest.
 *
 * TODO:
 * - Include sub-deltas in events.
 * - Add more events.
 *
 * @internal
 */
export declare interface AnchorEvents {
    /**
     * When the anchor node will never get reused by its AnchorSet.
     * This means that the content it corresponds to has been permanently destroyed.
     *
     * @remarks
     * When this happens depends entirely on how the anchorSet is used.
     * It's possible nodes removed from the tree will be kept indefinitely, and thus never trigger this event, or they may be discarded immediately.
     */
    afterDestroy(anchor: AnchorNode): void;
    /**
     * One or more of the children of this node is just about to change.
     *
     * @remarks
     * Does not include edits of child subtrees: instead only includes changes to nodes which are direct children in this node's fields.
     */
    childrenChanging(anchor: AnchorNode): void;
    /**
     * One or more of the children of this node has just changed.
     *
     * @remarks
     * Does not include edits of child subtrees: instead only includes changes to nodes which are direct children in this node's fields.
     */
    childrenChanged(anchor: AnchorNode): void;
    /**
     * Before a change in this subtree happens.
     *
     * @remarks
     * Includes edits of child subtrees.
     */
    beforeChange(anchor: AnchorNode): void;
    /**
     * After a change in this subtree happened.
     *
     * @remarks
     * Includes edits of child subtrees.
     */
    afterChange(anchor: AnchorNode): void;
    /**
     * Something in this tree is changing.
     * The event can optionally return a {@link PathVisitor} to traverse the subtree.
     * Called on every parent (transitively) when a change is occurring.
     * Includes changes to this node itself.
     */
    subtreeChanging(anchor: AnchorNode): PathVisitor | void;
    /**
     * Value on this node is changing.
     */
    valueChanging(anchor: AnchorNode, value: Value): void;
}

/**
 * Maps anchors (which must be ones this locator knows about) to paths.
 * @internal
 */
export declare interface AnchorLocator {
    /**
     * Get the current location of an Anchor.
     * The returned value should not be used after an edit has occurred.
     *
     * TODO: support extra/custom return types for specific/custom anchor types:
     * for now caller must rely on data in anchor + returned node location
     * (not ideal for anchors for places or ranges instead of nodes).
     */
    locate(anchor: Anchor): AnchorNode | undefined;
}

/**
 * Node in a tree of anchors.
 * @internal
 */
export declare interface AnchorNode extends UpPath<AnchorNode>, ISubscribable<AnchorEvents> {
    /**
     * Allows access to data stored on the Anchor in "slots".
     * Use {@link anchorSlot} to create slots.
     */
    readonly slots: BrandedMapSubset<AnchorSlot<any>>;
    /**
     * Gets a child of this node.
     *
     * @remarks
     * This does not return an AnchorNode since there might not be one, and lazily creating one here would have messy lifetime management (See {@link AnchorNode#getOrCreateChildRef})
     * If an AnchorNode is required, use the AnchorSet to track then locate the returned path.
     * TODO:
     * Revisit this API.
     * Perhaps if we use weak down pointers and remove ref counting, we can make this return a AnchorNode.
     *
     */
    child(key: FieldKey, index: number): UpPath<AnchorNode>;
    /**
     * Gets a child AnchorNode (creating it if needed), and an Anchor owning a ref to it.
     * Caller is responsible for freeing the returned Anchor, and must not use the AnchorNode after that.
     */
    getOrCreateChildRef(key: FieldKey, index: number): [Anchor, AnchorNode];
}

/**
 * Collection of Anchors at a specific revision.
 *
 * See `Rebaser` for how to update across revisions.
 *
 * TODO: this should not be package exported.
 * If it's needed outside the package an Interface should be used instead which can reduce its
 * API surface to a small subset.
 *
 * @sealed
 * @internal
 */
export declare class AnchorSet implements ISubscribable<AnchorSetRootEvents>, AnchorLocator {
    private readonly events;
    /**
     * Incrementing counter to give each anchor in this set a unique index for its identifier.
     * "0" is reserved for the `NeverAnchor`.
     */
    private anchorCounter;
    /**
     * Incrementing number that is bumped each time that the {@link AnchorSet} is changed.
     * This allows consumers to cache state associated with a particular generation number and later determine if that state may have been invalidated using a comparison with the current generation number.
     * For example, anchor slots can be used to cache the removal status of a node to memoize repeated walks up the tree.
     */
    generationNumber: number;
    /**
     * Special root node under which all anchors in this anchor set are transitively parented.
     * This does not appear in the UpPaths (instead they use undefined for the root).
     * Immediate children of this root are in detached fields (which have their identifiers used as the field keys).
     *
     * This is allocated with refCount one, which is never freed so it is never cleaned up
     * (as long as this AnchorSet is not garbage collected).
     *
     * There should never be any children other than the special root detached field under this between transactions:
     * TODO: check for and enforce this.
     */
    private readonly root;
    private readonly anchorToPath;
    private activeVisitor?;
    constructor();
    on<K extends keyof AnchorSetRootEvents>(eventName: K, listener: AnchorSetRootEvents[K]): () => void;
    /**
     * Check if there are currently no anchors tracked.
     * Mainly for testing anchor cleanup.
     */
    isEmpty(): boolean;
    locate(anchor: Anchor): AnchorNode | undefined;
    forget(anchor: Anchor): void;
    /**
     * TODO: Add APIs need to allow callers of this function to reduce copying here.
     * Ex: maybe return something extending UpPath here.
     * @param path - the path to the node to be tracked. If null, returns an anchor
     * which is permanently invalid.
     */
    track(path: UpPath | null): Anchor;
    /**
     * Finds a path node, creating if needed, and adds a ref count to it.
     */
    private trackInner;
    /**
     * Finds a path node if it already exists.
     */
    private find;
    /**
     * Returns an equivalent path making as much of it with PathNodes as possible.
     * This allows future operations (like find, track, locate) on this path (and derived ones) to be faster.
     * Note that the returned path may use AnchorNodes from this AnchorSet,
     * but does not have a tracked reference to them, so this should not be held onto across anything that might free an AnchorNode.
     *
     * @remarks
     * Also ensures that any PathNode in the path is from this AnchorSet.
     */
    internalizePath(originalPath: UpPath): UpPath;
    /**
     * Recursively marks the given `nodes` and their descendants as disposed and pointing to a deleted node.
     * Note that this does NOT detach the nodes.
     */
    private deepDelete;
    /**
     * Decouple nodes from their parent.
     * This removes the reference from the parent to the decoupled children, and updates the indexes of the remaining children accordingly.
     * This does NOT update the decoupled children: both their index and parent are left at their existing values.
     * To decouple and fixup the children, see `removeChildren` and `moveChildren`.
     * @param startPath - The path to the first node that is being decoupled.
     * @param count - number of siblings that are decoupled from the original tree.
     *
     * TODO: tests
     */
    private decoupleNodes;
    /**
     * Couple nodes to a parent.
     * @param destination - where the siblings are coupled to.
     * @param count - number of siblings that are coupled in the original tree.
     * @param coupleInfo - this object contains the nodes to couple and the parent index of the first node that is coupled in the original tree.
     *
     * TODO: tests
     */
    private coupleNodes;
    /**
     * Updates the parent indexes within `field` to account for `count` children being inserted at `fromParentIndex`. Note that
     * `fromParentIndex` is the logical position within the field, not the index with the sparse PathNode array.
     *
     * @param field - the field to update.
     * @param fromParentIndex - the logical index within the field to start updating from.
     * @param count - the number to increase parent indexes.
     * @returns the number of items in the field that are not increased.
     *
     * TODO: tests
     */
    private increaseParentIndexes;
    /**
     * Updates paths for a range move (including re-parenting path items and updating indexes).
     * @param sourceStart - where the siblings are removed from.
     * @param destination - where the siblings are moved to.
     * @param count - number of siblings to move.
     *
     * TODO:
     * How should anchors that become invalid, then valid again (ex: into content that was deleted, then undone) work?
     * Add an API to resurrect them? Store them in special detached fields? Store them in special non-detached fields?
     *
     * TODO:
     * How should custom anchors work (ex: ones not just tied to a specific Node)?
     * This design assumes they can be expressed in terms of a Node anchor + some extra stuff,
     * but we don't have an API for the extra stuff yet.
     *
     * TODO: tests
     */
    private moveChildren;
    private removeChildren;
    /**
     * Updates the parent indexes of all the nodes located at right side of the given path by the given offset.
     * @param firstSiblingToOffset - the path to offset children of.
     * @param offset - the offset to apply to the children.
     *
     */
    private offsetChildren;
    /**
     * Provides a visitor that can be used to mutate this {@link AnchorSet}.
     *
     * @returns A visitor that can be used to mutate this {@link AnchorSet}.
     *
     * @remarks
     * Mutating the {@link AnchorSet} does NOT update the forest.
     * The visitor must be released after use by calling {@link DeltaVisitor.free} on it.
     * It is invalid to acquire a visitor without releasing the previous one,
     * and this method will throw an error if this is attempted.
     */
    acquireVisitor(): AnnouncedVisitor & DeltaVisitor;
}

/**
 * Events for {@link AnchorSet}.
 * These events are triggered while the internal data structures are being updated.
 * Thus these events must not trigger reading of the anchorSet or forest.
 *
 * TODO:
 * - Design how events should be ordered.
 * - Include sub-deltas in events.
 * - Add more events.
 *
 * @internal
 */
export declare interface AnchorSetRootEvents {
    /**
     * What children are at the root is changing.
     */
    childrenChanging(anchors: AnchorSet): void;
    /**
     * Something in the tree is changing.
     */
    treeChanging(anchors: AnchorSet): void;
}

/**
 * Stores arbitrary, user-defined data on an {@link Anchor}.
 * This data is preserved over the course of that anchor's lifetime.
 * @see {@link anchorSlot} for creation and an example use case.
 * @internal
 */
export declare type AnchorSlot<TContent> = BrandedKey<Opaque<Brand<number, "AnchorSlot">>, TContent>;

/**
 * Define a strongly typed slot on anchors in which data can be stored.
 *
 * @remarks
 * This is mainly useful for caching data associated with a location in the tree.
 *
 * Example usage:
 * ```typescript
 * const counterSlot = anchorSlot<number>();
 *
 * function useSlot(anchor: AnchorNode): void {
 * 	anchor.slots.set(counterSlot, 1 + anchor.slots.get(counterSlot) ?? 0);
 * }
 * ```
 * @internal
 */
export declare function anchorSlot<TContent>(): AnchorSlot<TContent>;

/**
 * Visitor that is notified of changes before, after, and when changes are made.
 * Must be freed after use.
 * @internal
 */
export declare interface AnnouncedVisitor {
    /**
     * A hook that is called after all nodes have been created.
     */
    afterCreate(content: ProtoNodes, destination: FieldKey): void;
    beforeDestroy(field: FieldKey, count: number): void;
    beforeAttach(source: FieldKey, count: number, destination: PlaceIndex): void;
    afterAttach(source: FieldKey, destination: Range_2): void;
    beforeDetach(source: Range_2, destination: FieldKey): void;
    afterDetach(source: PlaceIndex, count: number, destination: FieldKey): void;
    beforeReplace(newContent: FieldKey, oldContent: Range_2, oldContentDestination: FieldKey): void;
    afterReplace(newContentSource: FieldKey, newContent: Range_2, oldContent: FieldKey): void;
}

/**
 * Allow any node (as long as it meets the schema for its own type).
 * @internal
 */
export declare const Any: "Any";

/**
 * Allow any node (as long as it meets the schema for its own type).
 * @internal
 */
export declare type Any = typeof Any;

/**
 * Suitable for output.
 * For input must error on side of excluding undefined instead.
 * @beta
 */
declare type ApplyKind<T, Kind extends FieldKind_2> = Kind extends FieldKind_2.Required ? T : undefined | T;

/**
 * Adjusts the API for a field based on its Multiplicity.
 * @internal
 */
export declare type ApplyMultiplicity<TMultiplicity extends Multiplicity, TypedChild> = {
    [Multiplicity.Forbidden]: undefined;
    [Multiplicity.Optional]: undefined | TypedChild;
    [Multiplicity.Sequence]: TypedChild[];
    [Multiplicity.Single]: TypedChild;
}[TMultiplicity];

/**
 * A symbol used to define a {@link MarkedArrayLike} interface.
 * @internal
 */
declare const arrayLikeMarkerSymbol: unique symbol;

/**
 * `ArrayLike` numeric indexed access, but writable.
 *
 * @remarks
 * Note that due to language limitations, this also allows reading as TSet.
 * This is why `TSet extends TGet` is required.
 *
 * See https://github.com/microsoft/TypeScript/issues/43826.
 * @internal
 */
export declare interface ArrayLikeMut<TGet, TSet extends TGet = TGet> extends ArrayLike<TGet> {
    [n: number]: TSet;
}

/**
 * Convert a Array type into the type of ReadonlySet.
 *
 * Same as `keyof ListToKeys<T, unknown>` but work for values that are not valid keys.
 * @beta
 */
declare type ArrayToUnion<T extends readonly unknown[]> = T extends readonly (infer TValue)[] ? TValue : never;

/**
 * Field kinds that allow value assignment.
 *
 * @internal
 */
export declare type AssignableFieldKinds = typeof FieldKinds.optional | typeof FieldKinds.required;

/**
 * Assume that `TInput` is a `TAssumeToBe`.
 *
 * @remarks
 * This is useful in generic code when it is impractical (or messy)
 * to to convince the compiler that a generic type `TInput` will extend `TAssumeToBe`.
 * In these cases `TInput` can be replaced with `Assume<TInput, TAssumeToBe>` to allow compilation of the generic code.
 * When the generic code is parameterized with a concrete type, if that type actually does extend `TAssumeToBe`,
 * it will behave like `TInput` was used directly.
 *
 * @internal
 */
export declare type Assume<TInput, TAssumeToBe> = [TInput] extends [TAssumeToBe] ? TInput : TAssumeToBe;

/**
 * Allows boxed iteration of a tree/field
 */
declare const boxedIterator: unique symbol;

/**
 * Constructs a "Branded" type, adding a type-checking only field to `ValueType`.
 *
 * Two usages of `Brand` should never use the same `Name`.
 * If they do, the resulting types will be assignable which defeats the point of this type.
 *
 * This type is constructed such that the first line of type errors when assigning
 * mismatched branded types will be:
 * `Type 'Name1' is not assignable to type 'Name2'.`
 *
 * These branded types are not opaque: A `Brand<A, B>` can still be used as a `B`.
 * @internal
 */
export declare type Brand<ValueType, Name extends string | ErasedType<string>> = ValueType & BrandedType<ValueType, Name extends Erased<infer TName> ? TName : Assume<Name, string>>;

/**
 * Adds a type {@link Brand} to a value.
 *
 * Only do this when specifically allowed by the requirements of the type being converted to.
 * @internal
 */
export declare function brand<T extends Brand<any, string>>(value: T extends BrandedType<infer ValueType, string> ? ValueType : never): T;

/**
 * Key in a {@link BrandedMapSubset}.
 * @internal
 */
export declare type BrandedKey<TKey, TContent> = TKey & Invariant<TContent>;

/**
 * @internal
 */
export declare type BrandedKeyContent<TKey extends BrandedKey<unknown, any>> = TKey extends BrandedKey<unknown, infer TContent> ? TContent : never;

/**
 * A Map where the keys carry the types of values which they correspond to.
 *
 * Example usage:
 * ```typescript
 * type FooSlot<TContent> = BrandedKey<Opaque<Brand<number, "FooSlot">>, TContent>;
 * const counterSlot = brandedSlot<FooSlot<number>>();
 * const slots: BrandedMapSubset<FooSlot<any>> = new Map();
 * slots.set(counterSlot, slots.get(counterSlot) ?? 0 + 1);
 * ```
 *
 * @remarks
 * These APIs are designed so that a Map can be used to implement this type.
 *
 * @internal
 */
export declare interface BrandedMapSubset<K extends BrandedKey<unknown, any>> {
    get<K2 extends K>(key: K2): BrandedKeyContent<K2> | undefined;
    has(key: K): boolean;
    set<K2 extends K>(key: K2, value: BrandedKeyContent<K2>): this;
    delete(key: K): boolean;
}

/**
 * Helper for {@link Brand}.
 * This is split out into its own as that's the only way to:
 * - have doc comments for the member.
 * - make the member protected (so you don't accidentally try and read it).
 * - get nominal typing (so types produced without using this class can never be assignable to it).
 * - allow use as {@link Opaque} branded type (not assignable to `ValueType`, but captures `ValueType`).
 *
 * See `InternalTypes.MakeNominal` for some more details.
 *
 * Do not use this class with `instanceof`: this will always be false at runtime,
 * but the compiler may think it's true in some cases.
 *
 * @remarks
 * This is covariant over ValueType.
 * This is suitable for when ValueType is immutable (like string or number),
 * which is the common use-case for branding.
 *
 * @sealed
 * @internal
 */
export declare abstract class BrandedType<out ValueType, Name extends string> {
    protected _typeCheck?: Covariant<ValueType>;
    /**
     * Compile time only marker to make type checking more strict.
     * This method will not exist at runtime and accessing it is invalid.
     * See {@link Brand} for details.
     *
     * @privateRemarks
     * `Name` is used as the return type of a method rather than a a simple readonly member as this allows types with two brands to be intersected without getting `never`.
     * The method takes in never to help emphasize that its not callable.
     */
    protected abstract brand(dummy: never): Name;
    /**
     * This class should never exist at runtime, so make it un-constructable.
     */
    private constructor();
}

/**
 * Options used to initialize (if needed) and schematize a `SharedTree`.
 * @remarks
 * Using this builder improves type safety and error quality over just constructing the configuration as a object.
 * @internal
 */
export declare function buildTreeConfiguration<T extends TreeFieldSchema>(config: InitializeAndSchematizeConfiguration<T>): InitializeAndSchematizeConfiguration<T>;

/**
 * An ID which is unique within a revision of a `ModularChangeset`.
 * A `ModularChangeset` which is a composition of multiple revisions may contain duplicate `ChangesetLocalId`s,
 * but they are unique when qualified by the revision of the change they are used in.
 * @internal
 */
export declare type ChangesetLocalId = Brand<number, "ChangesetLocalId">;

/**
 * Events for {@link ITreeCheckout}.
 * @beta
 */
export declare interface CheckoutEvents {
    /**
     * A batch of changes has finished processing and the view is in a consistent state.
     * It is once again safe to access the EditableTree, Forest and AnchorSet.
     *
     * @remarks
     * This is mainly useful for knowing when to do followup work scheduled during events from Anchors.
     */
    afterBatch(): void;
    /**
     * A revertible change has been made to this view.
     * Applications which subscribe to this event are expected to revert or discard revertibles they acquire, if they so choose (failure to do so will leak memory).
     * The provided revertible is inherently bound to the view that raised the event, calling `revert` won't apply to forked views.
     *
     * @remarks
     * This event provides a {@link Revertible} object that can be used to revert the change.
     */
    revertible(revertible: Revertible): void;
}

/**
 * Type to ensures two types overlap in at least one way.
 * It evaluates to the input type if this is true, and never otherwise.
 * Examples:
 * CheckTypesOverlap\<number | boolean, number | object\> = number | boolean
 * CheckTypesOverlap\<number | boolean, string | object\> = never
 * @internal
 */
export declare type CheckTypesOverlap<T, TCheck> = [Extract<T, TCheck> extends never ? never : T][0];

/**
 * Identifier for a child collection, either on a node/tree or at the root of a forest.
 * @internal
 */
export declare type ChildCollection = FieldKey | RootField;

/**
 * Location of a tree relative to is parent container (which can be a tree or forest).
 *
 * @internal
 */
export declare interface ChildLocation {
    readonly container: ChildCollection;
    readonly index: number;
}

/**
 * Collects the various parts of the API together.
 * @internal
 */
export declare type CollectOptions<TTypedFields, TValueSchema extends ValueSchema | undefined, TName> = TValueSchema extends undefined ? FlattenKeys<{
    [typeNameSymbol]?: UnbrandedName<TName>;
} & (TValueSchema extends ValueSchema ? {
    [valueSymbol]: TreeValue<TValueSchema>;
} : EmptyObject)> & TTypedFields : TValueSchema extends ValueSchema ? TreeValue<TValueSchema> : undefined;

/**
 * Compares two {@link LocalNodeKey}s.
 * All {@link LocalNodeKey}s retrieved from a single SharedTree client can be totally ordered using this comparator.
 * @param a - the first key to compare
 * @param b - the second key to compare
 * @returns `0` if `a` and `b` are the same key, otherwise `-1` if `a` is ordered before `b` or `1` if `a` is ordered after `b`.
 * @internal
 */
export declare function compareLocalNodeKeys(a: LocalNodeKey, b: LocalNodeKey): -1 | 0 | 1;

/**
 * Content of a field which needs external schema information to interpret.
 *
 * This format is intended for concise authoring of tree literals when the schema is statically known.
 *
 * Once schema aware APIs are implemented, they can be used to provide schema specific subsets of this type.
 * @internal
 */
export declare type ContextuallyTypedFieldData = ContextuallyTypedNodeData | undefined;

/**
 * Content of a tree which needs external schema information to interpret.
 *
 * This format is intended for concise authoring of tree literals when the schema is statically known.
 *
 * Once schema aware APIs are implemented, they can be used to provide schema specific subsets of this type.
 * @internal
 */
export declare type ContextuallyTypedNodeData = ContextuallyTypedNodeDataObject | number | string | boolean | null | readonly ContextuallyTypedNodeData[] | MarkedArrayLike<ContextuallyTypedNodeData>;

/**
 * Object case of {@link ContextuallyTypedNodeData}.
 * @internal
 */
export declare interface ContextuallyTypedNodeDataObject {
    /**
     * Value stored on this node.
     */
    readonly [valueSymbol]?: Value;
    /**
     * The type of the node.
     * If this node is well-formed, it must follow this schema.
     */
    readonly [typeNameSymbol]?: string;
    /**
     * Fields of this node, indexed by their field keys.
     *
     * Allow explicit undefined for compatibility with EditableTree, and type-safety on read.
     */
    [key: FieldKey]: ContextuallyTypedFieldData;
    /**
     * Fields of this node, indexed by their field keys as strings.
     *
     * Allow unbranded field keys as a convenience for literals.
     */
    [key: string]: ContextuallyTypedFieldData;
}

/**
 * Constrain generic type parameters to Contravariant.
 *
 * @example
 *
 * ```typescript
 * protected _typeCheck?: Contravariant<T>;
 * ```
 *
 * @internal
 */
export declare interface Contravariant<in T> {
    _removeCovariance?: (_: T) => void;
}

/**
 * Constrain generic type parameters to Covariant.
 *
 * @example
 *
 * ```typescript
 * protected _typeCheck?: Covariant<T>;
 * ```
 *
 * @internal
 */
export declare interface Covariant<out T> {
    _removeContravariance?: T;
}

/**
 * Non-exported symbol used to make IterableTreeListContent constructable only from within this file.
 */
declare const create: unique symbol;

/**
 * Create an {@link ISubscribable} that can be instructed to emit events via the {@link IEmitter} interface.
 *
 * A class can delegate handling {@link ISubscribable} to the returned value while using it to emit the events.
 * See also `EventEmitter` which be used as a base class to implement {@link ISubscribable} via extension.
 * @internal
 */
export declare function createEmitter<E extends Events<E>>(noListeners?: NoListenersCallback<E>): ISubscribable<E> & IEmitter<E> & HasListeners<E>;

/**
 * Provides functionality to allow a {@link stackTreeNodeCursor} and {@link stackTreeFieldCursor} to implement cursors.
 * @internal
 */
export declare interface CursorAdapter<TNode> {
    /**
     * @returns the value of the given node.
     */
    value(node: TNode): Value;
    /**
     * @returns the type of the given node.
     */
    type(node: TNode): TreeType;
    /**
     * @returns the keys for non-empty fields on the given node.
     */
    keysFromNode(node: TNode): readonly FieldKey[];
    /**
     * @returns the child nodes for the given node and key.
     */
    getFieldFromNode(node: TNode, key: FieldKey): readonly TNode[];
}

/**
 * Strongly typed {@link cursorFromContextualData} for a TreeNodeSchema.
 * @returns a cursor in Nodes mode for a single node containing the provided data.
 * @internal
 */
export declare function cursorForTypedTreeData<T extends FlexTreeNodeSchema>(context: TreeDataContext, schema: T, data: InsertableFlexNode<T>): ITreeCursorSynchronous;

/**
 * Construct a tree from ContextuallyTypedNodeData.
 *
 * TODO: this should probably be refactored into a `try` function which either returns a Cursor or a SchemaError with a path to the error.
 * @returns a cursor in Nodes mode for a single node containing the provided data.
 * @internal
 */
export declare function cursorFromContextualData(context: TreeDataContext, typeSet: AllowedTypeSet, data: ContextuallyTypedNodeData): ITreeCursorSynchronous;

/**
 * @internal
 */
export declare const enum CursorLocationType {
    /**
     * Can iterate through nodes in a field.
     * At a "current node".
     */
    Nodes = 0,
    /**
     * Can iterate through fields of a node.
     * At a "current field".
     */
    Fields = 1
}

/**
 * A symbol for marking an object as an {@link ITreeCursor}.
 *
 * Useful when APIs want to take in tree data in multiple formats, including cursors.
 */
declare const CursorMarker: unique symbol;

/**
 * {@link ITreeCursorSynchronous} that can return the underlying node objects.
 * @internal
 */
export declare interface CursorWithNode<TNode> extends ITreeCursorSynchronous {
    /**
     * Gets the underlying object for the current node.
     *
     * Only valid when `mode` is `Nodes`.
     */
    getNode(): TNode;
    /**
     * Create a copy of this cursor which navigates independently,
     * and is initially located at the same place as this one.
     *
     * Depending on the cursor implementation this may be significantly faster
     * than other ways to copy the cursor
     * (such as creating a new one and walking the path from this one).
     */
    fork(): CursorWithNode<TNode>;
}

/**
 * FullSchemaPolicy with the default field kinds.
 * @internal
 */
export declare const defaultSchemaPolicy: FullSchemaPolicy;

/**
 * Represents the creation of detached nodes.
 *
 * Tree creation is idempotent: if a tree with the same ID already exists,
 * then this build is ignored in favor of the existing tree.
 * @internal
 */
export declare interface DeltaDetachedNodeBuild<TTree = DeltaProtoNode> {
    readonly id: DeltaDetachedNodeId;
    readonly trees: readonly TTree[];
}

/**
 * Represents changes made to a detached node
 * @internal
 */
export declare interface DeltaDetachedNodeChanges<TTree = DeltaProtoNode> {
    readonly id: DeltaDetachedNodeId;
    readonly fields: DeltaFieldMap<TTree>;
}

/**
 * Represents the destruction of detached nodes
 * @internal
 */
export declare interface DeltaDetachedNodeDestruction {
    readonly id: DeltaDetachedNodeId;
    readonly count: number;
}

/**
 * A globally unique ID for a node in a detached field.
 * @internal
 */
export declare interface DeltaDetachedNodeId {
    readonly major?: string | number;
    readonly minor: number;
}

/**
 * Represents a detached node being assigned a new `DetachedNodeId`.
 * @internal
 */
export declare interface DeltaDetachedNodeRename {
    readonly count: number;
    readonly oldId: DeltaDetachedNodeId;
    readonly newId: DeltaDetachedNodeId;
}

/**
 * Represents the changes to perform on a given field.
 * @internal
 */
export declare interface DeltaFieldChanges<TTree = DeltaProtoNode> {
    /**
     * Represents a list of changes to the nodes in the field.
     * The index of each mark within the range of nodes, before
     * applying any of the changes, is not represented explicitly.
     * It corresponds to the sum of `mark.count` values for all previous marks for which `isAttachMark(mark)` is false.
     */
    readonly local?: readonly DeltaMark<TTree>[];
    /**
     * Changes to apply to detached nodes.
     * The ordering has no significance.
     *
     * Nested changes for a root that is undergoing a rename should be listed under the starting name.
     * For example, if one wishes to change a tree which is being renamed from ID A to ID B,
     * then the changes should be listed under ID A.
     */
    readonly global?: readonly DeltaDetachedNodeChanges<TTree>[];
    /**
     * New detached nodes to be constructed.
     * The ordering has no significance.
     *
     * @deprecated - Builds should be set at the root.
     * TODO:6308 migrate all reader/writers away from this and remove it.
     *
     * Build instructions for a root that is undergoing a rename should be listed under the starting name.
     * For example, if one wishes to build a tree which is being renamed from ID A to ID B,
     * then the build should be listed under ID A.
     */
    readonly build?: readonly DeltaDetachedNodeBuild<TTree>[];
    /**
     * New detached nodes to be destroyed.
     * The ordering has no significance.
     *
     * @deprecated - Destroys should be set at the root.
     * TODO:6308 migrate all reader/writers away from this and remove it.
     *
     * Destruction instructions for a root that is undergoing a rename should be listed under the final name.
     * For example, if one wishes to destroy a tree which is being renamed from ID A to ID B,
     * then the destruction should be listed under ID B.
     */
    readonly destroy?: readonly DeltaDetachedNodeDestruction[];
    /**
     * Detached whose associated ID needs to be updated.
     * The ordering has no significance.
     * Note that the renames may need to be performed in a specific order to avoid collisions.
     * This ordering problem is left to the consumer of this format.
     */
    readonly rename?: readonly DeltaDetachedNodeRename[];
}

/**
 * @internal
 */
export declare type DeltaFieldMap<TTree = DeltaProtoNode> = ReadonlyMap<FieldKey, DeltaFieldChanges<TTree>>;

/**
 * Represents a change being made to a part of the document tree.
 * @internal
 */
export declare interface DeltaMark<TTree = DeltaProtoNode> {
    /**
     * The number of nodes affected.
     * When `isAttachMark(mark)` is true, this is the number of new nodes being attached.
     * When `isAttachMark(mark)` is false, this the number of existing nodes affected.
     * Must be 1 when `fields` is populated.
     */
    readonly count: number;
    /**
     * Modifications to the pre-existing content.
     * Must be undefined when `attach` is set but `detach` is not.
     */
    readonly fields?: DeltaFieldMap<TTree>;
    /**
     * When set, indicates that some pre-existing content is being detached and sent to the given detached field.
     */
    readonly detach?: DeltaDetachedNodeId;
    /**
     * When set, indicates that some content is being attached from the given detached field.
     */
    readonly attach?: DeltaDetachedNodeId;
}

/**
 * The default representation for inserted content.
 *
 * TODO:
 * Ownership and lifetime of data referenced by this cursor is unclear,
 * so it is a poor abstraction for this use-case which needs to hold onto the data in a non-exclusive (readonly) way.
 * Cursors can be one supported way to input data, but aren't a good storage format.
 * @internal
 */
export declare type DeltaProtoNode = ITreeCursorSynchronous;

/**
 * This format describes changes that must be applied to a forest in order to update it.
 * Instances of this format are generated based on incoming changesets and consumed by a view layer (e.g., Forest) to
 * update itself.
 *
 * Because this format is only meant for updating document state, it does not fully represent user intentions.
 * For example, if some concurrent edits A and B insert content at the same location, then a Delta that represents
 * represents the state update for edit A would not include information that allows B's insertion to be ordered
 * relative to A's insertion. This format is therefore not fit to be rebased in the face of concurrent changes.
 * Instead this format is used to describe the end product of rebasing user intentions over concurrent edits.
 *
 * This format is self-contained in the following ways:
 *
 * 1. It uses integer indices (offsets, technically) to describe the locations of necessary changes.
 * As such, it does not rely on document nodes being accessible/locatable by ID.
 *
 * 2. This format does not require historical information in order to apply the changes it describes.
 * For example, if a user undoes the deletion of a subtree, then the Delta generated for the undo edit will contain all
 * information necessary to restore that subtree.
 *
 * This format can be generated from any Changeset without having access to the current document state.
 *
 * This format is meant to serve as the lowest common denominator to represent state changes resulting from any kind
 * of operation on any kind of field.
 * This means all such operations must be expressible in terms of this format.
 *
 * Within the above design constrains, this format is designed with the following goals in mind:
 *
 * 1. Make it easy to walk both a document tree and the delta tree to apply the changes described in the delta
 * with a minimum amount of backtracking over the contents of the tree.
 * This a boon for both code simplicity and performance.
 *
 * 2. Make the format terse.
 *
 * 3. Make the format uniform.
 *
 * These goals are reflected in the following design choices (this is very much optional reading for users of this
 * format):
 *
 * 1. All marks that apply to field elements are represented in a single linear structure where marks that affect later
 * elements of the document field appear after marks that affect earlier elements of the document field.
 *
 * If the marks were not ordered in this fashion then a consumer would need to backtrack within the document field.
 *
 * If the marks were represented in multiple such linear structures then it would be necessary to either:
 * - backtrack when iterating over one structure fully, then the next
 * - maintain a pointer within each such linear structure and advance them in lock-step (like in a k-way merge-sort but
 * more fiddly because of the offsets).
 *
 * 2. Nested changes are not inlined within `ProtoNode`s.
 *
 * Inlining them would force the consuming code to detect such changes within the `ProtoNode` and handle them
 * within the context of the content creation.
 * This would be cumbersome because either the code that is responsible for consuming the `ProtoNode` would need to
 * be aware of and have the context to handle such changes, or some caller of that code would need to find and extract such
 * change information ahead to calling that code.
 */
/**
 * Represents the change made to a document.
 * Immutable, therefore safe to retain for async processing.
 * @internal
 */
export declare interface DeltaRoot<TTree = DeltaProtoNode> {
    /**
     * Changes to apply to the root fields.
     */
    readonly fields?: DeltaFieldMap<TTree>;
    /**
     * New detached nodes to be constructed.
     * The ordering has no significance.
     *
     * Build instructions for a root that is undergoing a rename should be listed under the starting name.
     * For example, if one wishes to build a tree which is being renamed from ID A to ID B,
     * then the build should be listed under ID A.
     */
    readonly build?: readonly DeltaDetachedNodeBuild<TTree>[];
    /**
     * New detached nodes to be destroyed.
     * The ordering has no significance.
     *
     * Destruction instructions for a root that is undergoing a rename should be listed under the final name.
     * For example, if one wishes to destroy a tree which is being renamed from ID A to ID B,
     * then the destruction should be listed under ID B.
     */
    readonly destroy?: readonly DeltaDetachedNodeDestruction[];
}

/**
 * Visitor for changes in a delta.
 * Must be freed after use.
 * @internal
 */
export declare interface DeltaVisitor {
    /**
     * Frees/releases the visitor. Must be called once the visitor is no longer needed, since trying to acquire
     * a new one before freeing an existing one is invalid.
     */
    free(): void;
    /**
     * Creates nodes for the given content in a new detached field.
     * @param content - The content to create.
     * @param destination - The key for a new detached field.
     * A field with this key must not already exist.
     */
    create(content: ProtoNodes, destination: FieldKey): void;
    /**
     * Recursively destroys the given detached field and all of the nodes within it.
     * @param detachedField - The key for the detached field to destroy.
     * @param count - The number of nodes being destroyed.
     * Expected to match the number of nodes in the detached field being destroyed.
     */
    destroy(detachedField: FieldKey, count: number): void;
    /**
     * Transfers all the nodes from a detached field to the current field.
     * @param source - The detached field to transfer the nodes from.
     * @param count - The number of nodes being attached.
     * Expected to match the number of nodes in the source detached field.
     * @param destination - The index at which to attach the nodes.
     */
    attach(source: FieldKey, count: number, destination: PlaceIndex): void;
    /**
     * Transfers a range of nodes from the current field to a new detached field.
     * @param source - The bounds of the range of nodes to detach.
     * @param destination - The key for a new detached field.
     * A field with this key must not already exist.
     */
    detach(source: Range_2, destination: FieldKey): void;
    /**
     * Replaces a range of nodes in the current field by transferring them out to a new detached field
     * and transferring in all the nodes from an existing detached field in their place.
     * The number of nodes being detached must match the number of nodes being attached.
     * @param newContentSource - The detached field to transfer the new nodes from.
     * @param range - The bounds of the range of nodes to replace.
     * @param oldContentDestination - The key for a new detached field to transfer the old nodes to.
     */
    replace(newContentSource: FieldKey, range: Range_2, oldContentDestination: FieldKey): void;
    /**
     * Tells the visitor that it should update its "current location" to be the Node at the specified index
     * within the Field that is the current "current location".
     * Future calls to methods of {@link DeltaVisitor} will assume that's the location where their effects are happening.
     * @param index - The index (within the Field) of the node that should become the new "current location".
     *
     * @remarks This should only be called when the "current location" is a Field.
     */
    enterNode(index: NodeIndex): void;
    /**
     * Tells the visitor that it should update its "current location" to be the Field which contains the Node
     * that is the current "current location".
     * Future calls to methods of {@link DeltaVisitor} will assume that's the location where their effects are happening.
     * @param index - The index (within its Field) of the node that is being exited.
     *
     * @remarks This should only be called when the "current location" is a Node.
     */
    exitNode(index: NodeIndex): void;
    /**
     * Tells the visitor that it should update its "current location" to be the Field with the specified key,
     * within the Node that is the current "current location".
     * Future calls to methods of {@link DeltaVisitor} will assume that's the location where their effects are happening.
     * @param key - The key of the field that should become the new "current location".
     *
     * @remarks This should only be called when the "current location" is a Node.
     */
    enterField(key: FieldKey): void;
    /**
     * Tells the visitor that it should update its "current location" to be the Node which contains the Field
     * that is the current "current location".
     * Future calls to methods of {@link DeltaVisitor} will assume that's the location where their effects are happening.
     * @param key - The key of the field that is being exited.
     *
     * @remarks This should only be called when the "current location" is a Field.
     */
    exitField(key: FieldKey): void;
}

/**
 * A root in the forest.
 *
 * The range is a "container" like a field:
 * any additional content inserted before or after contents of this range will be included in the range.
 * This also means that moving the content from this range elsewhere will leave this range valid, but empty.
 *
 * In some APIs DetachedFields are used as FieldKeys on a special implicit root node
 * to simplify the APIs and implementation.
 * @internal
 */
export declare interface DetachedField extends Opaque<Brand<string, "tree.DetachedField">> {
}

/**
 * A place in a detached field.
 * @internal
 */
export declare type DetachedPlaceUpPath = Brand<Omit<PlaceUpPath, "parent">, "DetachedRangeUpPath">;

/**
 * A range of nodes in a detached field.
 * Unlike RangeUpPath, such ranges are generated by create and detach operations,
 * as well as consumed by destroy and attach operations.
 * @internal
 */
export declare type DetachedRangeUpPath = Brand<Omit<RangeUpPath, "parent">, "DetachedRangeUpPath">;

/**
 * The result of a discard operation.
 *
 * @beta
 */
export declare enum DiscardResult {
    /** The discard was successful. */
    Success = 0,
    /** The discard failed. */
    Failure = 1
}

/**
 * Placeholder for `Symbol.dispose`.
 *
 * Replace this with `Symbol.dispose` when it is available.
 * @beta
 */
export declare const disposeSymbol: unique symbol;

/**
 * A collection of events that can be raised by a {@link FlexTreeNode}.
 * These events are triggered while the internal data structures are being updated.
 * Thus these events must not trigger reading of the anchorSet or forest.
 *
 * TODO:
 * - Design how events should be ordered.
 * - Include sub-deltas in events.
 * - Add more events.
 * - Have some events (or a way to defer events) until the tree can be read.
 *
 * @internal
 */
export declare interface EditableTreeEvents {
    /**
     * Raised when a specific EditableTree node is changing.
     * This includes its fields.
     * @param upPath - the path corresponding to the location of the node being changed, upward.
     */
    changing(upPath: UpPath): void;
    /**
     * Raised when something in the tree is changing, including this node and its descendants.
     * The event can optionally return a {@link PathVisitor} to traverse the subtree
     * This event is called on every parent (transitively) when a change is occurring.
     * Includes changes to this node itself.
     * @param upPath - the path corresponding to the location of the node being changed, upward.
     * @returns a visitor to traverse the subtree or `void`.
     */
    subtreeChanging(upPath: UpPath): PathVisitor | void;
    /**
     * Raised on a node right before a change is applied to one of its fields or the fields of a descendant node.
     *
     * @param event - The event object. See {@link TreeEvent} for details.
     *
     * @remarks
     * What exactly qualifies as a change that triggers this event (or {@link EditableTreeEvents.afterChange}) is dependent
     * on the implementation of SharedTree. In general, these events will fire once for every atomic editing operation
     * supported by SharedTree; {@link EditableTreeEvents.beforeChange} before the change is applied, and
     * {@link EditableTreeEvents.afterChange} after it is.
     *
     * {@link FieldKinds.sequence} fields present two exceptions:
     *
     * The first one is that events will fire separately for each node involved in the operation (when inserting, removing,
     * or moving more than one node at a time). This means that, for example, when inserting two nodes into a {@link FieldKinds.sequence}
     * field the following will happen:
     * - {@link EditableTreeEvents.beforeChange} will fire once before either new node is present in the tree.
     * - {@link EditableTreeEvents.afterChange} will fire once after the first node is present in the tree, but the second one isn't.
     * - {@link EditableTreeEvents.beforeChange} will fire once before the second node is present in the tree, but the first one already is.
     * - {@link EditableTreeEvents.afterChange} will fire once after the second node is present in the tree (so at this point both nodes are).
     * Something similar applies to removing nodes from a sequence, and moving them to another sequence.
     *
     * The second one is that for an operation to move nodes, events will fire *twice* for each node being moved; once
     * while they are being detached from their source location, and once when they are being attached at the target location.
     */
    beforeChange(event: TreeEvent): void;
    /**
     * Raised on a node right after a change is applied to one of its fields or the fields of a descendant node.
     *
     * @param event - The event object. See {@link TreeEvent} for details.
     *
     * @remarks
     * What exactly qualifies as a change that triggers this event (or {@link EditableTreeEvents.beforeChange}) is dependent
     * on the implementation of SharedTree. In general, these events will fire once for every atomic editing operation supported
     * by SharedTree; {@link EditableTreeEvents.beforeChange} before the change is applied, and
     * {@link EditableTreeEvents.afterChange} after it is.
     *
     * {@link FieldKinds.sequence} present two exceptions:
     *
     * The first one is that events will fire separately for each node involved in the operation (when inserting, removing,
     * or moving more than one node at a time). This means that, for example, when inserting two nodes into a {@link FieldKinds.sequence}
     * field the following will happen:
     * - {@link EditableTreeEvents.beforeChange} will fire once before either new node is present in the tree.
     * - {@link EditableTreeEvents.afterChange} will fire once after the first node is present in the tree, but the second one isn't.
     * - {@link EditableTreeEvents.beforeChange} will fire once before the second node is present in the tree, but the first one already is.
     * - {@link EditableTreeEvents.afterChange} will fire once after the second node is present in the tree (so at this point both nodes are).
     * Something similar applies to removing nodes from a sequence, and moving them to another sequence.
     *
     * The second one is that for an operation to move nodes, events will fire *twice* for each node being moved; once
     * while they are being detached from their source location, and once when they are being attached at the target location.
     */
    afterChange(event: TreeEvent): void;
}

/**
 * The empty key ("") is used for unnamed relationships, such as the indexer
 * of an explicit array node.
 *
 * This key is a hint that this field is the primary function of the node,
 * and in some abstractions the APIs for this field should be inlined onto the node.
 *
 * TODO:
 * This has to be a FieldKey since different nodes will have different TreeFieldStoredSchema for it.
 * This makes it prone to collisions and suggests
 * that this intention may be better conveyed by metadata on the ITreeSchema.
 * @internal
 */
export declare const EmptyKey: FieldKey;

/**
 * Empty Object for use in type computations that should contribute no fields when `&`ed with another type.
 * @internal
 */
export declare type EmptyObject = {};

/**
 * Dumps schema into a deterministic JSON compatible semi-human readable but unspecified format.
 *
 * @remarks
 * This can be used to help inspect schema for debugging, and to save a snapshot of schema to help detect and review changes to an applications schema.
 *
 * This format may change across major versions of this package: such changes are considered breaking.
 * Beyond that, no compatibility guarantee is provided for this format: it should never be relied upon to load data, it should only be used for comparing outputs from this function.
 * @privateRemarks
 * This currently uses the schema summary format, but that could be changed to something more human readable (particularly if the encoded format becomes less human readable).
 * This intentionally does not leak the format types in the API.
 * @internal
 */
export declare function encodeTreeSchema(schema: TreeStoredSchema): JsonCompatible;

/**
 * Converts an array of distinct strings into a collection of schema which can be used like an enum style union.
 * @remarks
 * The returned collection is also a function which can be used to convert strings into {@link Unhydrated} nodes in the union.
 * Each node type has a `.value` getter which returns the associated string.
 *
 * The produced nodes use the provided strings as their `name`, and don't store any data beyond that.
 * @example
 * ```typescript
 * const Mode = enumFromStrings(schemaFactory, ["Fun", "Cool"]);
 * type Mode = NodeFromSchema<(typeof Mode)[keyof typeof Mode]>;
 * const nodeFromString: Mode = Mode("Fun");
 * const nodeFromSchema: Mode = new Mode.Fun();
 * const nameFromNode = nodeFromSchema.value;
 *
 * class Parent extends schemaFactory.object("Parent", { mode: typedObjectValues(Mode) }) {}
 * ```
 * @internal
 */
export declare function enumFromStrings<TScope extends string, const Members extends string>(factory: SchemaFactory<TScope>, members: Members[]): (<TValue extends Members>(value: TValue) => {
    readonly value: TValue;
}) & Record<Members, {
    new (data?: EmptyObject | undefined): {
        readonly value: Members;
    };
    readonly identifier: `${TScope}.${Members}`;
    readonly kind: NodeKind.Object;
    readonly info: unknown;
    readonly implicitlyConstructable: true;
}>;

/**
 * "opaque" handle which can be used to expose a branded type without referencing its value type.
 *
 * Recommended usage is to use `interface` instead of `type` so tooling (such as tsc and refactoring tools)
 * uses the type name instead of expanding it:
 * ```typescript
 * export interface MyType extends Erased<"myPackage.MyType">{}
 * ```
 * @internal
 */
export declare type Erased<Name extends string> = ErasedType<Name>;

/**
 * Opaque type erased handle to the encoded representation of the contents of a stored schema.
 * @internal
 */
export declare interface ErasedTreeNodeSchemaDataFormat extends Erased<"TreeNodeSchemaDataFormat"> {
}

/**
 * Helper for {@link Erased}.
 * This is split out into its own as that's the only way to:
 * - have doc comments for the member.
 * - make the member protected (so you don't accidentally try and read it).
 * - get nominal typing (so types produced without using this class can never be assignable to it).
 *
 * See `InternalTypes.MakeNominal` for some more details.
 *
 * Do not use this class with `instanceof`: this will always be false at runtime,
 * but the compiler may think it's true in some cases.
 *
 * @sealed
 * @internal
 */
export declare abstract class ErasedType<out Name extends string> {
    /**
     * Compile time only marker to make type checking more strict.
     * This method will not exist at runtime and accessing it is invalid.
     * See {@link Brand} for details.
     *
     * @privateRemarks
     * `Name` is used as the return type of a method rather than a a simple readonly member as this allows types with two brands to be intersected without getting `never`.
     * The method takes in never to help emphasize that its not callable.
     */
    protected abstract brand(dummy: never): Name;
    /**
     * This class should never exist at runtime, so make it un-constructable.
     */
    private constructor();
}

/**
 * Used to specify the kinds of events emitted by an {@link ISubscribable}.
 *
 * @remarks
 *
 * Any object type is a valid {@link Events}, but only the event-like properties of that
 * type will be included.
 *
 * @example
 *
 * ```typescript
 * interface MyEvents {
 *   load: (user: string, data: IUserData) => void;
 *   error: (errorCode: number) => void;
 * }
 * ```
 *
 * @beta
 */
export declare type Events<E> = {
    [P in (string | symbol) & keyof E as IsEvent<E[P]> extends true ? P : never]: E[P];
};

/**
 * See {@link extractFromOpaque}.
 * @internal
 */
export declare type ExtractFromOpaque<TOpaque extends BrandedType<any, string>> = TOpaque extends BrandedType<infer ValueType, infer Name> ? isAny<ValueType> extends true ? unknown : Brand<ValueType, Name> : never;

/**
 * Converts a {@link Opaque} handle to the underlying branded type.
 *
 * It is assumed that only code that produces these "opaque" handles does this conversion,
 * allowing these handles to be considered opaque.
 * @internal
 */
export declare function extractFromOpaque<TOpaque extends BrandedType<any, string>>(value: TOpaque): ExtractFromOpaque<TOpaque>;

/**
 * Get the `Item` type from a `LazyItem<Item>`.
 * @beta
 */
declare type ExtractItemType<Item extends LazyItem> = Item extends () => infer Result ? Result : Item;

/**
 * @internal
 */
export declare function fail(message: string): never;

/**
 * Anchor to a field.
 * This is structurally based on the parent, so it will move only as the parent moves.
 * @internal
 */
export declare interface FieldAnchor {
    /**
     * Node above this field.
     * If `undefined`, field is a detached field.
     */
    parent: Anchor | undefined;
    fieldKey: FieldKey;
}

/**
 * Generates field content for a MapTree on demand.
 * TODO:
 * Currently being exposed at the package level which also requires us to export MapTree at the package level.
 * Refactor the FieldGenerator to use JsonableTree instead of MapTree, and convert them internally.
 * @internal
 */
export declare type FieldGenerator = () => MapTree[];

/**
 * Key (aka Name or Label) for a field which is scoped to a specific TreeNodeStoredSchema.
 *
 * Stable identifier, used when persisting data.
 * @internal
 */
export declare type FieldKey = Brand<string, "tree.FieldKey">;

/**
 * Functionality for FieldKinds that is stable,
 * meaning that it can not change in any measurable way without providing a new identifier.
 *
 * It is assumed that this information and policy is available on all clients interacting with a document
 * using the identifier.
 *
 * This must contain enough information to process remote edits to this FieldKind consistently with all clients.
 * All behavior must be deterministic, and not change across versions of the app/library.
 *
 * These policies include the data encoding, change encoding, change rebase and change application.
 *
 * @sealed @internal
 * @privateRemarks
 * This being @sealed is for users of this package.
 * This package itself may provide implementations.
 * This pattern was picked instead of an interface since we do not have a convention for how to mark interfaces as only allowed to be implemented by the package declaring them,
 * and using a class also gets stronger (nominal) typing for objects which will be down cast.
 */
export declare abstract class FieldKind<TName extends string = string, TMultiplicity extends Multiplicity = Multiplicity> implements FieldKindSpecifier {
    readonly identifier: TName & FieldKindIdentifier;
    readonly multiplicity: TMultiplicity;
    /**
     * @param identifier - Globally scoped identifier.
     * @param multiplicity - bound on the number of children that fields of this kind may have.
     * TODO: consider replacing with numeric upper and lower bounds.
     */
    protected constructor(identifier: TName & FieldKindIdentifier, multiplicity: TMultiplicity);
}

/**
 * Kind of a field on a node.
 * @beta
 */
declare enum FieldKind_2 {
    /**
     * A field which can be empty or filled.
     * @remarks
     * Allows 0 or one child.
     */
    Optional = 0,
    /**
     * A field which must always be filled.
     * @remarks
     * Only allows exactly one child.
     */
    Required = 1
}

/**
 * Identifier for a FieldKind.
 * Refers to an exact stable policy (ex: specific version of a policy),
 * for how to handle (ex: edit and merge edits to) fields marked with this kind.
 * Persisted in documents as part of stored schema.
 * @internal
 */
export declare type FieldKindIdentifier = Brand<string, "tree.FieldKindIdentifier">;

/**
 * Default FieldKinds with their editor types erased.
 * @internal
 */
export declare const FieldKinds: {
    readonly required: Required_2;
    readonly optional: Optional;
    readonly sequence: Sequence;
    readonly nodeKey: NodeKeyFieldKind;
    readonly forbidden: Forbidden;
};

/**
 * Specifies which field kind to use.
 *
 * @remarks
 * This is used instead of just the FieldKindIdentifier so that it can be subtyped into a more expressive type with additional information.
 *
 * @internal
 */
export declare interface FieldKindSpecifier<T = FieldKindIdentifier> {
    identifier: T;
}

/**
 * Location of a field within a tree that is not a detached/root field.
 * @internal
 */
export declare interface FieldLocation {
    readonly key: FieldKey;
    readonly parent: ForestLocation;
}

/**
 * This modules provides a simple human readable (and editable) tree format.
 *
 * This implementation can handle all trees (so it does not need a fallback for any special cases),
 * and is not optimized.
 *
 * It's suitable for testing and debugging,
 * though it could also reasonably be used as a fallback for edge cases or for small trees.
 *
 * The serialized format is valid utf-8, and also includes a json compatible intermediate in memory format.
 *
 * This format is currently not stable: its internal contents are not considered public APIs and may change.
 * There is currently no guarantee that data serialized with this library will
 * be loadable with a different version of this library.
 *
 * TODO: stabilize this format (probably after schema are more stable).
 *
 * This format does not include schema: typically schema would be stored alongside data in this format.
 *
 * @privateRemarks A forked version of these types is available at `persistedTreeTextFormat.ts`.
 * Changes to them might necessitate changes to the analogous forked types, or codecs which transcode between
 * them.
 * See persistedTreeTextFormat's module documentation for more details.
 */
/**
 * Json compatible map as object.
 * Keys are FieldKey strings.
 * Values are the content of the field specified by the key.
 *
 * WARNING:
 * Be very careful when using objects as maps:
 * Use `Object.prototype.hasOwnProperty.call(fieldMap, key)` to safely check for keys.
 * Do NOT simply read the field and check for undefined as this will return values for `__proto__`
 * and various methods on Object.prototype, like `hasOwnProperty` and `toString`.
 * This exposes numerous bug possibilities, including prototype pollution.
 *
 * Due to the above issue, try to avoid this type (and the whole object as map pattern).
 * Only use this type when needed for json compatible maps,
 * but even in those cases consider lists of key value pairs for serialization and using `Map`
 * for runtime.
 *
 * @internal
 */
export declare interface FieldMapObject<TChild> {
    [key: string]: TChild[];
}

/**
 * TODO: replace (or subclass) this with more specific types, like "List".
 * @internal
 */
export declare class FieldNodeSchema<Name extends string = string, Specification extends Unenforced<TreeFieldSchema> = TreeFieldSchema> extends TreeNodeSchemaBase<Name, Specification> {
    protected _typeCheck2?: MakeNominal;
    static create<const Name extends string, const Specification extends TreeFieldSchema>(builder: Named<string>, name: TreeNodeSchemaIdentifier<Name>, specification: Specification): FieldNodeSchema<Name, Specification>;
    private constructor();
    getFieldSchema(field?: FieldKey): TreeFieldSchema;
}

/**
 * @internal
 */
export declare interface Fields {
    readonly [key: string]: TreeFieldSchema;
}

/**
 * All policy for a specific field,
 * including functionality that does not have to be kept consistent across versions or deterministic.
 *
 * This can include policy for how to use this schema for "view" purposes, and well as how to expose editing APIs.
 * @sealed @beta
 */
declare class FieldSchema<out Kind extends FieldKind_2 = FieldKind_2, out Types extends ImplicitAllowedTypes_2 = ImplicitAllowedTypes_2> {
    readonly kind: Kind;
    readonly allowedTypes: Types;
    /**
     * This class is used with instanceof, and therefore should have nominal typing.
     * This field enforces that.
     */
    protected _typeCheck?: MakeNominal;
    /**
     * @param kind - The [kind](https://en.wikipedia.org/wiki/Kind_(type_theory)) of this field.
     * Determine the multiplicity, viewing and editing APIs as well as the merge resolution policy.
     * @param allowedTypes - What types of tree nodes are allowed in this field.
     */
    constructor(kind: Kind, allowedTypes: Types);
}

/**
 * Path from a field in the tree upward.
 *
 * See {@link UpPath}.
 * @internal
 */
export declare interface FieldUpPath<TUpPath extends UpPath = UpPath> {
    /**
     * The parent, or undefined in the case where this path is to a detached sequence.
     */
    readonly parent: TUpPath | undefined;
    /**
     * The Field to which this path points.
     * Note that if `parent` returns `undefined`, this key  corresponds to a detached sequence.
     */
    readonly field: FieldKey;
}

/**
 * Utilities for manipulating types.
 */
/**
 * Return a type thats equivalent to the input, but with different IntelliSense.
 * This tends to convert unions and intersections into objects.
 * @internal
 */
export declare type FlattenKeys<T> = [{
    [Property in keyof T]: T[Property];
}][_InlineTrick];

/**
 * Strongly typed tree literals for inserting as the content of a field.
 *
 * If a cursor is provided, it must be in Fields mode.
 * @internal
 */
export declare type FlexibleFieldContent<TSchema extends TreeFieldSchema> = InsertableFlexField<TSchema> | ITreeCursorSynchronous;

/**
 * Strongly typed tree literals for inserting as a node.
 *
 * If a cursor is provided, it must be in Nodes mode.
 * @internal
 */
export declare type FlexibleNodeContent<TTypes extends AllowedTypes> = AllowedTypesToFlexInsertableTree<TTypes> | ITreeCursorSynchronous;

/**
 * Strongly typed tree literals for inserting a subsequence of nodes.
 *
 * Used to insert a batch of 0 or more nodes into some location in a {@link FlexTreeSequenceField}.
 *
 * If a cursor is provided, it must be in Fields mode.
 * @internal
 */
export declare type FlexibleNodeSubSequence<TTypes extends AllowedTypes> = Iterable<AllowedTypesToFlexInsertableTree<TTypes>> | ITreeCursorSynchronous;

/**
 * Type that when combined with a default {@link FieldKind} can be normalized into a {@link TreeFieldSchema}.
 * @internal
 */
export declare type FlexImplicitFieldSchema = TreeFieldSchema | ImplicitAllowedTypes;

/**
 * A flexible way to list values.
 * Each item in the list can either be an "eager" **value** or a "lazy" **function that returns a value** (the latter allows cyclic references to work).
 * @privateRemarks
 * By default, items that are of type `"function"` will be considered lazy and all other items will be considered eager.
 * To force a `"function"` item to be treated as an eager item, call `markEager` before putting it in the list.
 * This is necessary e.g. when the eager list items are function types and the lazy items are functions that _return_ function types.
 * `FlexList`s are processed by `normalizeFlexList` and `normalizeFlexListEager`.
 * @beta
 */
declare type FlexList<Item = unknown> = readonly LazyItem<Item>[];

/**
 * Normalize FlexList type to a union.
 * @beta
 */
declare type FlexListToUnion<TList extends FlexList> = ExtractItemType<ArrayToUnion<TList>>;

/**
 * Part of a tree.
 * Iterates over children.
 *
 * @privateRemarks
 * This exists mainly as a place to share common members between nodes and fields.
 * It is not expected to be useful or common to write code which handles this type directly.
 * If this assumption turns out to be false, and generically processing `UntypedEntity`s is useful,
 * then this interface should probably be extended with some down casting functionality (like `is`).
 *
 * TODO:
 * Design and document iterator invalidation rules and ordering rules.
 * Providing a custom iterator type with place anchor semantics would be a good approach.
 *
 * @internal
 */
export declare interface FlexTreeEntity<out TSchema = unknown> {
    /**
     * Indicates that an object is a specific kind of flex tree FlexTreeEntity.
     * This makes it possible to both down cast FlexTreeEntities safely as well as validate if an object is or is not a FlexTreeEntity.
     */
    readonly [flexTreeMarker]: FlexTreeEntityKind;
    /**
     * Schema for this entity.
     * If well-formed, it must follow this schema.
     */
    readonly schema: TSchema;
    /**
     * A common context of a "forest" of EditableTrees.
     */
    readonly context: TreeContext;
    /**
     * Gets the {@link TreeStatus} of this tree.
     *
     * @remarks
     * For non-root fields, this is the status of the parent node, since fields do not have a separate lifetime.
     */
    treeStatus(): TreeStatus;
    /**
     * Iterate through all nodes/fields in this field/node.
     *
     * @remarks
     * No mutations to the current view of the shared tree are permitted during iteration.
     */
    [boxedIterator](): IterableIterator<FlexTreeEntity>;
}

/**
 * @internal
 */
export declare enum FlexTreeEntityKind {
    Node = 0,
    Field = 1
}

/**
 * A collaboratively editable collection of nodes within a {@link FlexTreeEntity}.
 *
 * Fields are inherently part of their parent, and thus cannot be moved.
 * Instead their content can be moved, deleted or created.
 *
 * Editing operations are only valid on trees with the {@link TreeStatus#InDocument} `TreeStatus`.
 *
 * @remarks
 * Fields are used wherever an editable collection of nodes is required.
 * This is required in two places:
 * 1. To hold the children of non-leaf {@link FlexTreeNode}s.
 * 2. As the root of a {@link FlexTreeEntity}.
 *
 * Down-casting (via {@link FlexTreeField.is}) is required to access Schema-Aware APIs, including editing.
 * All content in the tree is accessible without down-casting, but if the schema is known,
 * the schema aware API may be more ergonomic.
 *
 * @internal
 */
export declare interface FlexTreeField extends FlexTreeEntity<TreeFieldSchema> {
    readonly [flexTreeMarker]: FlexTreeEntityKind.Field;
    /**
     * The `FieldKey` this field is under.
     * Defines what part of its parent this field makes up.
     */
    readonly key: FieldKey;
    /**
     * The node which has this field on it under `fieldKey`.
     * `undefined` iff this field is a detached field.
     */
    readonly parent?: FlexTreeNode;
    /**
     * Type guard for narrowing / down-casting to a specific schema.
     */
    is<TSchema extends TreeFieldSchema>(schema: TSchema): this is FlexTreeTypedField<TSchema>;
    [boxedIterator](): IterableIterator<FlexTreeNode>;
    /**
     * Check if this field is the same as a different field.
     * This is defined to mean that both are in the same editable tree, and are the same field on the same node.
     * This is more than just a reference comparison because unlike EditableTree nodes, fields are not cached on anchors and can be duplicated.
     *
     * @privateRemarks
     * TODO:
     * If practical, cache TreeField instances so use of this method can be replaced with `===` to compare object identity.
     * Implementing this will require some care to preserve lazy-ness and work efficiently (without leaks) for empty fields, particularly on MapNodes.
     */
    isSameAs(other: FlexTreeField): boolean;
}

/**
 * A {@link FlexTreeNode} that wraps a single {@link FlexTreeField} (which is placed under the {@link EmptyKey}).
 *
 * @remarks
 * A FieldNode is mostly identical to a struct node with a single field using the {@link EmptyKey}, but provides access to it via a field named "content".
 *
 * There are several use-cases where it makes sense to use a field node.
 * Here are a few:
 * - When it's necessary to differentiate between an empty sequence, and no sequence.
 * One case where this is needed is encoding Json.
 * - When polymorphism over {@link TreeFieldSchema} (and not just a union of {@link AllowedTypes}) is required.
 * For example when encoding a schema for a type like
 * `Foo[] | Bar[]`, `Foo | Foo[]` or `Optional<Foo> | Optional<Bar>` (Where `Optional` is the Optional field kind, not TypeScript's `Optional`).
 * Since this schema system only allows `|` of {@link TreeNodeSchema} (and only when declaring a {@link TreeFieldSchema}), see {@link SchemaBuilderBase.field},
 * these aggregate types are most simply expressed by creating fieldNodes for the terms like `Foo[]`, and `Optional<Foo>`.
 * Note that these are distinct from types like `(Foo | Bar)[]` and `Optional<Foo | Bar>` which can be expressed as single fields without extra nodes.
 * - When a distinct merge identity is desired for a field.
 * For example, if the application wants to be able to have an optional node or a sequence which it can pass around, edit and observe changes to,
 * in some cases (like when the content is moved to a different parent) this can be more flexible if a field node is introduced
 * to create a separate logical entity (node) which wraps the field.
 * This can even be useful with value fields to wrap terminal nodes if a stable merge
 * - When a field (such as a {@link FlexTreeSequenceField}) is desired in a location where {@link FlexTreeNode}s are required
 * (like the member of a union or the child of another {@link FlexTreeField}).
 * This can is typically just a different perspective on one of the above cases.
 * For example:
 * `Sequence<Foo> | Sequence<Bar>` or `OptionalField<Sequence<Foo>>` can't be expressed as simple fields
 * (unlike `Sequence<Foo | Bar>` or `OptionalField<Foo>` which can be done as simple fields).
 * Instead {@link FlexTreeFieldNode}s can be use to achieve something similar, more like:
 * `FieldNode<Sequence<Foo>> | FieldNode<Sequence<Bar>>` or `OptionalField<FieldNode<Sequence<Foo>>>`.
 *
 * @privateRemarks
 * FieldNodes do not unbox to their content, so in schema aware APIs which do unboxing, the FieldNode will NOT be skipped over.
 * This is a change from the old behavior to simplify unboxing and prevent cases where arbitrary deep chains of field nodes could unbox omitting information about the tree depth.
 * @internal
 */
export declare interface FlexTreeFieldNode<in out TSchema extends FieldNodeSchema> extends FlexTreeNode {
    readonly schema: TSchema;
    /**
     * The content this field node wraps.
     * @remarks
     * This is a version of {@link FlexTreeFieldNode.boxedContent} but does unboxing.
     * Since field node are usually used to wrap fields which don't do unboxing (like {@link FlexTreeSequenceField})
     */
    readonly content: FlexTreeUnboxField<TSchema["info"]>;
    /**
     * The field this field node wraps.
     *
     * @remarks
     * Since field nodes are usually used to wrap fields which don't do unboxing (like {@link FlexTreeSequenceField}),
     * this is usually the same as {@link FlexTreeFieldNode.content}.
     * This is also the same as `[...this][0]`.
     */
    readonly boxedContent: FlexTreeTypedField<TSchema["info"]>;
}

/**
 * Leaf holding a value.
 *
 * @remarks
 * Leaves are immutable and have no children.
 * Leaf unboxes its content, so in schema aware APIs which do unboxing, the Leaf itself will be skipped over and its value will be returned directly.
 * @internal
 */
export declare interface FlexTreeLeafNode<in out TSchema extends LeafNodeSchema> extends FlexTreeNode {
    readonly schema: TSchema;
    /**
     * Value stored on this node.
     */
    readonly value: TreeValue<TSchema["info"]>;
}

/**
 * A {@link FlexTreeNode} that behaves like a `Map<string, Field>` for a specific `Field` type.
 *
 * @remarks
 * Unlike TypeScript Map type, {@link FlexTreeMapNode.get} always provides a reference to any field looked up, even if it has never been set.
 *
 * This means that, for example, a `MapNode` of {@link FlexTreeSequenceField} fields will return an empty sequence when a previously unused key is looked up,
 * and that sequence can be used to insert new items into the field.
 * Additionally empty fields (those containing no nodes) are not distinguished from fields which do not exist.
 * This differs from JavaScript Maps which have a subtle distinction between storing undefined as a value in the map and deleting an entry from the map.
 *
 * @internal
 */
export declare interface FlexTreeMapNode<in out TSchema extends MapNodeSchema> extends FlexTreeNode {
    readonly schema: TSchema;
    /**
     * The number of elements in the map.
     *
     * @remarks
     * All fields under a map implicitly exist, but `size` will count only the fields which contain one or more nodes.
     */
    readonly size: number;
    /**
     * Checks whether a value exists for the given key.
     * @param key - Which map entry to look up.
     *
     * @remarks
     * All fields under a map implicitly exist, but `has` will only return true if there are one or more nodes present in the given field.
     */
    has(key: string): boolean;
    /**
     * Get the value associated with `key`.
     * @param key - which map entry to look up.
     */
    get(key: string): FlexTreeUnboxField<TSchema["info"]>;
    /**
     * Get the field for `key`.
     * @param key - which map entry to look up.
     *
     * @remarks
     * All fields under a map implicitly exist, so `get` can be called with any key and will always return a field.
     * Even if the field is empty, it will still be returned, and can be edited to insert content into the map.
     */
    getBoxed(key: string): FlexTreeTypedField<TSchema["info"]>;
    /**
     * Returns an iterable of keys in the map.
     *
     * @remarks
     * All fields under a map implicitly exist, but `keys` will yield only the keys of fields which contain one or more nodes.
     */
    keys(): IterableIterator<FieldKey>;
    /**
     * Returns an iterable of values in the map.
     *
     * @remarks
     * All fields under a map implicitly exist, but `values` will yield only the fields containing one or more nodes.
     */
    values(): IterableIterator<FlexTreeUnboxField<TSchema["info"], "notEmpty">>;
    /**
     * Returns an iterable of key, value pairs for every entry in the map.
     *
     * @remarks
     * All fields under a map implicitly exist, but `entries` will yield only the entries whose fields contain one or more nodes.
     *
     * This iteration provided by `entries()` is equivalent to that provided by direct iteration of the {@link FlexTreeMapNode} (a.k.a. `[Symbol.Iterator]()`).
     */
    entries(): IterableIterator<[FieldKey, FlexTreeUnboxField<TSchema["info"], "notEmpty">]>;
    /**
     * Executes a provided function once per each key/value pair in the map.
     * @param callbackFn - The function to run for each map entry
     * @param thisArg - If present, `callbackFn` will be bound to `thisArg`
     *
     * @privateRemarks
     * TODO: This should run over fields in insertion order if we want to match the javascript foreach spec.
     */
    forEach(callbackFn: (value: FlexTreeUnboxField<TSchema["info"], "notEmpty">, key: FieldKey, map: FlexTreeMapNode<TSchema>) => void, thisArg?: any): void;
    /**
     * Adds or updates an entry in the map with a specified `key` and a `value`.
     *
     * @param key - The key of the element to add to the map.
     * @param value - The value of the element to add to the map.
     */
    set(key: string, value: FlexibleFieldContent<TSchema["info"]>): void;
    /**
     * Removes the specified element from this map by its `key`.
     *
     * @remarks
     * Note: unlike JavaScript's Map API, this method does not return a flag indicating whether or not the value was
     * deleted.
     *
     * @privateRemarks
     * Regarding the choice to not return a boolean: Since this data structure is distributed in nature, it isn't
     * possible to tell whether or not the item was deleted as a result of this method call. Returning a "best guess"
     * is more likely to create issues / promote bad usage patterns than offer useful information.
     *
     * @param key - The key of the element to remove from the map.
     */
    delete(key: string): void;
    /**
     * Iterate through all fields in the map.
     *
     * @remarks
     * No mutations to the current view of the shared tree are permitted during iteration.
     * To iterate over the unboxed values of the map, use `Symbol.Iterator()`.
     */
    [boxedIterator](): IterableIterator<FlexTreeTypedField<TSchema["info"]>>;
    [Symbol.iterator](): IterableIterator<[
    FieldKey,
    FlexTreeUnboxField<TSchema["info"], "notEmpty">
    ]>;
    /**
     * An enumerable own property which allows JavaScript object traversals to access {@link FlexTreeSequenceField} content.
     * It is recommenced to NOT use this when possible (for performance and type safety reasons): instead use {@link FlexTreeMapNode.get} or iterate over fields with `Symbol.iterator`.
     * See [ReadMe](./README.md) for details.
     *
     * This object is not guaranteed to be kept up to date across edits and thus should not be held onto across edits.
     */
    readonly asObject: {
        readonly [P in FieldKey]?: FlexTreeUnboxField<TSchema["info"], "notEmpty">;
    };
}

/**
 * Indicates that an object is a flex tree.
 * @internal
 */
export declare const flexTreeMarker: unique symbol;

/**
 * Generic tree node API.
 *
 * Nodes are (shallowly) immutable and have a logical identity, a type and either a value or fields under string keys.
 *
 * This "logical identity" is exposed as the object identity: if a node is moved within a document,
 * the same {@link FlexTreeNode} instance will be used in the new location.
 * Similarly, edits applied to a node's sub-tree concurrently with the move of the node will still be applied to its subtree in its new location.
 *
 *
 * @remarks
 * Down-casting (via {@link FlexTreeNode#is}) is required to access Schema-Aware APIs, including editing.
 * All content in the tree is accessible without down-casting, but if the schema is known,
 * the schema aware API may be more ergonomic.
 * All editing is actually done via {@link FlexTreeField}s: the nodes are immutable other than that they contain mutable fields.
 *
 * @internal
 */
export declare interface FlexTreeNode extends FlexTreeEntity<FlexTreeNodeSchema> {
    readonly [flexTreeMarker]: FlexTreeEntityKind.Node;
    /**
     * Value stored on this node.
     */
    readonly value?: TreeValue;
    /**
     * {@inheritDoc ISubscribable#on}
     */
    on<K extends keyof EditableTreeEvents>(eventName: K, listener: EditableTreeEvents[K]): () => void;
    /**
     * Gets a field of this node, if it is not empty.
     */
    tryGetField(key: FieldKey): undefined | FlexTreeField;
    /**
     * The field this tree is in, and the index within that field.
     */
    readonly parentField: {
        readonly parent: FlexTreeField;
        readonly index: number;
    };
    /**
     * Type guard for narrowing / down-casting to a specific schema.
     */
    is<TSchema extends FlexTreeNodeSchema>(schema: TSchema): this is FlexTreeTypedNode<TSchema>;
    /**
     * Same as `this.schema.name`.
     * This is provided as an enumerable own property to aid with JavaScript object traversals of this data-structure.
     * See [ReadMe](./README.md) for details.
     */
    readonly type: TreeNodeSchemaIdentifier;
    [boxedIterator](): IterableIterator<FlexTreeField>;
    /**
     * Subscribe to the next change that affects this node's children.
     * @returns a function which will deregister the registered event.
     * It has no effect if the event was already deregistered.
     * @remarks
     * The given function will be run the next time that this node's direct children change.
     * It will only be run once, and thereafter automatically deregistered.
     * It does not run in response to changes beneath this node's direct children.
     * This event fires after the tree has been mutated but before {@link EditableTreeEvents.afterChange}.
     * Only one subscriber may register to this event at the same time.
     * @privateRemarks
     * This event allows the proxy-based API that is built on top of the editable tree to maintain invariants
     * around "hydrating" proxies that were created with schema-provided factory functions.
     * It is not a public API and thus the symbol for this property is not exported.
     */
    [onNextChange](fn: (node: FlexTreeNode) => void): () => void;
}

/**
 * Field that contains an immutable {@link StableNodeKey} identifying this node.
 * @internal
 */
export declare interface FlexTreeNodeKeyField extends FlexTreeField {
    readonly localNodeKey: LocalNodeKey;
    readonly stableNodeKey: StableNodeKey;
}

/**
 * @internal
 * @privateRemarks
 * This could be an exhaustive union, or just the common base type.
 * Using just the base type prevents exhaustive matching, which has both pros and cons.
 *
 * For now this is using just the base type since the union is causing issues with schema aware typing, likely due to it being a union and thus distributing over extends clauses.
 */
export declare type FlexTreeNodeSchema = TreeNodeSchemaBase;

/**
 * A {@link FlexTreeNode} that behaves like an "object" or "struct", providing properties to access its fields.
 *
 * ObjectNodes consist of a finite collection of fields, each with their own (distinct) key and {@link TreeFieldSchema}.
 *
 * @remarks
 * ObjectNodes require complex typing, and have been split into two parts for implementation purposes.
 * See {@link FlexTreeObjectNodeTyped} for the schema aware extensions to this that provide access to the fields.
 *
 * These "Objects" resemble "Structs" from a wide variety of programming languages
 * (Including Algol 68, C, Go, Rust, C# etc.).
 * ObjectNodes also somewhat resemble JavaScript objects: this analogy is less precise (objects don't have a fixed schema for example),
 * but for consistency with other systems in the JavaScript ecosystem (like JSON) is "ObjectNodes" nodes are named "Objects".
 *
 * Another common name for this abstraction is [record](https://en.wikipedia.org/wiki/Record_(computer_science)).
 * The name "Record" is avoided (in favor of Object) here because it has less precise connotations for most TypeScript developers.
 * For example, TypeScript has a built in `Record` type, but it requires all of the fields to have the same type,
 * putting its semantics half way between this library's "Object" schema and {@link FlexTreeMapNode}.
 *
 * @internal
 */
export declare interface FlexTreeObjectNode extends FlexTreeNode {
    readonly schema: ObjectNodeSchema;
    /**
     * {@link LocalNodeKey} that identifies this node.
     */
    readonly localNodeKey?: LocalNodeKey;
}

/**
 * Properties to access an object node's fields. See {@link FlexTreeObjectNodeTyped}.
 *
 * @privateRemarks TODOs:
 *
 * 1. Support custom field keys.
 *
 * 2. Do we keep assignment operator + "setFoo" methods, or just use methods?
 * Inconsistency in the API experience could confusing for consumers.
 *
 * @internal
 */
export declare type FlexTreeObjectNodeFields<TFields extends Fields> = FlattenKeys<{
    readonly [key in keyof TFields as `boxed${Capitalize<key & string>}`]: FlexTreeTypedField<TFields[key]>;
} & {
    readonly [key in keyof TFields as TFields[key]["kind"] extends AssignableFieldKinds ? never : key]: FlexTreeUnboxField<TFields[key]>;
} & {
    -readonly [key in keyof TFields as TFields[key]["kind"] extends AssignableFieldKinds ? key : never]: FlexTreeUnboxField<TFields[key]>;
} & {
    readonly [key in keyof TFields as TFields[key]["kind"] extends AssignableFieldKinds ? `set${Capitalize<key & string>}` : never]: (content: FlexibleFieldContent<TFields[key]>) => void;
}>;

/**
 * An {@link FlexTreeObjectNode} with schema aware accessors for its fields.
 *
 * @privateRemarks
 *
 * The corresponding implementation logic for this lives in `LazyTree.ts` under `buildStructClass`.
 * If you change the signature here, you will need to update that logic to match.
 *
 * @internal
 */
export declare type FlexTreeObjectNodeTyped<TSchema extends ObjectNodeSchema> = ObjectNodeSchema extends TSchema ? FlexTreeObjectNode : FlexTreeObjectNode & FlexTreeObjectNodeFields<TSchema["info"]>;

/**
 * Field that stores zero or one child.
 *
 * @remarks
 * Unboxes its content, so in schema aware APIs which do unboxing, the OptionalField itself will be skipped over and its content will be returned directly.
 *
 * @privateRemarks
 * TODO: Document merge semitics
 * TODO: Allow Optional fields to be used with last write wins OR first write wins merge resolution.
 * TODO:
 * Better centralize the documentation about what kinds of merge semantics are available for field kinds.
 * Maybe link editor?
 * @internal
 */
export declare interface FlexTreeOptionalField<in out TTypes extends AllowedTypes> extends FlexTreeField {
    get content(): FlexTreeUnboxNodeUnion<TTypes> | undefined;
    set content(newContent: FlexibleNodeContent<TTypes> | undefined);
    readonly boxedContent?: FlexTreeTypedNodeUnion<TTypes>;
}

/**
 * Field that stores exactly one child.
 *
 * @remarks
 * Unboxes its content, so in schema aware APIs which do unboxing, the RequiredField itself will be skipped over and its content will be returned directly.
 * @internal
 */
export declare interface FlexTreeRequiredField<in out TTypes extends AllowedTypes> extends FlexTreeField {
    get content(): FlexTreeUnboxNodeUnion<TTypes>;
    set content(content: FlexibleNodeContent<TTypes>);
    readonly boxedContent: FlexTreeTypedNodeUnion<TTypes>;
}

/**
 * Schema data that can be be used to view a document.
 * Strongly typed over its rootFieldSchema.
 *
 * @remarks
 * The type of the rootFieldSchema is used to implement SchemaAware APIs.
 * Cases that do not require being compile time schema aware can omit the explicit type for it.
 *
 * @internal
 */
export declare interface FlexTreeSchema<out T extends TreeFieldSchema = TreeFieldSchema> extends SchemaCollection {
    /**
     * Schema for the root field which contains the whole tree.
     */
    readonly rootFieldSchema: T;
    /**
     * Extra configuration for how this schema is handled at runtime.
     */
    readonly policy: FullSchemaPolicy;
    /**
     * Compatibility information how how to interact with content who's stored schema is not directly compatible with this schema.
     */
    readonly adapters: Adapters;
}

/**
 * {@link FlexTreeField} that stores a sequence of children.
 *
 * Sequence fields can contain an ordered sequence any number of {@link FlexTreeNode}s which must be of the {@link AllowedTypes} from the {@link TreeFieldSchema}).
 *
 * @remarks
 * Allows for concurrent editing based on index, adjusting the locations of indexes as needed so they apply to the same logical place in the sequence when rebased and merged.
 *
 * Edits to sequence fields are anchored relative to their surroundings, so concurrent edits can result in the indexes of nodes and edits getting shifted.
 * To hold onto locations in sequence across an edit, use anchors.
 *
 * @privateRemarks
 * TODO:
 * Add anchor API that can actually hold onto locations in a sequence.
 * Currently only nodes can be held onto with anchors, and this does not replicate the behavior implemented for editing.
 * @internal
 */
export declare interface FlexTreeSequenceField<in out TTypes extends AllowedTypes> extends FlexTreeField {
    /**
     * Gets a node of this field by its index with unboxing.
     * @param index - Zero-based index of the item to retrieve. Negative values are interpreted from the end of the sequence.
     *
     * @returns The element in the sequence matching the given index. Always returns undefined if index \< -sequence.length
     * or index \>= array.length.
     *
     * @remarks
     * Semantics match {@link https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/at | Array.at}.
     */
    at(index: number): FlexTreeUnboxNodeUnion<TTypes> | undefined;
    /**
     * Gets a node of this field by its index without unboxing.
     * @param index - Zero-based index of the item to retrieve. Negative values are interpreted from the end of the sequence.
     *
     * @returns The element in the sequence matching the given index. Always returns undefined if index \< -sequence.length
     * or index \>= array.length.
     *
     * @remarks
     * Semantics match {@link https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/at | Array.at}.
     */
    boxedAt(index: number): FlexTreeTypedNodeUnion<TTypes> | undefined;
    /**
     * Calls the provided callback function on each child of this sequence, and returns an array that contains the results.
     * @param callbackfn - A function that accepts the child and its index.
     */
    map<U>(callbackfn: (value: FlexTreeUnboxNodeUnion<TTypes>, index: number) => U): U[];
    /**
     * Calls the provided callback function on each child of this sequence, and returns an array that contains the results.
     * @param callbackfn - A function that accepts the child and its index.
     */
    mapBoxed<U>(callbackfn: (value: FlexTreeTypedNodeUnion<TTypes>, index: number) => U): U[];
    readonly length: number;
    /**
     * Inserts new item(s) at a specified location.
     * @param index - The index at which to insert `value`.
     * @param value - The content to insert.
     * @throws Throws if `index` is not in the range [0, `list.length`).
     */
    insertAt(index: number, value: FlexibleNodeSubSequence<TTypes>): void;
    /**
     * Inserts new item(s) at the start of the sequence.
     * @param value - The content to insert.
     */
    insertAtStart(value: FlexibleNodeSubSequence<TTypes>): void;
    /**
     * Inserts new item(s) at the end of the sequence.
     * @param value - The content to insert.
     */
    insertAtEnd(value: FlexibleNodeSubSequence<TTypes>): void;
    /**
     * Removes the item at the specified location.
     * @param index - The index at which to remove the item.
     * @throws Throws if `index` is not in the range [0, `list.length`).
     */
    removeAt(index: number): void;
    /**
     * Removes all items between the specified indices.
     * @param start - The starting index of the range to remove (inclusive). Defaults to the start of the sequence.
     * @param end - The ending index of the range to remove (exclusive).
     * @throws Throws if `start` is not in the range [0, `list.length`).
     * @throws Throws if `end` is less than `start`.
     * If `end` is not supplied or is greater than the length of the sequence, all items after `start` are deleted.
     */
    removeRange(start?: number, end?: number): void;
    /**
     * Moves the specified item to the start of the sequence.
     * @param sourceIndex - The index of the item to move.
     * @throws Throws if `sourceIndex` is not in the range [0, `list.length`).
     */
    moveToStart(sourceIndex: number): void;
    /**
     * Moves the specified item to the start of the sequence.
     * @param sourceIndex - The index of the item to move.
     * @param source - The source sequence to move the item out of.
     * @throws Throws if `sourceIndex` is not in the range [0, `list.length`).
     */
    moveToStart(sourceIndex: number, source: FlexTreeSequenceField<AllowedTypes>): void;
    /**
     * Moves the specified item to the end of the sequence.
     * @param sourceIndex - The index of the item to move.
     * @throws Throws if `sourceIndex` is not in the range [0, `list.length`).
     */
    moveToEnd(sourceIndex: number): void;
    /**
     * Moves the specified item to the end of the sequence.
     * @param sourceIndex - The index of the item to move.
     * @param source - The source sequence to move the item out of.
     * @throws Throws if `sourceIndex` is not in the range [0, `list.length`).
     */
    moveToEnd(sourceIndex: number, source: FlexTreeSequenceField<AllowedTypes>): void;
    /**
     * Moves the specified item to the desired location in the sequence.
     * @param index - The index to move the item to.
     * This is based on the state of the sequence before moving the source item.
     * @param sourceIndex - The index of the item to move.
     * @throws Throws if any of the input indices are not in the range [0, `list.length`).
     */
    moveToIndex(index: number, sourceIndex: number): void;
    /**
     * Moves the specified item to the desired location in the sequence.
     * @param index - The index to move the item to.
     * @param sourceIndex - The index of the item to move.
     * @param source - The source sequence to move the item out of.
     * @throws Throws if any of the input indices are not in the range [0, `list.length`).
     */
    moveToIndex(index: number, sourceIndex: number, source: FlexTreeSequenceField<AllowedTypes>): void;
    /**
     * Moves the specified items to the start of the sequence.
     * @param sourceStart - The starting index of the range to move (inclusive).
     * @param sourceEnd - The ending index of the range to move (exclusive)
     * @throws Throws if either of the input indices are not in the range [0, `list.length`) or if `sourceStart` is greater than `sourceEnd`.
     */
    moveRangeToStart(sourceStart: number, sourceEnd: number): void;
    /**
     * Moves the specified items to the start of the sequence.
     * @param sourceStart - The starting index of the range to move (inclusive).
     * @param sourceEnd - The ending index of the range to move (exclusive)
     * @param source - The source sequence to move items out of.
     * @throws Throws if the types of any of the items being moved are not allowed in the destination sequence,
     * if either of the input indices are not in the range [0, `list.length`) or if `sourceStart` is greater than `sourceEnd`.
     */
    moveRangeToStart(sourceStart: number, sourceEnd: number, source: FlexTreeSequenceField<AllowedTypes>): void;
    /**
     * Moves the specified items to the end of the sequence.
     * @param sourceStart - The starting index of the range to move (inclusive).
     * @param sourceEnd - The ending index of the range to move (exclusive)
     * @throws Throws if either of the input indices are not in the range [0, `list.length`) or if `sourceStart` is greater than `sourceEnd`.
     */
    moveRangeToEnd(sourceStart: number, sourceEnd: number): void;
    /**
     * Moves the specified items to the end of the sequence.
     * @param sourceStart - The starting index of the range to move (inclusive).
     * @param sourceEnd - The ending index of the range to move (exclusive)
     * @param source - The source sequence to move items out of.
     * @throws Throws if the types of any of the items being moved are not allowed in the destination sequence,
     * if either of the input indices are not in the range [0, `list.length`) or if `sourceStart` is greater than `sourceEnd`.
     */
    moveRangeToEnd(sourceStart: number, sourceEnd: number, source: FlexTreeSequenceField<AllowedTypes>): void;
    /**
     * Moves the specified items to the desired location within the sequence.
     * @param index - The index to move the items to.
     * This is based on the state of the sequence before moving the source items.
     * @param sourceStart - The starting index of the range to move (inclusive).
     * @param sourceEnd - The ending index of the range to move (exclusive)
     * @throws Throws if any of the input indices are not in the range [0, `list.length`) or if `sourceStart` is greater than `sourceEnd`.
     */
    moveRangeToIndex(index: number, sourceStart: number, sourceEnd: number): void;
    /**
     * Moves the specified items to the desired location within the sequence.
     * @param index - The index to move the items to.
     * @param sourceStart - The starting index of the range to move (inclusive).
     * @param sourceEnd - The ending index of the range to move (exclusive)
     * @param source - The source sequence to move items out of.
     * @throws Throws if the types of any of the items being moved are not allowed in the destination sequence,
     * if any of the input indices are not in the range [0, `list.length`) or if `sourceStart` is greater than `sourceEnd`.
     */
    moveRangeToIndex(index: number, sourceStart: number, sourceEnd: number, source: FlexTreeSequenceField<AllowedTypes>): void;
    [boxedIterator](): IterableIterator<FlexTreeTypedNodeUnion<TTypes>>;
    [Symbol.iterator](): IterableIterator<FlexTreeUnboxNodeUnion<TTypes>>;
    /**
     * An enumerable own property which allows JavaScript object traversals to access {@link FlexTreeSequenceField} content.
     * It is recommenced to NOT use this when possible (for performance and type safety reasons): instead use {@link FlexTreeSequenceField#at} or iterate over nodes with `Symbol.iterator`.
     * See [ReadMe](./README.md) for details.
     *
     * This array is not guaranteed to be kept up to date across edits and thus should not be held onto across edits.
     */
    readonly asArray: readonly FlexTreeUnboxNodeUnion<TTypes>[];
}

/**
 * Schema aware specialization of {@link FlexTreeField}.
 * @internal
 */
export declare type FlexTreeTypedField<TSchema extends TreeFieldSchema> = FlexTreeTypedFieldInner<TSchema["kind"], TSchema["allowedTypes"]>;

/**
 * Helper for implementing {@link FlexTreeTypedField}.
 * @internal
 */
export declare type FlexTreeTypedFieldInner<Kind extends FieldKind, Types extends AllowedTypes> = Kind extends typeof FieldKinds.sequence ? FlexTreeSequenceField<Types> : Kind extends typeof FieldKinds.required ? FlexTreeRequiredField<Types> : Kind extends typeof FieldKinds.optional ? FlexTreeOptionalField<Types> : Kind extends typeof FieldKinds.nodeKey ? FlexTreeNodeKeyField : FlexTreeField;

/**
 * Schema aware specialization of {@link FlexTreeNode} for a given {@link TreeNodeSchema}.
 * @internal
 */
export declare type FlexTreeTypedNode<TSchema extends FlexTreeNodeSchema> = TSchema extends LeafNodeSchema ? FlexTreeLeafNode<TSchema> : TSchema extends MapNodeSchema ? FlexTreeMapNode<TSchema> : TSchema extends FieldNodeSchema ? FlexTreeFieldNode<TSchema> : TSchema extends ObjectNodeSchema ? FlexTreeObjectNodeTyped<TSchema> : FlexTreeNode;

/**
 * Schema aware specialization of {@link FlexTreeNode} for a given {@link AllowedTypes}.
 * @internal
 */
export declare type FlexTreeTypedNodeUnion<T extends AllowedTypes> = T extends FlexList<FlexTreeNodeSchema> ? FlexTreeTypedNode<Assume<FlexListToUnion<T>, FlexTreeNodeSchema>> : FlexTreeNode;

/**
 * Schema aware unboxed field.
 * @remarks
 * Unboxes fields to their content if appropriate for the kind.
 * Recursively unboxes that content (then its content etc.) as well if the node union does unboxing.
 * @internal
 */
export declare type FlexTreeUnboxField<TSchema extends TreeFieldSchema, Emptiness extends "maybeEmpty" | "notEmpty" = "maybeEmpty"> = FlexTreeUnboxFieldInner<TSchema["kind"], TSchema["allowedTypes"], Emptiness>;

/**
 * Helper for implementing FlexTreeUnboxField.
 * @internal
 */
export declare type FlexTreeUnboxFieldInner<Kind extends FieldKind, TTypes extends AllowedTypes, Emptiness extends "maybeEmpty" | "notEmpty"> = Kind extends typeof FieldKinds.sequence ? FlexTreeSequenceField<TTypes> : Kind extends typeof FieldKinds.required ? FlexTreeUnboxNodeUnion<TTypes> : Kind extends typeof FieldKinds.optional ? FlexTreeUnboxNodeUnion<TTypes> | (Emptiness extends "notEmpty" ? never : undefined) : Kind extends typeof FieldKinds.nodeKey ? FlexTreeNodeKeyField : unknown;

/**
 * Schema aware unboxed tree type.
 * @remarks
 * Unboxes if the node kind does unboxing.
 * Recursively unboxes that content as well if it does unboxing.
 * @internal
 */
export declare type FlexTreeUnboxNode<TSchema extends FlexTreeNodeSchema> = TSchema extends LeafNodeSchema ? TreeValue<TSchema["info"]> : TSchema extends MapNodeSchema ? FlexTreeMapNode<TSchema> : TSchema extends FieldNodeSchema ? FlexTreeFieldNode<TSchema> : TSchema extends ObjectNodeSchema ? FlexTreeObjectNodeTyped<TSchema> : FlexTreeUnknownUnboxed;

/**
 * Schema aware unboxed union of tree types.
 * @remarks
 * Unboxes when not polymorphic.
 * Recursively unboxes that content as well if the node kind does unboxing.
 * @internal
 */
export declare type FlexTreeUnboxNodeUnion<TTypes extends AllowedTypes> = TTypes extends readonly [
InternalTypedSchemaTypes.LazyItem<infer InnerType>
] ? InnerType extends FlexTreeNodeSchema ? FlexTreeUnboxNode<InnerType> : InnerType extends Any ? FlexTreeNode : unknown : boolean extends IsArrayOfOne<TTypes> ? FlexTreeUnknownUnboxed : FlexTreeTypedNodeUnion<TTypes>;

/**
 * Unboxed tree type for unknown schema cases.
 * @internal
 */
export declare type FlexTreeUnknownUnboxed = TreeValue | FlexTreeNode;

/**
 * An editable view of a (version control style) branch of a shared tree.
 * @privateRemarks
 * TODO:
 * 1. Once ISharedTreeView is renamed this can become ISharedTreeView.
 * 2. This object should be combined with or accessible from the TreeContext to allow easy access to thinks like branching.
 * @internal
 */
export declare interface FlexTreeView<in out TRoot extends TreeFieldSchema> extends IDisposable {
    /**
     * Context for controlling the EditableTree nodes produced from {@link FlexTreeView.editableTree}.
     *
     * @remarks
     * This is an owning reference: disposing of this view disposes its context.
     */
    readonly context: TreeContext;
    /**
     * Access non-view schema specific aspects of of this branch.
     *
     * @remarks
     * This is a non-owning reference: disposing of this view does not impact the branch.
     */
    readonly checkout: ITreeCheckout;
    /**
     * Get a typed view of the tree content using the editable-tree-2 API.
     */
    readonly editableTree: FlexTreeTypedField<TRoot>;
    /**
     * Spawn a new view which is based off of the current state of this view.
     * Any mutations of the new view will not apply to this view until the new view is merged back into this view via `merge()`.
     */
    fork(): ITreeViewFork<TRoot>;
}

/**
 * @internal
 */
export declare interface Forbidden extends FieldKind<typeof forbiddenFieldKindIdentifier, Multiplicity.Forbidden> {
}

/**
 * Identifier used for the FieldKind for fields which must be empty.
 *
 * @remarks
 * This mainly show up in:
 * 1. The root default field for documents.
 * 2. The schema used for out of schema fields (which thus must be empty/not exist) on object and leaf nodes.
 *
 * @internal
 */
export declare const forbiddenFieldKindIdentifier = "Forbidden";

/**
 * APIs for forest designed so the implementation can be copy on write,
 * or mutate in place, and we can ensure no references are dangling into the forest to allow this.
 *
 * This results in rather manual memory management,
 * but makes it practical to provide highly optimized implementations,
 * for example WASM powered binary formats that can track reference counts and only copy when needed.
 */
/**
 * Events for {@link IForestSubscription}.
 *
 * TODO: consider having before and after events per subtree instead while applying anchor (and this just shows what happens at the root).
 * @internal
 */
export declare interface ForestEvents {
    /**
     * The forest is about to be changed.
     * Emitted before the first change in a batch of changes.
     */
    beforeChange(): void;
    /**
     * The forest was just changed.
     * Emitted after the last change in a batch of changes.
     */
    afterChange(): void;
}

/**
 * Ways to refer to a node in an IEditableForest.
 * @internal
 */
export declare type ForestLocation = ITreeSubscriptionCursor | Anchor;

/**
 * Used to distinguish between different forest types.
 * @internal
 */
export declare enum ForestType {
    /**
     * The "ObjectForest" forest type.
     */
    Reference = 0,
    /**
     * The "ChunkedForest" forest type.
     */
    Optimized = 1
}

/**
 * Policy from the app for interpreting the stored schema.
 * The app must ensure consistency for all users of the document.
 * @internal
 */
export declare interface FullSchemaPolicy {
    /**
     * Policy information about FieldKinds:
     * This is typically stored as code, not in documents, and defines how to handles fields based on their kind.
     * It is assumed that all users of a document will have exactly the same FieldKind policies,
     * though older applications might be missing some,
     * and will be unable to process any changes that use those FieldKinds.
     */
    readonly fieldKinds: ReadonlyMap<FieldKindIdentifier, FieldKind>;
}

/**
 * Json comparable field collection, generic over child type.
 * Json compatibility assumes `TChild` is also json compatible.
 * @internal
 */
export declare interface GenericFieldsNode<TChild> {
    fields?: FieldMapObject<TChild>;
}

/**
 * Json comparable tree node, generic over child type.
 * Json compatibility assumes `TChild` is also json compatible.
 * @internal
 */
export declare interface GenericTreeNode<TChild> extends GenericFieldsNode<TChild>, NodeData {
}

/**
 * @internal
 */
export declare interface HasListeners<E extends Events<E>> {
    /**
     * When no `eventName` is provided, returns true iff there are any listeners.
     *
     * When `eventName` is provided, returns true iff there are listeners for that event.
     *
     * @remarks
     * This can be used to know when its safe to cleanup data-structures which only exist to fire events for their listeners.
     */
    hasListeners(eventName?: keyof Events<E>): boolean;
}

/**
 * Options relating to handling of persisted data.
 * @internal
 */
export declare interface ICodecOptions {
    /**
     * {@link JsonValidator} which SharedTree uses to validate persisted data it reads & writes
     * matches the expected encoded format (i.e. the wire format for ops and summaries).
     *
     * See {@link noopValidator} and {@link typeboxValidator} for out-of-the-box implementations.
     *
     * This option is not "on-by-default" because JSON schema validation comes with a small but noticeable
     * runtime performance cost, and popular schema validation libraries have relatively large bundle size.
     *
     * SharedTree users are still encouraged to use a non-trivial validator (i.e. not `noopValidator`)
     * whenever reasonable: it gives better fail-fast behavior when unexpected encoded data is found,
     * which reduces the risk of unrecoverable data corruption.
     */
    readonly jsonValidator: JsonValidator;
}

/**
 * Used for allocating IDs unique to a particular instance of the allocator.
 * @internal
 */
export declare interface IdAllocator<TId = number> {
    /**
     * Allocates a block of `count` consecutive IDs and returns the first ID in the block.
     * For convenience can be called with no parameters to allocate a single ID.
     */
    allocate: (count?: number) => TId;
    /**
     * @returns The next ID that will be allocated by this allocator.
     */
    getNextId: () => TId;
}

/**
 * Default editor for transactional tree data changes.
 * @internal
 */
export declare interface IDefaultEditBuilder {
    /**
     * @param field - the value field which is being edited under the parent node
     * @returns An object with methods to edit the given field of the given parent.
     * The returned object can be used (i.e., have its methods called) multiple times but its lifetime
     * is bounded by the lifetime of this edit builder.
     */
    valueField(field: FieldUpPath): ValueFieldEditBuilder;
    /**
     * @param field - the optional field which is being edited under the parent node
     * @returns An object with methods to edit the given field of the given parent.
     * The returned object can be used (i.e., have its methods called) multiple times but its lifetime
     * is bounded by the lifetime of this edit builder.
     */
    optionalField(field: FieldUpPath): OptionalFieldEditBuilder;
    /**
     * @param field - the sequence field which is being edited under the parent node
     *
     * @returns An object with methods to edit the given field of the given parent.
     * The returned object can be used (i.e., have its methods called) multiple times but its lifetime
     * is bounded by the lifetime of this edit builder.
     */
    sequenceField(field: FieldUpPath): SequenceFieldEditBuilder;
    /**
     * Moves a subsequence from one sequence field to another sequence field.
     *
     * Note that the `destinationIndex` is interpreted based on the state of the sequence *before* the move operation.
     * For example, `move(field, 0, 1, field, 2)` changes `[A, B, C]` to `[B, A, C]`.
     */
    move(sourceField: FieldUpPath, sourceIndex: number, count: number, destinationField: FieldUpPath, destinationIndex: number): void;
    addNodeExistsConstraint(path: UpPath): void;
}

/**
 * An object with an explicit lifetime that can be ended.
 * @privateRemarks
 * TODO: align this with core-utils/IDisposable and {@link https://www.typescriptlang.org/docs/handbook/release-notes/typescript-5-2.html#using-declarations-and-explicit-resource-management| TypeScript's Disposable}.
 * @beta
 */
export declare interface IDisposable {
    /**
     * Call to end the lifetime of this object.
     *
     * It is invalid to use this object after this,
     * except for operations which explicitly document they are valid after disposal.
     *
     * @remarks
     * May cleanup resources retained by this object.
     * Often includes un-registering from events and thus preventing other objects from retaining a reference to this indefinably.
     *
     * Usually the only operations allowed after disposal are querying if an object is already disposed,
     * but this can vary between implementations.
     */
    [disposeSymbol](): void;
}

/**
 * Editing APIs.
 * @internal
 */
export declare interface IEditableForest extends IForestSubscription {
    /**
     * Provides a visitor that can be used to mutate the forest.
     *
     * @returns a visitor that can be used to mutate the forest.
     *
     * @remarks
     * Mutating the forest does NOT update anchors.
     * The visitor must be released after use by calling {@link DeltaVisitor.free} on it.
     * It is invalid to acquire a visitor without releasing the previous one.
     */
    acquireVisitor(): DeltaVisitor;
}

/**
 * Interface for an event emitter that can emit typed events to subscribed listeners.
 * @internal
 */
export declare interface IEmitter<E extends Events<E>> {
    /**
     * Emits an event with the specified name and arguments, notifying all subscribers by calling their registered listener functions.
     * @param eventName - the name of the event to fire
     * @param args - the arguments passed to the event listener functions
     */
    emit<K extends keyof Events<E>>(eventName: K, ...args: Parameters<E[K]>): void;
    /**
     * Emits an event with the specified name and arguments, notifying all subscribers by calling their registered listener functions.
     * It also collects the return values of all listeners into an array.
     *
     * Warning: This method should be used with caution. It deviates from the standard event-based integration pattern as creates substantial coupling between the emitter and its listeners.
     * For the majority of use-cases it is recommended to use the standard {@link IEmitter.emit} functionality.
     * @param eventName - the name of the event to fire
     * @param args - the arguments passed to the event listener functions
     * @returns An array of the return values of each listener, preserving the order listeners were called.
     */
    emitAndCollect<K extends keyof Events<E>>(eventName: K, ...args: Parameters<E[K]>): ReturnType<E[K]>[];
}

/**
 * Invalidates whenever the tree content changes.
 * For now (might change later) downloading new parts of the forest counts as a change.
 * Not invalidated when schema changes.
 *
 * When invalidating, all outstanding cursors must be freed or cleared.
 * @internal
 */
export declare interface IForestSubscription extends ISubscribable<ForestEvents> {
    /**
     * Set of anchors this forest is tracking.
     *
     * To keep these anchors usable, this AnchorSet must be updated / rebased for any changes made to the forest.
     * It is the responsibility of the caller of the forest-editing methods to do this, not the forest itself.
     * The caller performs these updates because it has more semantic knowledge about the edits, which can be needed to
     * update the anchors in a semantically optimal way.
     */
    readonly anchors: AnchorSet;
    /**
     * Create an independent copy of this forest, that uses the provided schema and anchors.
     *
     * The new copy will not invalidate observers (dependents) of the old one.
     */
    clone(schema: TreeStoredSchemaSubscription, anchors: AnchorSet): IEditableForest;
    /**
     * Allocates a cursor in the "cleared" state.
     */
    allocateCursor(): ITreeSubscriptionCursor;
    /**
     * Frees an Anchor, stopping tracking its position across edits.
     */
    forgetAnchor(anchor: Anchor): void;
    /**
     * It is an error not to free `cursorToMove` before the next edit.
     * Must provide a `cursorToMove` from this subscription (acquired via `allocateCursor`).
     */
    tryMoveCursorToNode(destination: Anchor, cursorToMove: ITreeSubscriptionCursor): TreeNavigationResult;
    /**
     * It is an error not to free `cursorToMove` before the next edit.
     * Must provide a `cursorToMove` from this subscription (acquired via `allocateCursor`).
     */
    tryMoveCursorToField(destination: FieldAnchor, cursorToMove: ITreeSubscriptionCursor): TreeNavigationResult;
    /**
     * Set `cursorToMove` to location described by path.
     * This is NOT a relative move: current position is discarded.
     * Path must point to existing node.
     */
    moveCursorToPath(destination: UpPath, cursorToMove: ITreeSubscriptionCursor): void;
    /**
     * The cursor is moved to a special dummy node above the detached fields.
     * This dummy node can be used to read the detached fields,
     * but other operations (such as inspecting the dummy node's type or path) should not be relied upon.
     * While this method does not return an {@link ITreeSubscriptionCursor}, similar restrictions apply to its use:
     * the returned cursor must not used after any edits are made to the forest.
     */
    getCursorAboveDetachedFields(): ITreeCursorSynchronous;
    /**
     * True if there are no nodes in the forest at all.
     *
     * @remarks
     * This means no nodes under any detached field, not just the special document root one.
     */
    readonly isEmpty: boolean;
}

/**
 * Generalized version of AllowedTypes allowing for more concise expressions in some cases.
 * @internal
 */
export declare type ImplicitAllowedTypes = AllowedTypes | FlexTreeNodeSchema | Any;

/**
 * Types allowed in a field.
 * @remarks
 * Implicitly treats a single type as an array of one type.
 * @beta
 */
declare type ImplicitAllowedTypes_2 = AllowedTypes_2 | TreeNodeSchema;

/**
 * Schema for a field of a tree node.
 * @remarks
 * Implicitly treats {@link ImplicitAllowedTypes} as a Required field of that type.
 * @beta
 */
export declare type ImplicitFieldSchema = FieldSchema | ImplicitAllowedTypes_2;

/**
 * Options used to initialize (if needed) and schematize a `SharedTree`.
 *
 * @internal
 */
export declare interface InitializeAndSchematizeConfiguration<TRoot extends TreeFieldSchema = TreeFieldSchema> extends TreeContent<TRoot>, SchematizeConfiguration<TRoot> {
}

/**
 * Use for trick to "inline" generic types.
 *
 * @remarks
 * The TypeScript compiler can be convinced to inline a generic type
 * (so the result of evaluating the generic type show up in IntelliSense and error messages instead of just the invocation of the generic type)
 * by creating an object with a field, and returning the type of that field.
 *
 * For example:
 * ```typescript
 * type MyGeneric<T1, T2> = {x: T1 extends [] ? T1 : T2 };
 * type MyGenericExpanded<T1, T2> = [{x: T1 extends [] ? T1 : T2 }][_InlineTrick]
 *
 * // Type is MyGeneric<5, string>
 * const foo: MyGeneric<5, string> = {x: "x"}
 * // Type is {x: "x"}
 * const foo2: MyGenericExpanded<5, string> = {x: "x"}
 * ```
 *
 * This constant is defined to provide a way to find this documentation from types which use this pattern,
 * and to locate types which use this pattern in case they need updating for compiler changes.
 * @internal
 */
export declare type _InlineTrick = 0;

/**
 * `TreeFieldSchema` to `TypedField`. May unwrap to child depending on FieldKind.
 * @internal
 */
export declare type InsertableFlexField<TField extends TreeFieldSchema> = [
ApplyMultiplicity<TField["kind"]["multiplicity"], AllowedTypesToFlexInsertableTree<TField["allowedTypes"]>>
][_InlineTrick];

/**
 * Generate a schema aware API for a single tree schema.
 * @internal
 */
export declare type InsertableFlexNode<TSchema extends FlexTreeNodeSchema> = FlattenKeys<CollectOptions<TSchema extends ObjectNodeSchema<string, infer TFields extends Fields> ? TypedFields<TFields> : TSchema extends FieldNodeSchema<string, infer TField extends TreeFieldSchema> ? InsertableFlexField<TField> : TSchema extends MapNodeSchema<string, infer TField extends TreeFieldSchema> ? {
    readonly [P in string]: InsertableFlexField<TField>;
} : EmptyObject, TSchema extends LeafNodeSchema<string, infer TValueSchema> ? TValueSchema : undefined, TSchema["name"]>>;

/**
 * Helper used to produce types for:
 *
 * 1. Insertable content which can be used to construct an object node.
 *
 * 2. Insertable content which is an unhydrated object node.
 *
 * 3. Union of 1 and 2.
 *
 * TODO: consider separating these cases into different types.
 * @beta
 */
declare type InsertableObjectFromSchemaRecord<T extends RestrictiveReadonlyRecord<string, ImplicitFieldSchema>> = {
    readonly [Property in keyof T]: InsertableTreeFieldFromImplicitField<T[Property]>;
};

/**
 * Type of content that can be inserted into the tree for a field of the given schema.
 * @beta
 */
declare type InsertableTreeFieldFromImplicitField<TSchema extends ImplicitFieldSchema = FieldSchema> = TSchema extends FieldSchema<infer Kind, infer Types> ? ApplyKind<InsertableTreeNodeFromImplicitAllowedTypes<Types>, Kind> : TSchema extends ImplicitAllowedTypes_2 ? InsertableTreeNodeFromImplicitAllowedTypes<TSchema> : unknown;

/**
 * Type of content that can be inserted into the tree for a node of the given schema.
 * @beta
 */
export declare type InsertableTreeNodeFromImplicitAllowedTypes<TSchema extends ImplicitAllowedTypes_2 = TreeNodeSchema> = TSchema extends TreeNodeSchema ? InsertableTypedNode<TSchema> : TSchema extends AllowedTypes_2 ? InsertableTypedNode<FlexListToUnion<TSchema>> : never;

/**
 * Data which can be used as a node to be inserted.
 * Either an unhydrated node, or content to build a new node.
 * @beta
 */
declare type InsertableTypedNode<T extends TreeNodeSchema> = (T extends {
    implicitlyConstructable: true;
} ? NodeBuilderData<T> : never) | Unhydrated<NodeFromSchema<T>>;

declare namespace InternalClassTreeTypes {
    export {
        ObjectFromSchemaRecord,
        InsertableObjectFromSchemaRecord,
        TreeNodeFromImplicitAllowedTypes,
        ImplicitAllowedTypes_2 as ImplicitAllowedTypes,
        FieldKind_2 as FieldKind,
        InsertableTreeFieldFromImplicitField,
        AllowedTypes_2 as AllowedTypes,
        FieldSchema,
        ApplyKind,
        InsertableTypedNode,
        NodeBuilderData
    }
}

declare namespace InternalTypedSchemaTypes {
    export {
        ArrayToUnion,
        FlexList,
        LazyItem,
        ExtractItemType,
        FlexListToUnion
    }
}
export { InternalTypedSchemaTypes }

declare namespace InternalTypes {
    export {
        MakeNominal,
        RestrictiveReadonlyRecord,
        NodeKind,
        TreeNodeSchemaClass,
        TreeNodeSchemaNonClass,
        TreeNodeSchemaCore,
        InternalClassTreeTypes,
        TreeArrayNodeBase
    }
}
export { InternalTypes }

/**
 * Constrain generic type parameters to Invariant.
 *
 * @example
 *
 * ```typescript
 * protected _typeCheck?: Invariant<T>;
 * ```
 *
 * @internal
 */
export declare interface Invariant<in out T> extends Contravariant<T>, Covariant<T> {
}

/**
 * Returns a type parameter that is true iff T is any.
 *
 * @privateRemarks
 * Only `never` is assignable to `never` (`any` isn't),
 * but `any` distributes over the `extends` here while nothing else should.
 * This can be used to detect `any`.
 * @internal
 */
export declare type isAny<T> = boolean extends (T extends never ? true : false) ? true : false;

/**
 * `true` if T is known to be an array of one item.
 * `false` if T is known not to be an array of one item.
 * `boolean` if it is unknown if T is an array of one item or not.
 * @internal
 */
export declare type IsArrayOfOne<T extends readonly unknown[]> = T["length"] extends 1 ? true : 1 extends T["length"] ? boolean : false;

/**
 * Editor for schema changes.
 * The only currently supported operation is to replace the stored schema.
 * @internal
 */
export declare interface ISchemaEditor {
    /**
     * Updates the stored schema.
     * @param oldSchema - The schema being overwritten.
     * @param newSchema - The new schema to apply.
     * @internal
     */
    setStoredSchema(oldSchema: TreeStoredSchema, newSchema: TreeStoredSchema): void;
}

/**
 * Checks the type of a `ContextuallyTypedNodeData`.
 * @internal
 */
export declare function isContextuallyTypedNodeDataObject(data: ContextuallyTypedNodeData | undefined): data is ContextuallyTypedNodeDataObject;

/**
 * `true` iff the given type is an acceptable shape for an event
 * @beta
 */
export declare type IsEvent<Event> = Event extends (...args: any[]) => any ? true : false;

/**
 * Collaboratively editable tree distributed data-structure,
 * powered by {@link @fluidframework/shared-object-base#ISharedObject}.
 *
 * See [the README](../../README.md) for details.
 * @internal
 */
export declare interface ISharedTree extends ISharedObject, ITree {
    /**
     * Provides a copy of the current content of the tree.
     * This can be useful for inspecting the tree when no suitable view schema is available.
     * This is only intended for use in testing and exceptional code paths: it is not performant.
     *
     * This does not include everything that is included in a tree summary, since information about how to merge future edits is omitted.
     */
    contentSnapshot(): SharedTreeContentSnapshot;
    /**
     * Like {@link ITree.schematize}, but uses the flex-tree schema system and exposes the tree as a flex-tree.
     * @privateRemarks
     * This has to avoid its name colliding with `schematize`.
     * TODO: Either ITree and ISharedTree should be split into separate objects, the methods should be merged or a better convention for resolving such name conflicts should be selected.
     */
    schematizeInternal<TRoot extends TreeFieldSchema>(config: InitializeAndSchematizeConfiguration<TRoot>): FlexTreeView<TRoot>;
    /**
     * Like {@link ISharedTree.schematizeInternal}, but will never modify the document.
     *
     * @param schema - The view schema to use.
     * @param onSchemaIncompatible - A callback.
     * Invoked when the returned ISharedTreeView becomes invalid to use due to a change to the stored schema which makes it incompatible with the view schema.
     * Called at most once.
     * @returns a view compatible with the provided schema, or undefined if the stored schema is not compatible with the provided view schema.
     * If this becomes invalid to use due to a change in the stored schema, onSchemaIncompatible will be invoked.
     *
     * @privateRemarks
     * TODO:
     * Once views actually have a view schema, onSchemaIncompatible can become an event on the view (which ends its lifetime),
     * instead of a separate callback.
     */
    requireSchema<TRoot extends TreeFieldSchema>(schema: FlexTreeSchema<TRoot>, onSchemaIncompatible: () => void): FlexTreeView<TRoot> | undefined;
}

/**
 * SharedTree editor for transactional tree data and schema changes.
 * @internal
 */
export declare interface ISharedTreeEditor extends IDefaultEditBuilder {
    /**
     * Editor for schema changes.
     * @internal
     */
    schema: ISchemaEditor;
}

/**
 * @internal
 */
export declare function isNeverField(policy: FullSchemaPolicy, originalData: TreeStoredSchema, field: TreeFieldStoredSchema): boolean;

/**
 * An object which allows the registration of listeners so that subscribers can be notified when an event happens.
 *
 * {@link createEmitter} can help implement this interface via delegation.
 * `EventEmitter` can be used as a base class to implement this via extension.
 * @param E - All the events that this emitter supports
 * @example
 * ```ts
 * type MyEventEmitter = IEventEmitter<{
 *   load: (user: string, data: IUserData) => void;
 *   error: (errorCode: number) => void;
 * }>
 * ```
 * @beta
 */
export declare interface ISubscribable<E extends Events<E>> {
    /**
     * Register an event listener.
     * @param eventName - the name of the event
     * @param listener - the handler to run when the event is fired by the emitter
     * @returns a function which will deregister the listener when run. This function has undefined behavior
     * if called more than once.
     */
    on<K extends keyof Events<E>>(eventName: K, listener: E[K]): () => void;
}

/**
 * Used to insert iterable content into a {@link (TreeArrayNode:interface)}.
 * Use {@link (TreeArrayNode:variable).inline} to create an instance of this type.
 * @beta
 */
export declare class IterableTreeListContent<T> implements Iterable<T> {
    private readonly content;
    private constructor();
    /**
     * Package internal construction API.
     * Use {@link (TreeArrayNode:variable).inline} to create an instance of this type instead.
     */
    static [create]<T>(content: Iterable<T>): IterableTreeListContent<T>;
    /**
     * Iterates over content for nodes to insert.
     */
    [Symbol.iterator](): Iterator<T>;
}

/**
 * A collection of functions for managing transactions.
 * Transactions allow edits to be batched into atomic units.
 * Edits made during a transaction will update the local state of the tree immediately, but will be squashed into a single edit when the transaction is committed.
 * If the transaction is aborted, the local state will be reset to what it was before the transaction began.
 * Transactions may nest, meaning that a transaction may be started while a transaction is already ongoing.
 *
 * To avoid updating observers of the view state with intermediate results during a transaction,
 * use {@link ITreeCheckout#fork} and {@link ISharedTreeFork#merge}.
 * @internal
 */
export declare interface ITransaction {
    /**
     * Start a new transaction.
     * If a transaction is already in progress when this new transaction starts, then this transaction will be "nested" inside of it,
     * i.e. the outer transaction will still be in progress after this new transaction is committed or aborted.
     *
     * @remarks - Asynchronous transactions are not supported on the root checkout,
     * since it is always kept up-to-date with the latest remote edits and the results of this rebasing (which might invalidate
     * the transaction) is not visible to the application author.
     * Instead,
     *
     * 1. fork the root checkout
     * 2. run the transaction on the fork
     * 3. merge the fork back into the root checkout
     *
     * @privateRemarks - There is currently no enforcement that asynchronous transactions don't happen on the root checkout.
     * AB#6488 tracks adding some enforcement to make it more clear to application authors that this is not supported.
     */
    start(): void;
    /**
     * Close this transaction by squashing its edits and committing them as a single edit.
     * If this is the root checkout and there are no ongoing transactions remaining, the squashed edit will be submitted to Fluid.
     */
    commit(): TransactionResult.Commit;
    /**
     * Close this transaction and revert the state of the tree to what it was before this transaction began.
     */
    abort(): TransactionResult.Abort;
    /**
     * True if there is at least one transaction currently in progress on this view, otherwise false.
     */
    inProgress(): boolean;
}

/**
 * Channel for a Fluid Tree DDS.
 * @remarks
 * Allows storing and collaboratively editing schema-aware hierarchial data.
 * @beta
 */
export declare interface ITree extends IChannel {
    /**
     * Returns a tree that is known to be compatible with the provided schema. The returned tree exposes an API that is schema-aware.
     *
     * @remarks
     * If the tree is uninitialized (has no schema and no content), the tree is initialized with the `initialTree` and
     * view `schema` in the provided `config`.
     *
     * The tree (now known to have been initialized) has its stored schema checked against the provided view schema.
     *
     * If the schemas are compatible (including updating the stored schema if permitted via `allowedSchemaModifications`),
     * a {@link TreeView} is returned, with a schema-aware API based on the provided view schema.
     *
     * If the schemas are not compatible, an exception is thrown.
     *
     * @privateRemarks
     * TODO: make the mismatch case recoverable.
     * - Provide a way to make a generic view schema for any document.
     * - Produce/throw the error in a document-recoverable way (ex: specific exception type or return value).
     * TODO: Document and handle what happens when the stored schema changes after schematize has returned. Is some invalidation contract needed? How does editable tree behave?
     * TODO: Clarify lifetimes. Ensure calling schematize multiple times does not leak.
     * TODO: Support adapters for handling out-of-schema data.
     *
     * Doing initialization here, regardless of `AllowedUpdateType`, allows a small API that is hard to use incorrectly.
     * Other approaches tend to have easy-to-make mistakes.
     * For example, having a separate initialization function means apps can forget to call it, making an app that can only open existing documents,
     * or call it unconditionally leaving an app that can only create new documents.
     * It also would require the schema to be passed into separate places and could cause issues if they didn't match.
     * Since the initialization function couldn't return a typed tree, the type checking wouldn't help catch that.
     * Also, if an app manages to create a document, but the initialization fails to get persisted, an app that only calls the initialization function
     * on the create code-path (for example how a schematized factory might do it),
     * would leave the document in an unusable state which could not be repaired when it is reopened (by the same or other clients).
     * Additionally, once out of schema content adapters are properly supported (with lazy document updates),
     * this initialization could become just another out of schema content adapter: at that point it clearly belongs here in schematize.
     */
    schematize<TRoot extends ImplicitFieldSchema>(config: TreeConfiguration<TRoot>): TreeView<TreeFieldFromImplicitField<TRoot>>;
}

/**
 * Provides a means for interacting with a SharedTree.
 * This includes reading data from the tree and running transactions to mutate the tree.
 * @remarks This interface should not have any implementations other than those provided by the SharedTree package libraries.
 * @privateRemarks
 * API for interacting with a {@link SharedTreeBranch}.
 * Implementations of this interface must implement the {@link branchKey} property.
 * @internal
 */
export declare interface ITreeCheckout extends AnchorLocator {
    /**
     * Read and Write access for schema stored in the document.
     *
     * These APIs are temporary and will be replaced with different abstractions (View Schema based) in a different place later.
     *
     * TODO:
     * Editing of this should be moved into transactions with the rest of tree editing to they can be intermixed.
     * This will be done after the relations between views, branches and Indexes are figured out.
     *
     * TODO:
     * Public APIs for dealing with schema should be in terms of View Schema, and schema update policies.
     * The actual stored schema should be hidden (or ar least not be the most prominent way to interact with schema).
     *
     * TODO:
     * Something should ensure the document contents are always in schema.
     */
    readonly storedSchema: TreeStoredSchemaSubscription;
    /**
     * Current contents.
     * Updated by edits (local and remote).
     * Use `editor` to create a local edit.
     */
    readonly forest: IForestSubscription;
    /**
     * Used to edit the state of the tree. Edits will be immediately applied locally to the tree.
     * If there is no transaction currently ongoing, then the edits will be submitted to Fluid immediately as well.
     */
    readonly editor: ISharedTreeEditor;
    /**
     * A collection of functions for managing transactions.
     */
    readonly transaction: ITransaction;
    /**
     * Spawn a new view which is based off of the current state of this view.
     * Any mutations of the new view will not apply to this view until the new view is merged back into this view via `merge()`.
     */
    fork(): ITreeCheckoutFork;
    /**
     * Apply all the new changes on the given view to this view.
     * @param view - a view which was created by a call to `fork()`.
     * It is automatically disposed after the merge completes.
     * @remarks All ongoing transactions (if any) in `view` will be committed before the merge.
     */
    merge(view: ITreeCheckoutFork): void;
    /**
     * Apply all the new changes on the given view to this view.
     * @param view - a view which was created by a call to `fork()`.
     * @param disposeView - whether or not to dispose `view` after the merge completes.
     * @remarks All ongoing transactions (if any) in `view` will be committed before the merge.
     */
    merge(view: ITreeCheckoutFork, disposeView: boolean): void;
    /**
     * Rebase the given view onto this view.
     * @param view - a view which was created by a call to `fork()`. It is modified by this operation.
     */
    rebase(view: ITreeCheckoutFork): void;
    /**
     * Replaces all schema with the provided schema.
     * Can over-write preexisting schema, and removes unmentioned schema.
     */
    updateSchema(newSchema: TreeStoredSchema): void;
    /**
     * Events about this view.
     */
    readonly events: ISubscribable<CheckoutEvents>;
    /**
     * Events about the root of the tree in this view.
     */
    readonly rootEvents: ISubscribable<AnchorSetRootEvents>;
    /**
     * Returns a JsonableTree for each tree that was removed from (and not restored to) the document.
     * This list is guaranteed to contain all nodes that are recoverable through undo/redo on this checkout.
     * The list may also contain additional nodes.
     *
     * This is only intended for use in testing and exceptional code paths: it is not performant.
     */
    getRemovedRoots(): [string | number | undefined, number, JsonableTree][];
}

/**
 * Branch (like in a version control system) of SharedTree.
 *
 * {@link ITreeCheckout} that has forked off of the main trunk/branch.
 * @internal
 */
export declare interface ITreeCheckoutFork extends ITreeCheckout {
    /**
     * Rebase the changes that have been applied to this view over all the new changes in the given view.
     * @param view - Either the root view or a view that was created by a call to `fork()`. It is not modified by this operation.
     */
    rebaseOnto(view: ITreeCheckout): void;
}

/**
 * A stateful low-level interface for reading tree data.
 * @internal
 *
 * @remarks Cursor exists so that specialized data formats can be viewed through
 * a common abstraction. This allows performance optimizations to be done based
 * on data.
 *
 * A tree cursor is similar to a database cursor in that it allows for the efficient
 * traversal over the contents of a tree. Note that unlike a database cursor,
 * tree cursors may be invalidated after any edit to the tree. For a cursor-like
 * structure that also remains valid across edits, see {@link AnchorNode}.
 */
export declare interface ITreeCursor {
    /**
     * Marks this object as a cursor.
     */
    readonly [CursorMarker]: true;
    /**
     * What kind of place the cursor is at.
     * Determines which operations are allowed.
     *
     * @remarks
     * Users of cursors frequently need to refer to places in trees, both fields and nodes.
     * Approaches other than having the cursor have separate modes for these
     * cases had issues even worse than having the two modes.
     *
     * For example, modeling fields as parent + key has issues when there is no
     * parent, and doesn't provide a great way to do iteration over fields while
     * also having a nice API and making it easy for the implementation to track
     * state (like its current location inside a sequence tree of fields) while
     * traversing without having to allocate some state management for that.
     *
     * Another approach, of using arrays of cursors for fields (like we currently
     * do for inserting content) is very inefficient and better addressed by a
     * dual mode cursor.
     *
     * Another approach, of using the first node in a field when referring to
     * the field gets confusing since it's unclear if a given cursor means that
     * node, or that node, and the ones after it, and in the second case, it's
     * hard to restore the cursor back to the right state when returning. It also
     * doesn't work for empty fields. Overall there just didn't seem to be a way
     * that sucked less than the dual mode API.
     */
    readonly mode: CursorLocationType;
    readonly pending: boolean;
    /**
     * Moves the "current field" forward one in an arbitrary field traversal order,
     * skipping any empty fields.
     *
     * If there is no remaining field to iterate to,
     * returns false and navigates up to the parent setting the mode to `Nodes`.
     *
     * Order of fields is only guaranteed to be consistent through a single iteration.
     *
     * If skipPending, skip past fields which are currently pending.
     * This can be used to skip to the end of a large number of consecutive pending fields.
     *
     * Allowed when `mode` is `Fields`.
     */
    nextField(): boolean;
    /**
     * Navigate up to parent node.
     * Sets mode to `Nodes`
     *
     * Only valid when `mode` is `Fields`.
     *
     * TODO: what to do if at root?
     */
    exitField(): void;
    /**
     * Moves the "current field" forward until `pending` is `false`.
     *
     * If there are no remaining field to iterate to,
     * returns false and navigates up to the parent setting the mode to `Nodes`.
     *
     * Order of fields is only guaranteed to be consistent through a single iteration.
     *
     * Allowed when `mode` is `Fields`.
     */
    skipPendingFields(): boolean;
    /**
     * Returns the FieldKey for the current field.
     *
     * Allowed when `mode` is `Fields`, and not `pending`.
     */
    getFieldKey(): FieldKey;
    /**
     * @returns the number of immediate children in the current field.
     *
     * Allowed when `mode` is `Fields`, and not `pending`.
     */
    getFieldLength(): number;
    /**
     * Moves to the first node of the selected field, setting mode to `Nodes`.
     *
     * If field is empty, returns false instead.
     *
     * Allowed when `mode` is `Fields`, and not `pending`.
     */
    firstNode(): boolean;
    /**
     * Sets current node to the node at the provided `index` of the current field.
     *
     * Allowed when `mode` is `Fields`, and not `pending`.
     * Sets mode to `Nodes`.
     */
    enterNode(childIndex: number): void;
    /**
     * Returns a path to the current field. See {@link FieldUpPath}.
     *
     * Only valid when `mode` is `Fields`.
     *
     * If no prefix is provided, assumes this cursor is treated as if it has a root node where its field keys are actually detached sequences.
     * If the cursor is not rooted at such a node, the `prefix` should be used to ensure the path has the correct root.
     * This requirement exists because {@link FieldUpPath}s are absolute paths
     * and thus must be rooted in a detached sequence.
     *
     * @param prefix - optional overrides to apply to the root of the returned path.
     * See {@link PathRootPrefix}.
     * This adjusts the path as if the tree data accessible to this cursor is part of a larger tree.
     *
     * @returns a path to the current field.
     */
    getFieldPath(prefix?: PathRootPrefix): FieldUpPath;
    /**
     * Returns a path to the current node. See {@link UpPath}.
     *
     * Only valid when `mode` is `Nodes`.
     *
     * If no prefix is provided, assumes this cursor is treated as if it has a root node where its field keys are actually detached sequences.
     * If the cursor is not rooted at such a node, the `prefix` should be used to ensure the path has the correct root.
     * This requirement exists because {@link UpPath}s are absolute paths
     * and thus must be rooted in a detached sequence.
     *
     * @param prefix - optional overrides to apply to the root of the returned path.
     * See {@link PathRootPrefix}.
     * This adjusts the path as if the tree data accessible to this cursor is part of a larger tree.
     *
     * @returns a path to the current node.
     */
    getPath(prefix?: PathRootPrefix): UpPath | undefined;
    /**
     * Index (within its parent field) of the current node.
     *
     * Only valid when `mode` is `Nodes`.
     */
    readonly fieldIndex: number;
    /**
     * Index (within its parent field) of the first node in the current chunk.
     * Always less than or equal to `currentIndexInField`.
     *
     * Only valid when `mode` is `Nodes`.
     */
    readonly chunkStart: number;
    /**
     * Length of current chunk.
     * Since an entire chunk always has the same `pending` value,
     * can be used to help skip over all of a pending chunk at once.
     *
     * TODO:
     * Add optional APIs to access underlying chunks so readers can
     * accelerate processing of chunk formats they understand.
     *
     * Only valid when `mode` is `Nodes`.
     */
    readonly chunkLength: number;
    /**
     * Moves `offset` nodes in the field.
     * If seeking to exactly past either end,
     * returns false and navigates up to the parent field (setting mode to `Fields`).
     *
     * Allowed if mode is `Nodes`.
     */
    seekNodes(offset: number): boolean;
    /**
     * The same as `seekNodes(1)`, but might be faster.
     */
    nextNode(): boolean;
    /**
     * Navigate up to parent field.
     * Sets mode to `Fields`
     *
     * Same as seek Number.POSITIVE_INFINITY, but only valid when `mode` is `Nodes`.
     *
     * TODO: what to do if at root?
     * TODO: Maybe merge with upToNode to make a single "Up"?
     */
    exitNode(): void;
    /**
     * Enters the first non-empty field (setting mode to `Fields`)
     * so fields can be iterated with `nextField` and `skipPendingFields`.
     *
     * If there are no fields, mode is returned to `Nodes` and false is returned.
     *
     * Allowed when `mode` is `Nodes` and not `pending`.
     */
    firstField(): boolean;
    /**
     * Navigate to the field with the specified `key` and set the mode to `Fields`.
     *
     * Only valid when `mode` is `Nodes`, and not `pending`.
     */
    enterField(key: FieldKey): void;
    /**
     * The type of the currently selected node.
     *
     * Only valid when `mode` is `Nodes`, and not `pending`.
     */
    readonly type: TreeType;
    /**
     * The value associated with the currently selected node.
     *
     * Only valid when `mode` is `Nodes`, and not `pending`.
     */
    readonly value: Value;
}

/**
 * {@link ITreeCursor} that is never pending.
 * @internal
 */
export declare interface ITreeCursorSynchronous extends ITreeCursor {
    readonly pending: false;
}

/**
 * ITreeCursor supporting IForestSubscription and its changes over time.
 * @internal
 */
export declare interface ITreeSubscriptionCursor extends ITreeCursor {
    /**
     * @returns an independent copy of this cursor at the same location in the tree.
     */
    fork(): ITreeSubscriptionCursor;
    /**
     * Release any resources this cursor is holding onto.
     * After doing this, further use of this object other than reading `state` is forbidden (undefined behavior).
     */
    free(): void;
    /**
     * Release any resources this cursor is holding onto.
     * After doing this, further use of this object other than reading `state` or passing to `tryGet`
     * or calling `free` is forbidden (undefined behavior).
     */
    clear(): void;
    /**
     * Construct an `Anchor` which the IForestSubscription will keep rebased to `current`.
     * Note that maintaining an Anchor has cost: free them to stop incurring that cost.
     *
     * Only valid when `mode` is `Nodes`.
     */
    buildAnchor(): Anchor;
    /**
     * Construct a `FieldAnchor` which the IForestSubscription will keep rebased to `current`.
     * Note that maintaining an Anchor has cost: free them to stop incurring that cost.
     *
     * Only valid when `mode` is `Fields`.
     */
    buildFieldAnchor(): FieldAnchor;
    /**
     * Current state.
     */
    readonly state: ITreeSubscriptionCursorState;
}

/**
 * @internal
 */
export declare enum ITreeSubscriptionCursorState {
    /**
     * On the current revision of the forest.
     */
    Current = 0,
    /**
     * Empty, but can be reused.
     */
    Cleared = 1,
    /**
     * Freed and must not be used.
     */
    Freed = 2
}

/**
 * Branch (like in a version control system) of SharedTree.
 *
 * {@link FlexTreeView} that has forked off of the main trunk/branch.
 * @internal
 */
export declare interface ITreeViewFork<in out TRoot extends TreeFieldSchema> extends FlexTreeView<TRoot> {
    readonly checkout: ITreeCheckoutFork;
}

/**
 * A tree represented using plain JavaScript objects.
 * Can be passed to `JSON.stringify()` to produce a human-readable/editable JSON tree.
 *
 * JsonableTrees should not store empty fields.
 * @internal
 */
export declare interface JsonableTree extends GenericTreeNode<JsonableTree> {
}

/**
 * Extract a JsonableTree from the contents of the given ITreeCursor's current node.
 * @internal
 */
export declare function jsonableTreeFromCursor(cursor: ITreeCursor): JsonableTree;

/**
 * Use for Json compatible data.
 *
 * Note that this does not robustly forbid non json comparable data via type checking,
 * but instead mostly restricts access to it.
 * @internal
 */
export declare type JsonCompatible = string | number | boolean | null | JsonCompatible[] | JsonCompatibleObject;

/**
 * Use for Json object compatible data.
 *
 * Note that this does not robustly forbid non json comparable data via type checking,
 * but instead mostly restricts access to it.
 * @internal
 */
export declare type JsonCompatibleObject = {
    [P in string]?: JsonCompatible;
};

/**
 * Use for readonly view of Json compatible data.
 *
 * Note that this does not robustly forbid non json comparable data via type checking,
 * but instead mostly restricts access to it.
 * @internal
 */
export declare type JsonCompatibleReadOnly = string | number | boolean | null | readonly JsonCompatibleReadOnly[] | {
    readonly [P in string]?: JsonCompatibleReadOnly;
};

/**
 * JSON schema validator compliant with draft 6 schema. See https://json-schema.org.
 * @internal
 */
export declare interface JsonValidator {
    /**
     * Compiles the provided JSON schema into a validator for that schema.
     * @param schema - A valid draft 6 JSON schema
     * @remarks IFluidHandles--which have circular property references--are used in various places in SharedTree's persisted
     * format. Handles should only be contained in sections of data which are validated against the empty schema `{}`
     * (see https://datatracker.ietf.org/doc/html/draft-wright-json-schema-01#section-4.4).
     *
     * Implementations of `JsonValidator` must therefore tolerate these values, despite the input not being valid JSON.
     */
    compile<Schema extends TSchema>(schema: Schema): SchemaValidationFunction<Schema>;
}

/**
 * An "eager" or "lazy" Item in a `FlexList`.
 * Lazy items are wrapped in a function to allow referring to themselves before they are declared.
 * This makes recursive and co-recursive items possible.
 * @beta
 */
declare type LazyItem<Item = unknown> = Item | (() => Item);

/**
 * Tree type, but can be wrapped in a function to allow referring to types before they are declared.
 * This makes recursive and co-recursive types possible.
 * @internal
 */
export declare type LazyTreeNodeSchema = FlexTreeNodeSchema | (() => FlexTreeNodeSchema);

/**
 * Built-in {@link LeafNodeSchema}.
 * @internal
 */
export declare const leaf: {
    /**
     * {@link LeafNodeSchema} for holding a JavaScript `number`.
     *
     * @remarks
     * The number is a [double-precision 64-bit binary format IEEE 754](https://en.wikipedia.org/wiki/Double-precision_floating-point_format) value, however there are some exceptions:
     * - `NaN`, and the infinities should not be used.
     * - `-0` may be converted to `0` in some cases.
     *
     * These limitations match the limitations of JSON.
     * @privateRemarks
     * TODO:
     * We should be much more clear about what happens if you use problematic values.
     * We should validate and/or normalize them when inserting content.
     */
    number: LeafNodeSchema<"com.fluidframework.leaf.number", ValueSchema.Number>;
    /**
     * {@link LeafNodeSchema} for holding a boolean.
     */
    boolean: LeafNodeSchema<"com.fluidframework.leaf.boolean", ValueSchema.Boolean>;
    /**
     * {@link LeafNodeSchema} for holding a JavaScript `string`.
     *
     * @remarks
     * Strings containing unpaired UTF-16 surrogate pair code units may not be handled correctly.
     *
     * These limitations come from the use of UTF-8 encoding of the strings, which requires them to be valid unicode.
     * JavaScript does not make this requirement for its strings so not all possible JavaScript strings are supported.
     * @privateRemarks
     * TODO:
     * We should be much more clear about what happens if you use problematic values.
     * We should validate and/or normalize them when inserting content.
     */
    string: LeafNodeSchema<"com.fluidframework.leaf.string", ValueSchema.String>;
    /**
     * {@link LeafNodeSchema} for holding an {@link @fluidframework/core-interfaces#IFluidHandle}.
     */
    handle: LeafNodeSchema<"com.fluidframework.leaf.handle", ValueSchema.FluidHandle>;
    /**
     * {@link LeafNodeSchema} for JavaScript `null`.
     *
     * @remarks
     * There are good [reasons to avoid using null](https://www.npmjs.com/package/%40rushstack/eslint-plugin#rushstackno-new-null) in JavaScript, however sometimes it is desired.
     * This {@link LeafNodeSchema} node provides the option to include nulls in trees when desired.
     * Unless directly inter-operating with existing data using null, consider other approaches, like wrapping the value in an optional field, or using a more specifically named empty object node.
     */
    null: LeafNodeSchema<"com.fluidframework.leaf.null", ValueSchema.Null>;
    /**
     * The set of {@link LeafNodeSchema} which correspond to JavaScript primitive (non-object) types.
     */
    primitives: readonly [LeafNodeSchema<"com.fluidframework.leaf.number", ValueSchema.Number>, LeafNodeSchema<"com.fluidframework.leaf.boolean", ValueSchema.Boolean>, LeafNodeSchema<"com.fluidframework.leaf.string", ValueSchema.String>];
    /**
     * All {@link LeafNodeSchema} defined in this library..
     */
    all: readonly [LeafNodeSchema<"com.fluidframework.leaf.handle", ValueSchema.FluidHandle>, LeafNodeSchema<"com.fluidframework.leaf.null", ValueSchema.Null>, LeafNodeSchema<"com.fluidframework.leaf.number", ValueSchema.Number>, LeafNodeSchema<"com.fluidframework.leaf.boolean", ValueSchema.Boolean>, LeafNodeSchema<"com.fluidframework.leaf.string", ValueSchema.String>];
    /**
     * {@link SchemaLibrary} of the {@link LeafNodeSchema}.
     *
     * @privateRemarks
     * This is included by default in schema produced with {@link SchemaBuilder}.
     */
    library: SchemaLibrary;
};

/**
 * @internal
 */
export declare class LeafNodeSchema<const out Name extends string = string, const out Specification extends Unenforced<ValueSchema> = ValueSchema> extends TreeNodeSchemaBase<Name, Specification> {
    get leafValue(): ValueSchema;
    protected _typeCheck2?: MakeNominal;
    static create<const Name extends string, const Specification extends ValueSchema>(builder: Named<string>, name: TreeNodeSchemaIdentifier<Name>, specification: Specification): LeafNodeSchema<Name, Specification>;
    getFieldSchema(field: FieldKey): TreeFieldSchema;
}

/**
 * A key which uniquely identifies a node in the tree within this session.
 * @remarks {@link LocalNodeKey}s must not be serialized and stored as data without first being converted into a {@link StableNodeKey}.
 * They are local to the current session and equivalent nodes in another session will not necessarily have the same {@link LocalNodeKey}.
 * However, they are otherwise preferential to use over {@link StableNodeKey}s as they are much smaller.
 * For example, they are more efficient than {@link StableNodeKey}s when used as keys in a map.
 * {@link LocalNodeKey}s may be compared or equated via {@link compareLocalNodeKeys}.
 * @internal
 */
export declare interface LocalNodeKey extends Opaque<Brand<SessionSpaceCompressedId, "Local Node Key">> {
}

/**
 * Utilities for manipulating the typescript typechecker.
 *
 * @remarks
 * While it appears the the variance parts of this library are made obsolete by TypeScript 4.7's explicit variance annotations,
 * many cases still type check with incorrect variance even when using the explicit annotations,
 * and are fixed by using the patterns in this library.
 *
 * TypeScript uses structural typing if there are no private or protected members,
 * and variance of generic type parameters depends on their usages.
 * Thus when trying to constrain code by adding extra type information,
 * it often fails to actually constrain as desired, and these utilities can help with those cases.
 *
 * This library is designed so that the desired variance can be documented in a way that is easy to read, concise,
 * and allows easy navigation to documentation explaining what is being done
 * for readers who are not familiar with this library.
 * Additionally it constrains the types so the undesired usage patterns will not compile,
 * and will give somewhat intelligible errors.
 *
 * Additionally this library provides the tools needed to test that the type constraints are working as expected,
 * or test any other similar typing constraints in an application.
 *
 * This library assumes you are compiling with --strictFunctionTypes:
 * (Covariance and Contravariance is explained along with how these helpers cause it in typescript at this link)
 * {@link https://www.typescriptlang.org/docs/handbook/release-notes/typescript-2-6.html#strict-function-types}.
 * If compiled with a TypeScript configuration that is not strict enough for these features to work,
 * the test suite should fail to build.
 *
 * Classes in TypeScript by default allow all assignments:
 * its only though adding members that any type constraints actually get applied.
 * This library provides types that can be used on a protected member of a class to add the desired constraints.
 *
 * Typical usages (use one field like this at the top of a class):
 * ```typescript
 * protected _typeCheck?: MakeNominal;
 * protected _typeCheck?: Contravariant<T>;
 * protected _typeCheck?: Covariant<T>;
 * protected _typeCheck?: Invariant<T>;
 * protected _typeCheck?: Contravariant<T> & Invariant<K>;
 * ```
 *
 * See tests for examples.
 *
 * Note that all of these cause nominal typing.
 * If constraints on generic type parameter variance are desired, but nominal typing is not,
 * these types can be used on a public field. This case also works with interfaces.
 *
 * Be aware that other members of your type might apply further constraints
 * (ex: you might try and write a Contravariant<T> class, but it ends up being Invariant<T> due to a field of type T).
 *
 * Be aware of TypeScript Bug:
 * {@link https://github.com/microsoft/TypeScript/issues/36906}.
 * This bug is why the fields here are protected not private.
 * Note that this bug is closed as a duplicate of {@link https://github.com/microsoft/TypeScript/issues/20979}
 * which was closed because fixing it would be too large of a breaking change.
 * Thus we expect this bug to be part of TypeScript for the forseeable future.
 */
/**
 * Use this as the type of a protected field to cause a type to use nominal typing instead of structural.
 *
 * @remarks
 * Using nominal typing in this way prevents assignment of objects which are not instances of this class to values of this class's type.
 * Classes which are used with "instanceof", or are supposed to be instantiated in particular ways (not just made with object literals)
 * can use this to prevent undesired assignments.
 * @example
 * ```typescript
 * protected _typeCheck?: MakeNominal;
 * ```
 * @privateRemarks
 * See: {@link https://dev.azure.com/intentional/intent/_wiki/wikis/NP%20Platform/7146/Nominal-vs-Structural-Types}
 * @beta
 */
declare interface MakeNominal {
}

/**
 * Subset of TreeFieldSchema thats legal in maps.
 * This requires empty to be a valid value for the map.
 * @internal
 */
export declare type MapFieldSchema = TreeFieldSchema<typeof FieldKinds.optional | typeof FieldKinds.sequence>;

/**
 * @internal
 */
export declare class MapNodeSchema<const out Name extends string = string, const out Specification extends Unenforced<MapFieldSchema> = MapFieldSchema> extends TreeNodeSchemaBase<Name, Specification> {
    get mapFields(): MapFieldSchema;
    protected _typeCheck2?: MakeNominal;
    static create<const Name extends string, const Specification extends MapFieldSchema>(builder: Named<string>, name: TreeNodeSchemaIdentifier<Name>, specification: Specification): MapNodeSchema<Name, Specification>;
    getFieldSchema(field: FieldKey): MapFieldSchema;
}

/**
 * This modules provides a simple in memory tree format.
 */
/**
 * Simple in memory tree representation based on Maps.
 * MapTrees should not store empty fields.
 * @internal
 */
export declare interface MapTree extends NodeData {
    fields: Map<FieldKey, MapTree[]>;
}

/**
 * Can be used to mark a type which works like an array, but is not compatible with `Array.isArray`.
 * @internal
 */
export declare interface MarkedArrayLike<TGet, TSet extends TGet = TGet> extends ArrayLikeMut<TGet, TSet> {
    readonly [arrayLikeMarkerSymbol]: true;
    [Symbol.iterator](): IterableIterator<TGet>;
}

/**
 * Describes how a particular field functions.
 *
 * This determine its reading and editing APIs, multiplicity, and what merge resolution policies it will use.
 * @internal
 */
export declare enum Multiplicity {
    /**
     * Exactly one item.
     */
    Single = 0,
    /**
     * 0 or 1 items.
     */
    Optional = 1,
    /**
     * 0 or more items.
     */
    Sequence = 2,
    /**
     * Exactly 0 items.
     *
     * Using Forbidden makes what types are listed for allowed in a field irrelevant
     * since the field will never have values in it.
     *
     * Using Forbidden is equivalent to picking a kind that permits empty (like sequence or optional)
     * and having no allowed types (or only never types).
     * Because of this, its possible to express everything constraint wise without Forbidden,
     * but using Forbidden can be more semantically clear than optional with no allowed types.
     *
     * For view schema, this can be useful if you need to:
     * - run a specific out of schema handler when a field is present,
     * but otherwise are ignoring or tolerating (ex: via extra fields) unmentioned fields.
     * - prevent a specific field from being used as an extra field
     * (perhaps for some past of future compatibility reason)
     * - keep a field in a schema for metadata purposes
     * (ex: for improved error messaging, error handling or documentation)
     * that is not used in this specific version of the schema (ex: to document what it was or will be used for).
     *
     * For stored schema, this can be useful if you need to:
     * - have a field which can have its schema updated to Optional or Sequence of any type.
     * - to exclude a field from extra fields
     * - for the schema system to use as a default for fields which aren't declared
     * (ex: when updating a field that did not exist into one that does)
     *
     * @privateRemarks
     * See storedEmptyFieldSchema for a constant, reusable field using Forbidden.
     */
    Forbidden = 3
}

/**
 * Type with a name describing what it is.
 * Typically used with values (like schema) that can be stored in a map, but in some representations have their name/key as a field.
 * @internal
 */
export declare interface Named<TName> {
    readonly name: TName;
}

/**
 * Implementation detail of type branding. Should not be used directly outside this file,
 * but shows up as part of branded types so API-Extractor requires it to be exported.
 * @internal
 */
export declare type NameFromBranded<T extends BrandedType<any, string>> = T extends BrandedType<any, infer Name> ? Name : never;

/**
 * A dictionary whose values are keyed off of two objects (key1, key2).
 * As it is a nested map, size() will return the number of distinct key1s.
 * If you need constant-time access to the number of values, use SizedNestedMap instead.
 *
 * This code assumes values will not be undefined (keys can be undefined).
 *
 * @internal
 */
export declare type NestedMap<Key1, Key2, Value> = Map<Key1, Map<Key2, Value>>;

/**
 * Content to use for a field.
 *
 * When used, this content will be deeply copied into the tree, and must comply with the schema.
 *
 * The content must follow the {@link Multiplicity} of the {@link FieldKind}:
 * - use a single cursor for an `optional` or `value` field;
 * - use array of cursors for a `sequence` field;
 *
 * TODO: this should allow a field cursor instead of an array of cursors.
 * TODO: Make this generic so a variant of this type that allows placeholders for detached sequences to consume.
 * @internal
 */
export declare type NewFieldContent = ITreeCursorSynchronous | readonly ITreeCursorSynchronous[] | ContextuallyTypedFieldData;

/**
 * Given a node's schema, return the corresponding object from which the node could be built.
 * @privateRemarks
 * Currently this assumes factory functions take exactly one argument.
 * This could be changed if needed.
 *
 * These factory functions can also take a FlexTreeNode, but this is not exposed in the public facing types.
 * @beta
 */
declare type NodeBuilderData<T extends TreeNodeSchema> = T extends TreeNodeSchema<string, NodeKind, unknown, infer TBuild> ? TBuild : never;

/**
 * The fields required by a node in a tree.
 *
 * @privateRemarks A forked version of this type is used in `persistedTreeTextFormat.ts`.
 * Changes to this type might necessitate changes to `EncodedNodeData` or codecs.
 * See persistedTreeTextFormat's module documentation for more details.
 *
 * @internal
 */
export declare interface NodeData {
    /**
     * A payload of arbitrary serializable data.
     *
     * TODO: clarify rules for mutating this value.
     * For now, avoid mutating the TreeValue itself.
     * For example, if its an object, make a modified copy of the object instead of mutating it.
     */
    value?: TreeValue;
    /**
     * The meaning of this node.
     * Provides contexts/semantics for this node and its content.
     * Typically used to associate a node with metadata (including a schema) and source code (types, behaviors, etc).
     */
    readonly type: TreeNodeSchemaIdentifier;
}

/**
 * @internal
 */
export declare interface NodeExistsConstraint {
    violated: boolean;
}

/**
 * Takes in `TreeNodeSchema[]` and returns a TypedNode union.
 * @beta
 */
export declare type NodeFromSchema<T extends TreeNodeSchema> = T extends TreeNodeSchema<string, NodeKind, infer TNode> ? TNode : never;

/**
 * Indicates the index of a node in a field.
 * Zero indicates the first node in a field.
 * @internal
 */
export declare type NodeIndex = number;

/**
 * The key for the special field for {@link LocalNodeKey}s,
 * which allows nodes to be given keys that can be used to find the nodes via the node key index.
 * @internal
 * @privateRemarks TODO: Come up with a unified and collision-resistant naming schema for fields defined by the system.
 * For now, we'll use `__` to reduce the change of collision, since this is what other internal properties use in Fluid.
 */
export declare const nodeKeyFieldKey = "__n_id__";

/**
 * @internal
 */
export declare interface NodeKeyFieldKind extends FieldKind<"NodeKey", Multiplicity.Single> {
}

/**
 * A collection of utilities for managing {@link StableNodeKey}s.
 * A node key can be assigned to a node and allows that node to be easily retrieved from the tree at a later time. (see `nodeKey.map`).
 * @remarks {@link LocalNodeKey}s are put on {@link FlexTreeObjectNode}s via a special field.
 * A node with a node key in its schema must always have a node key.
 * @internal
 */
export declare interface NodeKeys {
    /**
     * Create a new {@link LocalNodeKey} which can be used as the key for a node in the tree.
     */
    generate(): LocalNodeKey;
    /**
     * Convert the given {@link LocalNodeKey} into a UUID that can be serialized.
     * @param key - the key to convert
     */
    stabilize(key: LocalNodeKey): StableNodeKey;
    /**
     * Convert a {@link StableNodeKey} back into its {@link LocalNodeKey} form.
     * @param key - the key to convert
     */
    localize(key: StableNodeKey): LocalNodeKey;
    /**
     * A map of all {@link LocalNodeKey}s in the document to their corresponding nodes.
     */
    readonly map: ReadonlyMap<LocalNodeKey, FlexTreeObjectNode>;
}

/**
 * Kind of tree node.
 * @beta
 */
declare enum NodeKind {
    /**
     * A node which serves as a map, storing children under string keys.
     */
    Map = 0,
    /**
     * A node which serves as an array, storing children in an ordered sequence.
     */
    Array = 1,
    /**
     * A node which stores a heterogenous collection of children in named fields.
     * @remarks
     * Each field gets its own schema.
     */
    Object = 2,
    /**
     * A node which stores a single leaf value.
     */
    Leaf = 3
}

/**
 * Called when the last listener for `eventName` is removed.
 * Useful for determining when to clean up resources related to detecting when the event might occurs.
 * @internal
 */
export declare type NoListenersCallback<E extends Events<E>> = (eventName: keyof Events<E>) => void;

/**
 * A {@link JsonValidator} implementation which performs no validation and accepts all data as valid.
 * @privateRemarks Having this as an option unifies opting out of validation with selection of
 * validators, simplifying code performing validation.
 * @internal
 */
export declare const noopValidator: JsonValidator;

/**
 * Normalizes an {@link ImplicitAllowedTypes} into  {@link AllowedTypes}.
 * @internal
 */
export declare type NormalizeAllowedTypes<TSchema extends ImplicitAllowedTypes> = TSchema extends FlexTreeNodeSchema ? readonly [TSchema] : TSchema extends Any ? readonly [Any] : TSchema;

/**
 * Normalizes an {@link ImplicitFieldSchema} into a {@link TreeFieldSchema}.
 * @internal
 */
export declare type NormalizeField<TSchema extends FlexImplicitFieldSchema, TDefault extends FieldKind> = TSchema extends TreeFieldSchema ? TSchema : TreeFieldSchema<TDefault, NormalizeAllowedTypes<Assume<TSchema, ImplicitAllowedTypes>>>;

/**
 * Convert FieldSchemaSpecification | undefined into TreeFieldSchema.
 * @internal
 */
export declare type NormalizeFieldSchema<T extends TreeFieldSchema | undefined> = T extends TreeFieldSchema ? T : TreeFieldSchema<typeof FieldKinds.forbidden, []>;

/**
 * @internal
 */
export declare type NormalizeObjectNodeFields<T extends Fields> = {
    readonly [Property in keyof T]: NormalizeFieldSchema<T[Property]>;
};

/**
 * Helper used to produce types for object nodes.
 * @beta
 */
declare type ObjectFromSchemaRecord<T extends RestrictiveReadonlyRecord<string, ImplicitFieldSchema>> = {
    -readonly [Property in keyof T]: TreeFieldFromImplicitField<T[Property]>;
};

/**
 * @internal
 */
export declare class ObjectNodeSchema<const out Name extends string = string, const out Specification extends Unenforced<Fields> = Fields> extends TreeNodeSchemaBase<Name, Specification> {
    readonly objectNodeFieldsObject: NormalizeObjectNodeFields<Assume<Specification, Fields>>;
    readonly objectNodeFields: ReadonlyMap<FieldKey, TreeFieldSchema>;
    protected _typeCheck2?: MakeNominal;
    static create<const Name extends string, const Specification extends Fields>(builder: Named<string>, name: TreeNodeSchemaIdentifier<Name>, specification: Specification): ObjectNodeSchema<Name, Specification>;
    private constructor();
    getFieldSchema(field: FieldKey): TreeFieldSchema;
}

/**
 * Returns the value from `set` if it contains exactly one item, otherwise `undefined`.
 * @internal
 */
export declare function oneFromSet<T>(set: ReadonlySet<T> | undefined): T | undefined;

/**
 * {@inheritdoc TreeNode.[onNextChange]}
 */
declare const onNextChange: unique symbol;

/**
 * Converts a Branded type into an "opaque" handle.
 * This prevents the value from being used directly, but does not fully type erase it
 * (and this its not really fully opaque):
 * The type can be recovered using {@link extractFromOpaque},
 * however if we assume only code that produces these "opaque" handles does that conversion,
 * they can function like opaque handles.
 *
 * Recommended usage is to use `interface` instead of `type` so tooling (such as tsc and refactoring tools)
 * uses the type name instead of expanding it:
 * ```typescript
 * export interface MyType extends Opaque<Brand<string, "myPackage.MyType">>{}
 * ```
 * @internal
 */
export declare type Opaque<T extends Brand<any, string>> = T extends BrandedType<infer ValueType, infer Name> ? BrandedType<ValueType, Name> : never;

/**
 * @internal
 */
export declare interface Optional extends FieldKind<"Optional", Multiplicity.Optional> {
}

/**
 * @internal
 */
export declare interface OptionalFieldEditBuilder {
    /**
     * Issues a change which replaces the current newContent of the field with `newContent`
     * @param newContent - the new content for the field. Must be in Nodes mode.
     * @param wasEmpty - whether the field is empty when creating this change
     */
    set(newContent: ITreeCursor | undefined, wasEmpty: boolean): void;
}

/**
 * Extract fields which permit undefined but can also hold other types.
 * @internal
 */
export declare type OptionalFields<T> = [
    {
    [P in keyof T as undefined extends T[P] ? T[P] extends undefined ? never : P : never]?: T[P];
}
][_InlineTrick];

/**
 * Prefix to apply as the root of a {@link UpPath} or {@link FieldUpPath}.
 *
 * @remarks This can be used to take a path relative to a subtree, and make it relative to a larger containing tree.
 * For example, if a node is being inserted in the 5th position in a field "Foo", you can update a path in that node's subtree to its new path by prefixing it with
 * `{ parent: theNodeAboveTheMovedNode, rootFieldOverride: Foo, indexOffset: 5 }`.
 * See {@link prefixPath} and {@link prefixFieldPath} for how to apply the prefix to the paths.
 * @internal
 */
export declare interface PathRootPrefix {
    /**
     * The new parent to place above root of the path which is being prefixed.
     * This replaces the `undefined` at the root of the path.
     *
     * @remarks specifying `undefined` here results in no change to the path.
     */
    parent?: UpPath | undefined;
    /**
     * The field of `parent` that the original path will be included under.
     *
     * If `undefined` the root field key from the original path will be used.
     */
    rootFieldOverride?: FieldKey;
    /**
     * Offset to add to the uppermost `parentIndex` in the original path.
     */
    indexOffset?: number;
}

/**
 * Delta visitor for the path tree.
 *
 * For any of these events, the paths are guaranteed to be valid at the time of the event,
 * and it is valid to read from the Forest at that path.
 *
 * @internal
 */
export declare interface PathVisitor {
    /**
     * Invoked after the creation of a range of nodes.
     * @param content - The content that was created.
     */
    afterCreate(content: DetachedRangeUpPath): void;
    /**
     * Invoked before the attaching of a range of nodes.
     * Not invoked for replacements.
     * @param source - The content that will be attached.
     * @param destination - The location where the content will be attached.
     * @param kind - The kind of replacement that will occur.
     */
    beforeAttach(source: DetachedRangeUpPath, destination: PlaceUpPath): void;
    /**
     * Invoked after the attaching of a range of nodes.
     * Not invoked for replacements.
     * @param source - The location where the content originated.
     * @param destination - The content that was attached.
     * @param kind - The kind of replacement that will occur.
     */
    afterAttach(source: DetachedPlaceUpPath, destination: RangeUpPath): void;
    /**
     * Invoked before the detaching of a range of nodes.
     * Not invoked for replacements.
     * @param source - The content that will be detached.
     * @param destination - The location where the content will be sent to.
     * @param kind - The kind of replacement that will occur.
     */
    beforeDetach(source: RangeUpPath, destination: DetachedPlaceUpPath): void;
    /**
     * Invoked after the detaching of a range of nodes.
     * Not invoked for replacements.
     * @param source - The content that was detached.
     * @param destination - The location where the content will be attached.
     * @param kind - The kind of replacement that will occur.
     */
    afterDetach(source: PlaceUpPath, destination: DetachedRangeUpPath): void;
    /**
     * Invoked before the replacement of a range of nodes.
     *
     * Note that the `newContent` range length will always match the `oldContent` range length.
     * A replace might actually be separate detaches and attaches which have been coalesced.
     *
     * @param newContent - The content that will be attached in place of the old.
     * @param oldContent - The old that will be replaced.
     * @param oldContentDestination - The destination of the old content.
     * @param kind - The kind of replacement that will occur.
     */
    beforeReplace(newContent: DetachedRangeUpPath, oldContent: RangeUpPath, oldContentDestination: DetachedPlaceUpPath): void;
    /**
     * Invoked after the replacement of a range of nodes.
     *
     * Note that the `newContent` range length will always match the `oldContent` range length.
     * A replace might actually be separate detaches and attaches which have been coalesced.
     *
     * @param newContentSource - The place that the new content came from.
     * @param newContent - The new content.
     * @param oldContent - The content that was replaced.
     */
    afterReplace(newContentSource: DetachedPlaceUpPath, newContent: RangeUpPath, oldContent: DetachedRangeUpPath): void;
    /**
     * Invoked before content is destroyed.
     * @param content - The content that will be destroyed
     */
    beforeDestroy(content: DetachedRangeUpPath): void;
    /**
     * A sequence of nodes of length `count` is being deleted starting with `path`.
     * Called when these nodes are no longer parented under their previous parent, and do not have a new parent.
     * It is possible they may be un-deleted in the future (for example by a conflicted merge or undo).
     *
     * Not called for children of deleted nodes.
     *
     * @param path - first node in the deleted range.
     * @param count - length of deleted range.
     *
     * @deprecated Migrate to using the other events.
     */
    onDelete(path: UpPath, count: number): void;
    /**
     * @param path - location which first node of inserted range will have after insert.
     * Any nodes at this index (or after it) will be moved to the right (have their indexes increased by `content.length`).
     * @param content - content which is being inserted.
     *
     * @deprecated Migrate to using the other events.
     */
    onInsert(path: UpPath, content: ProtoNodes): void;
}

/**
 * Indicates a place between nodes in a field or at the extremity of a field.
 * Zero indicates the place at the start of the field (before the first node if any).
 * @internal
 */
export declare type PlaceIndex = number;

/**
 * A place in a field.
 * @internal
 */
export declare interface PlaceUpPath<TUpPath extends UpPath = UpPath> extends FieldUpPath<TUpPath> {
    /**
     * The location in the field.
     */
    readonly index: PlaceIndex;
}

/**
 * Apply `prefix` to `path`.
 * @internal
 */
export declare function prefixFieldPath(prefix: PathRootPrefix | undefined, path: FieldUpPath): FieldUpPath;

/**
 * Apply `prefix` to `path`.
 * @internal
 */
export declare function prefixPath(prefix: PathRootPrefix | undefined, path: UpPath | undefined): UpPath | undefined;

/**
 * The default representation a chunk (sub-sequence) of inserted content.
 *
 * TODO:
 * See issue TODO with ProtoNode.
 * Additionally, Cursors support sequences, so if using cursors, there are better ways to handle this than an array of cursors,
 * like using a cursor over all the content (starting in fields mode).
 * Long term something like TreeChunk should probably be used here.
 * @internal
 */
export declare type ProtoNodes = readonly DeltaProtoNode[];

/**
 * Delimits the extend of a range.
 * @internal
 */
declare interface Range_2 {
    /**
     * The location before the first node.
     * Must be less than or equal to `end`.
     */
    readonly start: PlaceIndex;
    /**
     * The location after the last node.
     * Must be greater than or equal to `start`.
     */
    readonly end: PlaceIndex;
}
export { Range_2 as Range }

/**
 * The result of a query about a range of keys.
 *
 * @internal
 */
export declare interface RangeQueryResult<T> {
    /**
     * The value of the first key in the query range.
     */
    value: T | undefined;
    /**
     * The length of the prefix of the query range which have the same value.
     * For example, if a RangeMap has the same value for keys 5, 6, and 7,
     * a query about the range [5, 10] would give a result with length 3.
     */
    length: number;
}

/**
 * A possibly empty range of nodes in a field.
 * This object only characterizes the location of the range, it does not own/contain the nodes in the range.
 * @internal
 */
export declare interface RangeUpPath<TUpPath extends UpPath = UpPath> extends FieldUpPath<TUpPath>, Range_2 {
}

/**
 * Use for trick to prevent self reference error `ts(2456)`.
 *
 * Prefix a type expression with `K extends _RecursiveTrick ? _RecursiveTrick : ` for some K to break the cycle.
 *
 * @remarks
 * The TypeScript compiler handles some cases of recursive types, but not others.
 * Sometimes adding an otherwise needless conditional can make a type compile.
 * Use this type in such cases.
 *
 *For example:
 * ```typescript
 * // The TypeScript compiler can't handle this case
 * type Broken<T> = FlattenKeys<
 *	{
 * 		[K in keyof T]: 0;
 * 	} & {
 * 		[K in keyof T]: Broken<T[K]>;
 * 	}
 * >;
 *
 * // Adding `K extends _RecursiveTrick ? _RecursiveTrick :` makes it compile, and has no effect on the type produced.
 * type Works<T> = FlattenKeys<
 * 	{
 * 		[K in keyof T]: 0;
 * 	} & {
 * 		// Trick added here. Since `k` never extends `never`, the second conditional option is always taken,
 * 		// making this equivalent to the broken version, except this one compiles.
 * 		[K in keyof T]: K extends _RecursiveTrick ? _RecursiveTrick : Works<T[K]>;
 * 	}
 * >;
 * ```
 *
 * This trick appears to start working in TypeScript 4.1 and is confirmed to still work in 5.0.4.
 *
 * This constant is defined to provide a way to find this documentation from types which use this pattern,
 * and to locate types which use this pattern in case they need updating for compiler changes.
 * @internal
 */
export declare type _RecursiveTrick = never;

/**
 * @internal
 */
declare interface Required_2 extends FieldKind<"Value", Multiplicity.Single> {
}
export { Required_2 as Required }

/**
 * Remove all fields which permit undefined from `T`.
 * @internal
 */
export declare type RequiredFields<T> = [
    {
    [P in keyof T as undefined extends T[P] ? never : P]: T[P];
}
][_InlineTrick];

/**
 * Alternative to the built in Record type which does not permit unexpected members,
 * and is readonly.
 *
 * @privateRemarks
 * `number` is not allowed as a key here since doing so causes the compiler to reject recursive schema.
 * The cause for this is unclear, but empirically it was the case when this comment was written.
 * @beta
 */
declare type RestrictiveReadonlyRecord<K extends symbol | string, T> = {
    readonly [P in symbol | string]: P extends K ? T : never;
};

/**
 * Allows reversion of a change made to SharedTree.
 *
 * Applications wanting to implement undo/redo support might typically maintain two stacks of Revertibles, with optional eviction policy to free up memory.
 * @beta
 */
export declare interface Revertible {
    /** Indicates the type of edit that produced this revertible. */
    readonly kind: RevertibleKind;
    /**
     * Information about which client created the edit.
     */
    readonly origin: {
        /**
         * Indicates if the {@link Revertible} is from the local client (true) or a remote client (false).
         */
        readonly isLocal: boolean;
    };
    /**
     * Can be called in order to revert a change. A successful revert will automatically discard resources.
     */
    revert(): RevertResult;
    /**
     * Should be called to garbage collect any resources associated with the revertible.
     */
    discard(): DiscardResult;
}

/**
 * The type of revertible commit.
 *
 * @beta
 */
export declare enum RevertibleKind {
    /** A typical local commit */
    Default = 0,
    /** A revertible that is the result of an undo. */
    Undo = 1,
    /** A revertible that is the result of a redo. */
    Redo = 2,
    /**
     * A revertible that is the result of a rebase and should replace a previously generated revertible.
     * todo: improve error reporting in this case
     */
    Rebase = 3
}

/**
 * The result of a revert operation.
 *
 * @beta
 */
export declare enum RevertResult {
    /** The revert was successful. */
    Success = 0,
    /** The revert failed. */
    Failure = 1
}

/**
 * Wrapper around DetachedField that can be detected at runtime.
 * @internal
 */
export declare interface RootField {
    readonly key: DetachedField;
}

/**
 * @internal
 */
export declare const rootField: DetachedField;

/**
 * FieldKey to use for the root of documents in places that need to refer to detached sequences or the root.
 * TODO: if we do want to standardize on a single value for this,
 * it likely should be namespaced or a UUID to avoid risk of collisions.
 * @internal
 */
export declare const rootFieldKey: FieldKey;

/**
 * Run a synchronous transaction on the given shared tree view.
 * This is a convenience helper around the {@link SharedTreeFork#transaction} APIs.
 * @param view - the view on which to run the transaction
 * @param transaction - the transaction function. This will be executed immediately. It is passed `view` as an argument for convenience.
 * If this function returns an `Abort` result then the transaction will be aborted. Otherwise, it will be committed.
 * @returns whether or not the transaction was committed or aborted
 * @internal
 */
export declare function runSynchronous(view: ITreeCheckout, transaction: (view: ITreeCheckout) => TransactionResult | void): TransactionResult;

/**
 * Builds schema libraries, and the schema within them.
 * @internal
 *
 * @privateRemarks
 * This class does not directly depend on any specific field kinds,
 * or bake in any defaults that might have compatibility implications.
 * All use of such implicit defaults is done by subclasses, which thus get versioning implications.
 */
export declare class SchemaBuilderBase<TScope extends string, TDefaultKind extends FieldKind, TName extends number | string = string> {
    private readonly defaultKind;
    private readonly lintConfiguration;
    private readonly libraries;
    private finalized;
    private readonly treeNodeSchema;
    private readonly adapters;
    /**
     * Prefix appended to the identifiers of all {@link TreeNodeSchema} produced by this builder.
     */
    readonly scope: TScope;
    /**
     * Used in error messages to identify content produced by this builder.
     */
    readonly name: string;
    /**
     * @param defaultKind - The default field kind to use when inferring a {@link TreeFieldSchema} from {@link ImplicitAllowedTypes}.
     */
    constructor(defaultKind: TDefaultKind, options: SchemaBuilderOptions<TScope>);
    protected scoped<Name extends TName>(name: Name): TreeNodeSchemaIdentifier<`${TScope}.${Name}`>;
    /**
     * Adds more libraries to this one.
     *
     * Unlike adding of individual schema, adding of libraries is idempotent.
     * If a single library is added multiple times (even indirectly via libraries it was added into),
     * only a single copy will be included, so they will not conflict.
     * This allows adding any library this one depends on without risk of conflicts for users of this library.
     * Contents within the added libraries can still conflict however.
     * Such errors will be reported when finalizing this builder into a library of document schema.
     */
    private addLibraries;
    protected addNodeSchema<T extends FlexTreeNodeSchema>(schema: T): void;
    private finalizeCommon;
    /**
     * Produce SchemaLibraries which capture the content added to this builder, as well as any additional SchemaLibraries that were added to it.
     * May only be called once after adding content to builder is complete.
     */
    intoLibrary(): SchemaLibrary;
    /**
     * Produce a TreeSchema which captures the content added to this builder, any additional SchemaLibraries that were added to it and a root field.
     * Can be used with schematize to provide schema aware access to document content.
     *
     * @remarks
     * May only be called once after adding content to builder is complete.
     */
    intoSchema<const TSchema extends FlexImplicitFieldSchema>(root: TSchema): FlexTreeSchema<NormalizeField<TSchema, TDefaultKind>>;
    /**
     * Define (and add to this library) a {@link ObjectNodeSchema}.
     *
     * The name must be unique among all TreeNodeSchema in the the document schema.
     */
    object<const Name extends TName, const T extends RestrictiveReadonlyRecord<string, FlexImplicitFieldSchema>>(name: Name, t: T): ObjectNodeSchema<`${TScope}.${Name}`, {
        [key in keyof T]: NormalizeField<T[key], TDefaultKind>;
    }>;
    /**
     * Same as `object` but with less type safety and works for recursive objects.
     * Reduced type safety is a side effect of a workaround for a TypeScript limitation.
     *
     * See {@link Unenforced} for details.
     *
     * TODO: Make this work with ImplicitFieldSchema.
     */
    objectRecursive<const Name extends TName, const T extends Unenforced<RestrictiveReadonlyRecord<string, FlexImplicitFieldSchema>>>(name: Name, t: T): ObjectNodeSchema<`${TScope}.${Name}`, T>;
    /**
     * Define (and add to this library) a {@link MapNodeSchema}.
     */
    map<Name extends TName, const T extends MapFieldSchema>(name: Name, fieldSchema: T): MapNodeSchema<`${TScope}.${Name}`, T>;
    /**
     * Same as `map` but with less type safety and works for recursive objects.
     * Reduced type safety is a side effect of a workaround for a TypeScript limitation.
     *
     * See {@link Unenforced} for details.
     *
     * TODO: Make this work with ImplicitFieldSchema.
     */
    mapRecursive<Name extends TName, const T extends Unenforced<MapFieldSchema>>(name: Name, t: T): MapNodeSchema<`${TScope}.${Name}`, T>;
    /**
     * Define (and add to this library) a {@link FieldNodeSchema}.
     *
     * The name must be unique among all TreeNodeSchema in the the document schema.
     *
     * @privateRemarks
     * TODO: Write and link document outlining field vs node data model and the separation of concerns related to that.
     * TODO: Maybe find a better name for this.
     */
    fieldNode<Name extends TName, const T extends FlexImplicitFieldSchema>(name: Name, fieldSchema: T): FieldNodeSchema<`${TScope}.${Name}`, NormalizeField<T, TDefaultKind>>;
    /**
     * Same as `fieldNode` but with less type safety and works for recursive objects.
     * Reduced type safety is a side effect of a workaround for a TypeScript limitation.
     *
     * See {@link Unenforced} for details.
     *
     * TODO: Make this work with ImplicitFieldSchema.
     */
    fieldNodeRecursive<Name extends TName, const T extends Unenforced<FlexImplicitFieldSchema>>(name: Name, t: T): FieldNodeSchema<`${TScope}.${Name}`, T>;
    /**
     * Define a {@link TreeFieldSchema}.
     *
     * @param kind - The [kind](https://en.wikipedia.org/wiki/Kind_(type_theory)) of this field.
     * Determine the multiplicity, viewing and editing APIs as well as the merge resolution policy.
     * @param allowedTypes - What types of children are allowed in this field.
     * @returns a {@link TreeFieldSchema} which can be used as a object field (see {@link SchemaBuilderBase.object}),
     * a map field (see {@link SchemaBuilderBase.map}), a field node (see {@link SchemaBuilderBase.fieldNode}) or the root field (see {@link SchemaBuilderBase.intoSchema}).
     *
     * @privateRemarks
     * TODO:
     * If a solution to TreeFieldSchema not being able to have extends clauses gets found,
     * consider just having users do `new TreeFieldSchema` instead?
     */
    static field<Kind extends FieldKind, T extends ImplicitAllowedTypes>(kind: Kind, allowedTypes: T): TreeFieldSchema<Kind, NormalizeAllowedTypes<T>>;
    /**
     * Define a schema for a field.
     * Same as {@link SchemaBuilderBase.field} but is less type safe and supports recursive types.
     * This API is less safe to work around a [limitation of TypeScript](https://github.com/microsoft/TypeScript/issues/55758).
     *
     * T must extends `AllowedTypes`: This cannot be enforced via TypeScript since such an "extends" clauses cause recursive types to error with:
     * "'theSchema' implicitly has type 'any' because it does not have a type annotation and is referenced directly or indirectly in its own initializer."
     *
     * TODO: Try and find a way to provide a more specific type without triggering the above error.
     */
    static fieldRecursive<Kind extends FieldKind, T extends FlexList<Unenforced<FlexTreeNodeSchema>>>(kind: Kind, ...allowedTypes: T): TreeFieldSchema<Kind, T>;
    /**
     * Normalizes an {@link ImplicitFieldSchema} into a {@link TreeFieldSchema} using this schema builder's `defaultKind`.
     */
    protected normalizeField<TSchema extends FlexImplicitFieldSchema>(schema: TSchema): NormalizeField<TSchema, TDefaultKind>;
}

/**
 * Configuration for a SchemaBuilder.
 * @internal
 */
export declare interface SchemaBuilderOptions<TScope extends string = string> {
    /**
     * Prefix appended to the identifiers of all {@link TreeNodeSchema} produced by this builder.
     * Use of [Reverse domain name notation](https://en.wikipedia.org/wiki/Reverse_domain_name_notation) or a UUIDv4 is recommended to avoid collisions.
     */
    scope: TScope;
    /**
     * Name used to refer to this builder in error messages.
     * Has no impact on the actual generated schema.
     * Defaults to scope.
     */
    name?: string;
    /**
     * Optional configuration for "linting".
     * See {@link SchemaLintConfiguration}. Currently defaults to enabling all lints.
     */
    lint?: Partial<SchemaLintConfiguration>;
    /**
     * Libraries to include in this one.
     *
     * @remarks
     * Unlike adding of individual schema, adding of libraries is idempotent.
     * If a single library is added multiple times (even indirectly via libraries it was added into),
     * only a single copy will be included, so they will not conflict.
     * This allows adding any library this one depends on without risk of conflicts for users of this library.
     * Contents within the added libraries can still conflict however.
     * Such errors will be reported when finalizing this builder into a library or document schema.
     */
    libraries?: SchemaLibrary[];
}

/**
 * Schema data that can be be used to view a document.
 * @internal
 *
 * @privateRemarks
 * It is convenient that this can be used as a StoredSchemaCollection with no conversion.
 * There there isn't a design requirement for this however, so this extends clause can be removed later if needed.
 */
export declare interface SchemaCollection {
    /**
     * {@inheritdoc SchemaCollection}
     */
    readonly nodeSchema: ReadonlyMap<TreeNodeSchemaIdentifier, FlexTreeNodeSchema>;
}

/**
 * View Schema for a `SharedTree`.
 *
 * @internal
 */
export declare interface SchemaConfiguration<TRoot extends TreeFieldSchema = TreeFieldSchema> {
    /**
     * The schema which the application wants to view the tree with.
     */
    readonly schema: FlexTreeSchema<TRoot>;
}

/**
 * Events for {@link TreeStoredSchemaSubscription}.
 *
 * TODO: consider having before and after events per subtree instead while applying anchor (and this just shows what happens at the root).
 * @internal
 */
export declare interface SchemaEvents {
    /**
     * Schema change is about to be applied.
     */
    beforeSchemaChange(newSchema: TreeStoredSchema): void;
    /**
     * Schema change was just applied.
     */
    afterSchemaChange(newSchema: TreeStoredSchema): void;
}

/**
 * Builds schema libraries, and the schema within them.
 *
 * @typeParam TScope - Scope added as a prefix to the name of every schema produced by this factory.
 * @typeParam TName - Type of names used to identify each schema produced in this factory.
 * Typically this is just `string` but it is also possible to use `string` or `number` based enums if you prefer to identify your types that way.
 *
 * @sealed @beta
 */
export declare class SchemaFactory<TScope extends string = string, TName extends number | string = string> {
    readonly scope: TScope;
    private readonly structuralTypes;
    /**
     * @param scope - Prefix appended to the identifiers of all {@link TreeNodeSchema} produced by this builder.
     * Use of [Reverse domain name notation](https://en.wikipedia.org/wiki/Reverse_domain_name_notation) or a UUIDv4 is recommended to avoid collisions.
     */
    constructor(scope: TScope);
    private scoped;
    /**
     * {@link TreeNodeSchema} for holding a JavaScript `string`.
     *
     * @remarks
     * Strings containing unpaired UTF-16 surrogate pair code units may not be handled correctly.
     *
     * These limitations come from the use of UTF-8 encoding of the strings, which requires them to be valid unicode.
     * JavaScript does not make this requirement for its strings so not all possible JavaScript strings are supported.
     * @privateRemarks
     * TODO:
     * We should be much more clear about what happens if you use problematic values.
     * We should validate and/or normalize them when inserting content.
     */
    readonly string: TreeNodeSchema<"com.fluidframework.leaf.string", NodeKind.Leaf, string, string>;
    /**
     * {@link TreeNodeSchema} for holding a JavaScript `number`.
     *
     * @remarks
     * The number is a [double-precision 64-bit binary format IEEE 754](https://en.wikipedia.org/wiki/Double-precision_floating-point_format) value, however there are some exceptions:
     * - `NaN`, and the infinities are converted to `null` (and may therefore only be used where `null` is allowed by the schema).
     * - `-0` may be converted to `0` in some cases.
     *
     * These limitations match the limitations of JSON.
     * @privateRemarks
     * TODO:
     * We should be much more clear about what happens if you use problematic values.
     * We should validate and/or normalize them when inserting content.
     */
    readonly number: TreeNodeSchema<"com.fluidframework.leaf.number", NodeKind.Leaf, number, number>;
    /**
     * {@link TreeNodeSchema} for holding a boolean.
     */
    readonly boolean: TreeNodeSchema<"com.fluidframework.leaf.boolean", NodeKind.Leaf, boolean, boolean>;
    /**
     * {@link TreeNodeSchema} for JavaScript `null`.
     *
     * @remarks
     * There are good [reasons to avoid using null](https://www.npmjs.com/package/%40rushstack/eslint-plugin#rushstackno-new-null) in JavaScript, however sometimes it is desired.
     * This {@link TreeNodeSchema} node provides the option to include nulls in trees when desired.
     * Unless directly inter-operating with existing data using null, consider other approaches, like wrapping the value in an optional field, or using a more specifically named empty object node.
     */
    readonly null: TreeNodeSchema<"com.fluidframework.leaf.null", NodeKind.Leaf, null, null>;
    /**
     * {@link TreeNodeSchema} for holding an {@link @fluidframework/core-interfaces#IFluidHandle}.
     */
    readonly handle: TreeNodeSchema<"com.fluidframework.leaf.handle", NodeKind.Leaf, IFluidHandle<FluidObject & IFluidLoadable>, IFluidHandle<FluidObject & IFluidLoadable>>;
    /**
     * Construct a class that provides the common parts all TreeNodeSchemaClass share.
     * More specific schema extend this class.
     */
    private nodeSchema;
    /**
     * Define a {@link TreeNodeSchema} for an object node.
     *
     * @param name - Unique identifier for this schema within this factory's scope.
     */
    object<const Name extends TName, const T extends RestrictiveReadonlyRecord<string, ImplicitFieldSchema>>(name: Name, t: T): TreeNodeSchemaClass<`${TScope}.${Name}`, NodeKind.Object, object & TreeNode & ObjectFromSchemaRecord<T>, object & InsertableObjectFromSchemaRecord<T>, true>;
    /**
     * Define a structurally typed {@link TreeNodeSchema} for a {@link TreeMapNodeBase}.
     *
     * @remarks
     * The {@link TreeNodeSchemaIdentifier} for this Map is defined as a function of the provided types.
     * It is still scoped to this SchemaBuilder, but multiple calls with the same arguments will return the same schema object, providing somewhat structural typing.
     * This does not support recursive types.
     *
     * If using these structurally named maps, other types in this schema builder should avoid names of the form `Map<${string}>`.
     *
     * @example
     * The returned schema should be used as a schema directly:
     * ```typescript
     * const MyMap = factory.map(factory.number);
     * type MyMap = NodeFromSchema<typeof MyMap>;
     * ```
     * Or inline:
     * ```typescript
     * factory.object("Foo", {myMap: factory.map(factory.number)});
     * ```
     * @privateRemarks
     * See note on array.
     */
    map<const T extends TreeNodeSchema | readonly TreeNodeSchema[]>(allowedTypes: T): TreeNodeSchema<`${TScope}.Map<${string}>`, NodeKind.Map, TreeMapNode<T>, ReadonlyMap<string, TreeNodeFromImplicitAllowedTypes<T>>, true>;
    /**
     * Define a {@link TreeNodeSchema} for a {@link TreeMapNodeBase}.
     *
     * @param name - Unique identifier for this schema within this factory's scope.
     *
     * @example
     * ```typescript
     * class NamedMap extends factory.map("name", factory.number) {}
     * ```
     */
    map<Name extends TName, const T extends ImplicitAllowedTypes_2>(name: Name, allowedTypes: T): TreeNodeSchemaClass<`${TScope}.${Name}`, NodeKind.Map, TreeMapNode<T>, ReadonlyMap<string, InsertableTreeNodeFromImplicitAllowedTypes<T>>, true>;
    /**
     * Define a {@link TreeNodeSchema} for a {@link (TreeMapNode:interface)}.
     *
     * @param name - Unique identifier for this schema within this factory's scope.
     *
     * @remarks See remarks on {@link SchemaFactory.namedArray_internal}.
     */
    namedMap_internal<Name extends TName | string, const T extends ImplicitAllowedTypes_2, const ImplicitlyConstructable extends boolean>(name: Name, allowedTypes: T, customizable: boolean, implicitlyConstructable: ImplicitlyConstructable): TreeNodeSchemaClass<`${TScope}.${Name}`, NodeKind.Map, TreeMapNode<T>, ReadonlyMap<string, InsertableTreeNodeFromImplicitAllowedTypes<T>>, ImplicitlyConstructable>;
    /**
     * Define a structurally typed {@link TreeNodeSchema} for a {@link (TreeArrayNode:interface)}.
     *
     * @remarks
     * The identifier for this Array is defined as a function of the provided types.
     * It is still scoped to this SchemaFactory, but multiple calls with the same arguments will return the same schema object, providing somewhat structural typing.
     * This does not support recursive types.
     *
     * If using these structurally named arrays, other types in this schema builder should avoid names of the form `Array<${string}>`.
     *
     * @example
     * The returned schema should be used as a schema directly:
     * ```typescript
     * const MyArray = factory.array(factory.number);
     * type MyArray = NodeFromSchema<typeof MyArray>;
     * ```
     * Or inline:
     * ```typescript
     * factory.object("Foo", {myArray: factory.array(factory.number)});
     * ```
     * @privateRemarks
     * The name produced at the type level here is not as specific as it could be, however doing type level sorting and escaping is a real mess.
     * There are cases where not having this full type provided will be less than ideal since TypeScript's structural types.
     * For example attempts to narrow unions of structural arrays by name won't work.
     * Planned future changes to move to a class based schema system as well as factor function based node construction should mostly avoid these issues,
     * though there may still be some problematic cases even after that work is done.
     *
     * The return value is a class, but its the type is intentionally not specific enough to indicate it is a class.
     * This prevents callers of this from sub-classing it, which is unlikely to work well (due to the ease of accidentally giving two different calls o this different subclasses)
     * when working with structural typing.
     */
    array<const T extends TreeNodeSchema | readonly TreeNodeSchema[]>(allowedTypes: T): TreeNodeSchema<`${TScope}.Array<${string}>`, NodeKind.Array, TreeArrayNode<T>, Iterable<InsertableTreeNodeFromImplicitAllowedTypes<T>>, true>;
    /**
     * Define (and add to this library) a {@link FieldNodeSchema} for a {@link (TreeArrayNode:interface)}.
     *
     * @param name - Unique identifier for this schema within this factory's scope.
     *
     * @example
     * ```typescript
     * class NamedArray extends factory.array("name", factory.number) {}
     * ```
     */
    array<const Name extends TName, const T extends ImplicitAllowedTypes_2>(name: Name, allowedTypes: T): TreeNodeSchemaClass<`${TScope}.${Name}`, NodeKind.Array, TreeArrayNode<T>, Iterable<InsertableTreeNodeFromImplicitAllowedTypes<T>>, true>;
    /**
     * Define a {@link TreeNodeSchema} for a {@link (TreeArrayNode:interface)}.
     *
     * @param name - Unique identifier for this schema within this factory's scope.
     *
     * @remarks
     * This is not intended to be used directly, use the overload of `array` which takes a name instead.
     * This is only public to work around a compiler limitation.
     *
     * @privateRemarks
     * TODO: this should be made private or protected.
     * Doing so breaks due to:
     * `src/class-tree/schemaFactoryRecursive.ts:42:9 - error TS2310: Type 'Array' recursively references itself as a base type.`
     * Once recursive APIs are better sorted out and integrated into this class, switch this back to private.
     */
    namedArray_internal<Name extends TName | string, const T extends ImplicitAllowedTypes_2, const ImplicitlyConstructable extends boolean>(name: Name, allowedTypes: T, customizable: boolean, implicitlyConstructable: ImplicitlyConstructable): TreeNodeSchemaClass<`${TScope}.${Name}`, NodeKind.Array, TreeArrayNode<T>, Iterable<InsertableTreeNodeFromImplicitAllowedTypes<T>>, ImplicitlyConstructable>;
    /**
     * Make a field optional instead of the default which is required.
     */
    optional<const T extends ImplicitAllowedTypes_2>(t: T): FieldSchema<FieldKind_2.Optional, T>;
    /**
     * Function which can be used for its compile time side-effects to tweak the evaluation order of recursive types to make them compile.
     * @remarks
     * Some related information in https://github.com/microsoft/TypeScript/issues/55758.
     *
     * Also be aware that code which relies on this tends to break VSCode's IntelliSense every time anything related to that code (even comments) is edited.
     * Running the command `TypeScript: Restart TS Server` with the schema file focused should fix it.
     * Sometimes this does not work: closing all open files except the schema before running the command can help.
     * Real compile errors (for example elsewhere in the file) can also cause the IntelliSense to not work correctly ever after `TypeScript: Restart TS Server`.
     *
     * Intellisense has also shown problems when schema files with recursive types are part of a cyclic file dependency.
     * Splitting the schema into its own file with minimal dependencies can help with this.
     *
     * Ensure `"noImplicitAny": true` is set in the `tsconfig.json`.
     * Without it, recursive types that are not working properly can infer `any` and give very non-type-safe results instead of erroring.
     *
     * @example
     * ```typescript
     * const factory = new SchemaFactory("example");
     * const recursiveReference = () => RecursiveObject;
     * factory.fixRecursiveReference(recursiveReference);
     * export class RecursiveObject extends factory.object("exampleObject", {
     * 	recursive: [recursiveReference],
     * }) {}
     * ```
     */
    fixRecursiveReference<T extends AllowedTypes_2>(...types: T): void;
}

/**
 * Checks if a {@link TreeNodeSchema} is a {@link FieldNodeSchema}.
 * @internal
 */
export declare function schemaIsFieldNode(schema: FlexTreeNodeSchema): schema is FieldNodeSchema;

/**
 * Checks if a {@link TreeNodeSchema} is a {@link LeafNodeSchema}.
 * @internal
 */
export declare function schemaIsLeaf(schema: FlexTreeNodeSchema): schema is LeafNodeSchema;

/**
 * Checks if a {@link TreeNodeSchema} is a {@link MapNodeSchema}.
 * @internal
 */
export declare function schemaIsMap(schema: FlexTreeNodeSchema): schema is MapNodeSchema;

/**
 * Checks if a {@link TreeNodeSchema} is a {@link ObjectNodeSchema}.
 * @internal
 */
export declare function schemaIsObjectNode(schema: FlexTreeNodeSchema): schema is ObjectNodeSchema;

/**
 * Schema information collected by a SchemaBuilder, including referenced libraries.
 * Can be aggregated into other libraries by adding to their builders.
 * @internal
 */
export declare interface SchemaLibrary extends SchemaCollection {
    /**
     * Schema data aggregated from a collection of libraries by a SchemaBuilder.
     */
    readonly libraries: ReadonlySet<SchemaLibraryData>;
}

/**
 * Schema data collected by a single SchemaBuilder (does not include referenced libraries).
 * @internal
 */
export declare interface SchemaLibraryData extends SchemaCollection {
    readonly name: string;
    readonly adapters: Adapters;
}

/**
 * Allows opting into and out of errors for some unusual schema patterns which are usually bugs.
 * @internal
 */
export declare interface SchemaLintConfiguration {
    readonly rejectForbidden: boolean;
    readonly rejectEmpty: boolean;
}

/**
 * Options used to schematize a `SharedTree`.
 *
 * @internal
 */
export declare interface SchematizeConfiguration<TRoot extends TreeFieldSchema = TreeFieldSchema> extends SchemaConfiguration<TRoot> {
    /**
     * Controls if and how schema from existing documents can be updated to accommodate the view schema.
     */
    readonly allowedSchemaModifications: AllowedUpdateType;
}

/**
 * Validates data complies with some particular schema.
 * Implementations are typically created by a {@link JsonValidator}.
 * @internal
 */
export declare interface SchemaValidationFunction<Schema extends TSchema> {
    /**
     * @returns Whether the data matches a schema.
     */
    check(data: unknown): data is Static<Schema>;
}

/**
 * @internal
 */
export declare interface Sequence extends FieldKind<"Sequence", Multiplicity.Sequence> {
}

/**
 * @internal
 */
export declare interface SequenceFieldEditBuilder {
    /**
     * Issues a change which inserts the `newContent` at the given `index`.
     * @param index - the index at which to insert the `newContent`.
     * @param newContent - the new content to be inserted in the field. Cursors must be in Nodes mode.
     */
    insert(index: number, newContent: ITreeCursor | readonly ITreeCursor[]): void;
    /**
     * Issues a change which deletes `count` elements starting at the given `index`.
     * @param index - The index of the first deleted element.
     * @param count - The number of elements to delete.
     */
    delete(index: number, count: number): void;
    /**
     * Issues a change which moves `count` elements starting at `sourceIndex` to `destIndex`.
     * @param sourceIndex - the index of the first moved element.
     * @param count - the number of elements to move.
     * @param destIndex - the index the elements are moved to, interpreted before detaching the moved elements.
     */
    move(sourceIndex: number, count: number, destIndex: number): void;
}

/**
 * SharedTree is a hierarchical data structure for collaboratively editing JSON-like trees
 * of objects, arrays, and other data types.
 *
 * @beta
 */
export declare class SharedTree implements ITree {
    static getFactory(): IChannelFactory;
    schematize<TRoot extends ImplicitFieldSchema>(config: TreeConfiguration<TRoot>): TreeView<TreeFieldFromImplicitField<TRoot>>;
    private useFactory;
    get id(): string;
    get attributes(): IChannelAttributes;
    get handle(): IFluidHandle;
    get IFluidLoadable(): IFluidLoadable;
    getAttachSummary(fullTree?: boolean | undefined, trackState?: boolean | undefined, telemetryContext?: ITelemetryContext | undefined): ISummaryTreeWithStats;
    summarize(fullTree?: boolean | undefined, trackState?: boolean | undefined, telemetryContext?: ITelemetryContext | undefined, incrementalSummaryContext?: IExperimentalIncrementalSummaryContext | undefined): Promise<ISummaryTreeWithStats>;
    isAttached(): boolean;
    connect(services: IChannelServices): void;
    getGCData(fullGC?: boolean | undefined): IGarbageCollectionData;
}

/**
 * Copy of data from an {@link ISharedTree} at some point in time.
 * @remarks
 * This is unrelated to Fluids concept of "snapshots".
 * @internal
 */
export declare interface SharedTreeContentSnapshot {
    /**
     * The schema stored in the document.
     *
     * @remarks
     * Edits to the schema can mutate the schema stored of the tree which took this snapshot (but this snapshot will remain the same)
     * This is mainly useful for debugging cases where schematize reports an incompatible view schema.
     */
    readonly schema: TreeStoredSchema;
    /**
     * All {@link TreeStatus#InDocument} content.
     */
    readonly tree: JsonableTree[];
    /**
     * All {@link TreeStatus#Removed} content.
     */
    readonly removed: [string | number | undefined, number, JsonableTree][];
}

/**
 * A channel factory that creates {@link ISharedTree}s.
 * @internal
 */
export declare class SharedTreeFactory implements IChannelFactory {
    private readonly options;
    readonly type: string;
    readonly attributes: IChannelAttributes;
    constructor(options?: SharedTreeOptions);
    load(runtime: IFluidDataStoreRuntime, id: string, services: IChannelServices, channelAttributes: Readonly<IChannelAttributes>): Promise<ISharedTree>;
    create(runtime: IFluidDataStoreRuntime, id: string): ISharedTree;
}

/**
 * @internal
 */
export declare interface SharedTreeOptions extends Partial<ICodecOptions> {
    /**
     * The {@link ForestType} indicating which forest type should be created for the SharedTree.
     */
    forest?: ForestType;
    summaryEncodeType?: TreeCompressionStrategy;
}

/**
 * This module provides support for reading and writing a human readable (and
 * editable) tree format.
 *
 * This implementation can handle all trees (so it does not need a fallback for any special cases),
 * and is not optimized.
 *
 * It's suitable for testing and debugging,
 * though it could also reasonably be used as a fallback for edge cases or for small trees.
 *
 * TODO: Use placeholders.
 * build / add operations should be able to include detached ranges instead of children directly.
 * summaries should be able to reference unloaded chunks instead of having children directly.
 * Leverage placeholders in the types below to accomplish this.
 * Determine how this relates to Cursor: should cursor be generic over placeholder values?
 * (Could use them for errors to allow non erroring cursors?)
 *
 * Note:
 * Currently a lot of Tree's codebase is using json for serialization.
 * Because putting json strings inside json works poorly (adds lots of escaping),
 * for now this library actually outputs and inputs the Json compatible type JsonableTree
 * rather than actual strings.
 */
/**
 * Create a cursor, in `nodes` mode at the root of the provided tree.
 *
 * @returns an {@link ITreeCursorSynchronous} in nodes mode for a single {@link JsonableTree}.
 * @remarks
 * Do not confuse this with {@link JsonableTree} with the JSON domain:
 * this takes in data in a specific format that is json compatible (except for FluidHandle values).
 * That is distinct from treating arbitrary JSON data as a tree in the JSON domain.
 * @internal
 */
export declare function singleTextCursor(root: JsonableTree): ITreeCursorSynchronous;

/**
 * Create a schema for a node with no state.
 * @remarks
 * This is commonly used in unions when the only information needed is which kind of node the value is.
 * Enums are a common example of this pattern.
 * @internal
 */
export declare function singletonSchema<TScope extends string, TName extends string | number>(factory: SchemaFactory<TScope, TName>, name: TName): {
    new (data?: EmptyObject): {
        readonly value: TName;
    };
    readonly identifier: `${TScope}.${TName}`;
    readonly kind: NodeKind.Object;
    readonly info: unknown;
    readonly implicitlyConstructable: true;
};

/**
 * A UUID which identifies a node in the tree.
 * This key is universally unique and stable forever; therefore it is safe to persist as data in a SharedTree or other DDS/database.
 * When not persisted or serialized, it is preferable to use a {@link LocalNodeKey} instead for better performance.
 * @internal
 */
export declare type StableNodeKey = Brand<StableId, "Stable Node Key">;

/**
 * Create a cursor, in `fields` mode at the `detachedField` under the provided `root`.
 *
 * @returns an {@link ITreeCursorSynchronous} for `detachedField` of `root` in `fields` mode.
 * @internal
 */
export declare function stackTreeFieldCursor<TNode>(adapter: CursorAdapter<TNode>, root: TNode, detachedField?: DetachedField): CursorWithNode<TNode>;

/**
 * Create a cursor, in `nodes` mode at the root of the provided tree.
 *
 * @returns an {@link ITreeCursorSynchronous} for a single root in `nodes` mode.
 * @internal
 */
export declare function stackTreeNodeCursor<TNode>(adapter: CursorAdapter<TNode>, root: TNode): CursorWithNode<TNode>;

/**
 * Collection of TreeNodeSchema data that can be stored in a document.
 *
 * @remarks
 * Note: the owner of this may modify it over time:
 * thus if needing to hang onto a specific version, make a copy.
 * @internal
 */
export declare interface StoredSchemaCollection {
    /**
     * {@inheritdoc StoredSchemaCollection}
     */
    readonly nodeSchema: ReadonlyMap<TreeNodeSchemaIdentifier, TreeNodeStoredSchema>;
}

/**
 * @internal
 */
export declare class test_RecursiveObject extends test_RecursiveObject_base {
}

/**
 * To make API-Extractor happy, the base type has to be exported in addition to the actual schema class.
 * Ideally this would be inlined into the class definition below.
 *
 * TODO:
 * Fix API-Extractor to support class based schema, and remove this workaround (inline this type into `RecursiveObject extends` below).
 * Error: src/class-tree/testRecursiveDomain.ts:32:1 - (ae-forgotten-export) The symbol "RecursiveObject_base" needs to be exported by the entry point index.d.ts
 * See https://github.com/microsoft/rushstack/issues/4429
 * @internal
 */
export declare const test_RecursiveObject_base: TreeNodeSchemaClass<"Test Recursive Domain.testObject", NodeKind.Object, object & TreeNode & ObjectFromSchemaRecord<    {
readonly recursive: FieldSchema<import("./schemaTypes").FieldKind.Optional, readonly [() => typeof test_RecursiveObject]>;
readonly number: TreeNodeSchema<"com.fluidframework.leaf.number", NodeKind.Leaf, number, number>;
}>, object & InsertableObjectFromSchemaRecord<    {
readonly recursive: FieldSchema<import("./schemaTypes").FieldKind.Optional, readonly [() => typeof test_RecursiveObject]>;
readonly number: TreeNodeSchema<"com.fluidframework.leaf.number", NodeKind.Leaf, number, number>;
}>, true>;

/**
 * Describes the result of a transaction.
 * Transactions may either succeed and commit, or fail and abort.
 * @internal
 */
export declare enum TransactionResult {
    /**
     * Indicates the transaction failed.
     */
    Abort = 0,
    /**
     * Indicates the transaction succeeded.
     */
    Commit = 1
}

/**
 * The `Tree` object holds various functions for analyzing {@link TreeNode}s.
 * @beta
 */
export declare const Tree: TreeApi;

/**
 * @internal
 */
export declare interface TreeAdapter {
    readonly output: TreeNodeSchemaIdentifier;
    readonly input: TreeNodeSchemaIdentifier;
}

/**
 * Provides various functions for analyzing {@link TreeNode}s.
 *
 * @privateRemarks
 * Inlining the typing of this interface onto the `Tree` object provides slightly different .d.ts generation,
 * which avoids typescript expanding the type of TreeNodeSchema and thus encountering
 * https://github.com/microsoft/rushstack/issues/1958.
 * @beta
 */
export declare interface TreeApi {
    /**
     * The schema information for this node.
     */
    schema<T extends TreeNode | TreeLeafValue>(node: T): TreeNodeSchema<string, NodeKind, unknown, T>;
    /**
     * Narrow the type of the given value if it satisfies the given schema.
     * @example
     * ```ts
     * if (node.is(myNode, point)) {
     *     const y = myNode.y; // `myNode` is now known to satisfy the `point` schema and therefore has a `y` coordinate.
     * }
     * ```
     */
    is<TSchema extends TreeNodeSchema>(value: unknown, schema: TSchema): value is NodeFromSchema<TSchema>;
    /**
     * Return the node under which this node resides in the tree (or undefined if this is a root node of the tree).
     */
    parent(node: TreeNode): TreeNode | undefined;
    /**
     * The key of the given node under its parent.
     * @remarks
     * If `node` is an element in a {@link (TreeArrayNode:interface)}, this returns the index of `node` in the list (a `number`).
     * Otherwise, this returns the key of the field that it is under (a `string`).
     */
    key(node: TreeNode): string | number;
    /**
     * Register an event listener on the given node.
     * @returns A callback function which will deregister the event.
     * This callback should be called only once.
     */
    on<K extends keyof TreeNodeEvents>(node: TreeNode, eventName: K, listener: TreeNodeEvents[K]): () => void;
    /**
     * Returns the {@link TreeStatus} of the given node.
     */
    readonly status: (node: TreeNode) => TreeStatus;
}

/**
 * A {@link TreeNode} which implements 'readonly T[]' and the list mutation APIs.
 * @beta
 */
export declare interface TreeArrayNode<T extends ImplicitAllowedTypes_2 = ImplicitAllowedTypes_2> extends TreeNode, TreeArrayNodeBase<TreeNodeFromImplicitAllowedTypes<T>, InsertableTreeNodeFromImplicitAllowedTypes<T>, TreeArrayNode> {
}

/**
 * A {@link TreeNode} which implements 'readonly T[]' and the list mutation APIs.
 * @beta
 */
export declare const TreeArrayNode: {
    /**
     * Wrap an iterable of items to inserted as consecutive items in a list.
     * @remarks
     * The object returned by this function can be inserted into a {@link (TreeArrayNode:interface)}.
     * Its contents will be inserted consecutively in the corresponding location in the list.
     * @example
     * ```ts
     * list.insertAtEnd(TreeArrayNode.inline(iterable))
     * ```
     */
    inline: <T>(content: Iterable<T>) => IterableTreeListContent<T>;
};

/**
 * A generic List type, used to defined types like {@link (TreeArrayNode:interface)}.
 * @beta
 */
declare interface TreeArrayNodeBase<out T, in TNew, in TMoveFrom> extends ReadonlyArray<T>, TreeNode {
    /**
     * Inserts new item(s) at a specified location.
     * @param index - The index at which to insert `value`.
     * @param value - The content to insert.
     * @throws Throws if `index` is not in the range [0, `list.length`).
     */
    insertAt(index: number, ...value: (TNew | IterableTreeListContent<TNew>)[]): void;
    /**
     * Inserts new item(s) at the start of the list.
     * @param value - The content to insert.
     */
    insertAtStart(...value: (TNew | IterableTreeListContent<TNew>)[]): void;
    /**
     * Inserts new item(s) at the end of the list.
     * @param value - The content to insert.
     */
    insertAtEnd(...value: (TNew | IterableTreeListContent<TNew>)[]): void;
    /**
     * Removes the item at the specified location.
     * @param index - The index at which to remove the item.
     * @throws Throws if `index` is not in the range [0, `list.length`).
     */
    removeAt(index: number): void;
    /**
     * Removes all items between the specified indices.
     * @param start - The starting index of the range to remove (inclusive). Defaults to the start of the list.
     * @param end - The ending index of the range to remove (exclusive).
     * @throws Throws if `start` is not in the range [0, `list.length`).
     * @throws Throws if `end` is less than `start`.
     * If `end` is not supplied or is greater than the length of the list, all items after `start` are deleted.
     */
    removeRange(start?: number, end?: number): void;
    /**
     * Moves the specified item to the start of the list.
     * @param sourceIndex - The index of the item to move.
     * @throws Throws if `sourceIndex` is not in the range [0, `list.length`).
     */
    moveToStart(sourceIndex: number): void;
    /**
     * Moves the specified item to the start of the list.
     * @param sourceIndex - The index of the item to move.
     * @param source - The source list to move the item out of.
     * @throws Throws if `sourceIndex` is not in the range [0, `list.length`).
     */
    moveToStart(sourceIndex: number, source: TMoveFrom): void;
    /**
     * Moves the specified item to the end of the list.
     * @param sourceIndex - The index of the item to move.
     * @throws Throws if `sourceIndex` is not in the range [0, `list.length`).
     */
    moveToEnd(sourceIndex: number): void;
    /**
     * Moves the specified item to the end of the list.
     * @param sourceIndex - The index of the item to move.
     * @param source - The source list to move the item out of.
     * @throws Throws if `sourceIndex` is not in the range [0, `list.length`).
     */
    moveToEnd(sourceIndex: number, source: TMoveFrom): void;
    /**
     * Moves the specified item to the desired location in the list.
     * @param index - The index to move the item to.
     * This is based on the state of the list before moving the source item.
     * @param sourceIndex - The index of the item to move.
     * @throws Throws if any of the input indices are not in the range [0, `list.length`).
     */
    moveToIndex(index: number, sourceIndex: number): void;
    /**
     * Moves the specified item to the desired location in the list.
     * @param index - The index to move the item to.
     * @param sourceIndex - The index of the item to move.
     * @param source - The source list to move the item out of.
     * @throws Throws if any of the input indices are not in the range [0, `list.length`).
     */
    moveToIndex(index: number, sourceIndex: number, source: TMoveFrom): void;
    /**
     * Moves the specified items to the start of the list.
     * @param sourceStart - The starting index of the range to move (inclusive).
     * @param sourceEnd - The ending index of the range to move (exclusive)
     * @throws Throws if either of the input indices are not in the range [0, `list.length`) or if `sourceStart` is greater than `sourceEnd`.
     */
    moveRangeToStart(sourceStart: number, sourceEnd: number): void;
    /**
     * Moves the specified items to the start of the list.
     * @param sourceStart - The starting index of the range to move (inclusive).
     * @param sourceEnd - The ending index of the range to move (exclusive)
     * @param source - The source list to move items out of.
     * @throws Throws if the types of any of the items being moved are not allowed in the destination list,
     * if either of the input indices are not in the range [0, `list.length`) or if `sourceStart` is greater than `sourceEnd`.
     */
    moveRangeToStart(sourceStart: number, sourceEnd: number, source: TMoveFrom): void;
    /**
     * Moves the specified items to the end of the list.
     * @param sourceStart - The starting index of the range to move (inclusive).
     * @param sourceEnd - The ending index of the range to move (exclusive)
     * @throws Throws if either of the input indices are not in the range [0, `list.length`) or if `sourceStart` is greater than `sourceEnd`.
     */
    moveRangeToEnd(sourceStart: number, sourceEnd: number): void;
    /**
     * Moves the specified items to the end of the list.
     * @param sourceStart - The starting index of the range to move (inclusive).
     * @param sourceEnd - The ending index of the range to move (exclusive)
     * @param source - The source list to move items out of.
     * @throws Throws if the types of any of the items being moved are not allowed in the destination list,
     * if either of the input indices are not in the range [0, `list.length`) or if `sourceStart` is greater than `sourceEnd`.
     */
    moveRangeToEnd(sourceStart: number, sourceEnd: number, source: TMoveFrom): void;
    /**
     * Moves the specified items to the desired location within the list.
     * @param index - The index to move the items to.
     * This is based on the state of the list before moving the source items.
     * @param sourceStart - The starting index of the range to move (inclusive).
     * @param sourceEnd - The ending index of the range to move (exclusive)
     * @throws Throws if any of the input indices are not in the range [0, `list.length`) or if `sourceStart` is greater than `sourceEnd`.
     */
    moveRangeToIndex(index: number, sourceStart: number, sourceEnd: number): void;
    /**
     * Moves the specified items to the desired location within the list.
     * @param index - The index to move the items to.
     * @param sourceStart - The starting index of the range to move (inclusive).
     * @param sourceEnd - The ending index of the range to move (exclusive)
     * @param source - The source list to move items out of.
     * @throws Throws if the types of any of the items being moved are not allowed in the destination list,
     * if any of the input indices are not in the range [0, `list.length`) or if `sourceStart` is greater than `sourceEnd`.
     */
    moveRangeToIndex(index: number, sourceStart: number, sourceEnd: number, source: TMoveFrom): void;
}

/**
 * Selects which heuristics to use when encoding tree content.
 * All encoding options here are compatible with the same decoder:
 * the selection here does not impact compatibility.
 * @internal
 */
export declare enum TreeCompressionStrategy {
    /**
     * Optimized for encoded size.
     * Use this in production to reduce bandwidth and storage use.
     */
    Compressed = 0,
    /**
     * Optimized for human readability.
     * Use this when debugging or testing and needing to inspect encoded tree content.
     */
    Uncompressed = 1
}

/**
 * Configuration for how to {@link ITree.schematize|schematize} a tree.
 * @beta
 */
export declare class TreeConfiguration<TSchema extends ImplicitFieldSchema = ImplicitFieldSchema> {
    readonly schema: TSchema;
    readonly initialTree: () => InsertableTreeFieldFromImplicitField<TSchema>;
    /**
     * @param schema - The schema which the application wants to view the tree with.
     * @param initialTree - A function that returns the default tree content to initialize the tree with iff the tree is uninitialized
     * (meaning it does not even have any schema set at all).
     * If `initialTree` returns any actual node instances, they should be recreated each time `initialTree` runs.
     * This is because if the config is used a second time any nodes that were not recreated could error since nodes cannot be inserted into the tree multiple times.
     */
    constructor(schema: TSchema, initialTree: () => InsertableTreeFieldFromImplicitField<TSchema>);
}

/**
 * Content that can populate a `SharedTree`.
 *
 * @internal
 */
export declare interface TreeContent<TRoot extends TreeFieldSchema = TreeFieldSchema> extends SchemaConfiguration<TRoot> {
    /**
     * Default tree content to initialize the tree with iff the tree is uninitialized
     * (meaning it does not even have any schema set at all).
     */
    readonly initialTree: InsertableFlexField<TRoot> | readonly ITreeCursorSynchronous[] | ITreeCursorSynchronous;
}

/**
 * A common context of a "forest" of EditableTrees.
 * It handles group operations like transforming cursors into anchors for edits.
 * @internal
 */
export declare interface TreeContext extends ISubscribable<ForestEvents> {
    /**
     * Gets the root field of the tree.
     */
    get root(): FlexTreeField;
    /**
     * Schema used within this context.
     * All data must conform to these schema.
     */
    readonly schema: FlexTreeSchema;
    readonly nodeKeys: NodeKeys;
}

/**
 * Information needed to interpret a subtree described by {@link ContextuallyTypedNodeData} and {@link ContextuallyTypedFieldData}.
 * TODO:
 * Currently being exposed at the package level which also requires us to export MapTree at the package level.
 * Refactor the FieldGenerator to use JsonableTree instead of MapTree, and convert them internally.
 * @internal
 */
export declare interface TreeDataContext {
    /**
     * Schema for the document which the tree will be used in.
     */
    readonly schema: FlexTreeSchema;
    /**
     * Procedural data generator for fields.
     * Fields which provide generators here can be omitted in the input contextually typed data.
     *
     * @remarks
     * TODO:
     * For implementers of this which are not pure (like identifier generation),
     * order of invocation should be made consistent and documented.
     * This will be important for identifier elision optimizations in tree encoding for session based identifier generation.
     */
    fieldSource?(key: FieldKey, schema: TreeFieldStoredSchema): undefined | FieldGenerator;
}

/**
 * This file provides an API for working with trees which is type safe even when schema is not known.
 * This means no editing is allowed.
 *
 * Schema aware APIs for working with trees should superset this, while sub-setting EditableTree.
 *
 * TODO:
 * This API should replace EditableTree as the default public API for tree access.
 * SchemaAware builds on this, adding editing and type safe APIs which can be accessed via SchematizeView.
 * Once this is finished, the unsafe EditableTree types can be removed (or converted to package internal documentation for the proxies).
 */
/**
 * An event raised on a {@link FlexTreeNode}.
 *
 * @internal
 */
export declare interface TreeEvent {
    /**
     * The node of the tree where the listener receiving the event is attached.
     */
    readonly target: FlexTreeNode;
}

/**
 * A channel factory that creates an {@link ITree}.
 * @internal
 */
export declare class TreeFactory implements IChannelFactory {
    private readonly options;
    readonly type: string;
    readonly attributes: IChannelAttributes;
    constructor(options: TreeOptions);
    load(runtime: IFluidDataStoreRuntime, id: string, services: IChannelServices, channelAttributes: Readonly<IChannelAttributes>): Promise<ITree>;
    create(runtime: IFluidDataStoreRuntime, id: string): ITree;
}

/**
 * Converts ImplicitFieldSchema to the corresponding tree node's field type.
 * @beta
 */
export declare type TreeFieldFromImplicitField<TSchema extends ImplicitFieldSchema = FieldSchema> = TSchema extends FieldSchema<infer Kind, infer Types> ? ApplyKind<TreeNodeFromImplicitAllowedTypes<Types>, Kind> : TSchema extends ImplicitAllowedTypes_2 ? TreeNodeFromImplicitAllowedTypes<TSchema> : unknown;

/**
 * All policy for a specific field,
 * including functionality that does not have to be kept consistent across versions or deterministic.
 *
 * This can include policy for how to use this schema for "view" purposes, and well as how to expose editing APIs.
 *
 * @remarks
 * `Types` here must extend `AllowedTypes`, but this cannot be enforced with an "extends" clause: see {@link Unenforced} for details.
 *
 * @typeParam TKind - The kind of field.
 * @typeParam TTypes - The types allowed by the field.
 *
 * @sealed
 * @internal
 */
export declare class TreeFieldSchema<out TKind extends FieldKind = FieldKind, const out TTypes extends Unenforced<AllowedTypes> = AllowedTypes> implements TreeFieldStoredSchema {
    readonly kind: TKind;
    readonly allowedTypes: TTypes;
    /**
     * Schema for a field which must always be empty.
     */
    static readonly empty: TreeFieldSchema<Forbidden, readonly []>;
    /**
     * Constructs a TreeFieldSchema.
     * @privateRemarks
     * Alias for the constructor, but with extends clause for the `Types` parameter that {@link TreeFieldSchema} can not have (due to recursive type issues).
     */
    static create<TKind extends FieldKind, const Types extends AllowedTypes>(kind: TKind, allowedTypes: Types): TreeFieldSchema<TKind, Types>;
    /**
     * Constructs a TreeFieldSchema, but missing the extends clause which breaks most recursive types.
     * @remarks
     * `Types` here must extend `AllowedTypes`, but this cannot be enforced with an "extends" clause: see {@link Unenforced} for details.
     * Prefer {@link TreeFieldSchema.create} when possible.
     */
    static createUnsafe<TKind extends FieldKind, const Types extends Unenforced<AllowedTypes>>(kind: TKind, allowedTypes: Types): TreeFieldSchema<TKind, Types>;
    protected _typeCheck?: MakeNominal;
    /**
     * This is computed lazily since types can be recursive, which makes evaluating this have to happen after all the schema are defined.
     */
    private readonly lazyTypes;
    /**
     * @param kind - The {@link https://en.wikipedia.org/wiki/Kind_(type_theory) | kind} of this field.
     * Determine the multiplicity, viewing and editing APIs as well as the merge resolution policy.
     * @param allowedTypes - What types of tree nodes are allowed in this field.
     */
    private constructor();
    /**
     * Types which are allowed in this field (by {@link TreeNodeSchemaIdentifier}), in a format optimized for stored schema.
     * This is the same set of types in {@link TreeFieldSchema.allowedTypes}, just in a different format.
     */
    get types(): TreeTypeSet;
    /**
     * Types which are allowed in this field.
     * This is the same set of types in {@link TreeFieldSchema.allowedTypes}, just as a set with laziness removed.
     * @privateRemarks
     * TODO:
     * 3 ways to access the allowed types are now exposed.
     * Reducing this and/or renaming the more friendly options to take the shorter name (`types`)
     * would be a good idea.
     */
    get allowedTypeSet(): AllowedTypeSet;
    /**
     * If exactly one type of child is allowed in this field, it is provided here.
     * @remarks
     * Some code paths (like unboxing and compressed tree encoding) special case schema with exactly one allowed type.
     * This field allows for simple and optimized handling of this case.
     */
    get monomorphicChildType(): FlexTreeNodeSchema | undefined;
    /**
     * Compare this schema to another.
     *
     * @returns true iff the schema are identical.
     */
    equals(other: TreeFieldSchema): boolean;
}

/**
 * Schema for a field.
 * Object implementing this interface should never be modified.
 * @internal
 */
export declare interface TreeFieldStoredSchema {
    readonly kind: FieldKindSpecifier;
    /**
     * The set of allowed child types.
     * If not specified, types are unconstrained.
     */
    readonly types?: TreeTypeSet;
}

/**
 * Value that may be stored as a leaf node.
 * @beta
 */
export declare type TreeLeafValue = number | string | boolean | IFluidHandle | null;

/**
 * @internal
 */
export declare interface TreeLocation {
    readonly range: FieldLocation | DetachedField;
    readonly index: number;
}

/**
 * A map of string keys to tree objects.
 *
 * @beta
 */
export declare interface TreeMapNode<T extends ImplicitAllowedTypes_2> extends TreeMapNodeBase<TreeNodeFromImplicitAllowedTypes<T>, InsertableTreeNodeFromImplicitAllowedTypes<T>> {
}

/**
 * A map of string keys to tree objects.
 *
 * @privateRemarks
 * Add support for `clear` once we have established merge semantics for it.
 *
 * @beta
 */
export declare interface TreeMapNodeBase<TOut, TIn = TOut> extends ReadonlyMap<string, TOut>, TreeNode {
    /**
     * Adds or updates an entry in the map with a specified `key` and a `value`.
     *
     * @param key - The key of the element to add to the map.
     * @param value - The value of the element to add to the map.
     *
     * @remarks
     * Setting the value at a key to `undefined` is equivalent to calling {@link TreeMapNodeBase.delete} with that key.
     */
    set(key: string, value: TIn | undefined): void;
    /**
     * Removes the specified element from this map by its `key`.
     *
     * @remarks
     * Note: unlike JavaScript's Map API, this method does not return a flag indicating whether or not the value was
     * deleted.
     *
     * @privateRemarks
     * Regarding the choice to not return a boolean: Since this data structure is distributed in nature, it isn't
     * possible to tell whether or not the item was deleted as a result of this method call. Returning a "best guess"
     * is more likely to create issues / promote bad usage patterns than offer useful information.
     *
     * @param key - The key of the element to remove from the map.
     */
    delete(key: string): void;
}

/**
 * @internal
 */
export declare const enum TreeNavigationResult {
    /**
     * Attempt to navigate cursor to a key or index that is outside the client's view.
     */
    NotFound = -1,
    /**
     * Attempt to navigate cursor to a portion of the tree that has not yet been loaded.
     */
    Pending = 0,
    /**
     * ITreeReader successfully navigated to the desired node.
     */
    Ok = 1
}

/**
 * A non-{@link LeafNodeSchema|leaf} SharedTree node. Includes objects, lists, and maps.
 *
 * @remarks
 * Base type which all nodes implement.
 *
 * This can be used as a type to indicate/document values which should be tree nodes,
 * but currently does not provide strong type checking for this.
 * Additionally runtime use of this class object (for example when used with `instanceof` or subclassed), is not supported:
 * it may be replaced with an interface or union in the future.
 * @privateRemarks
 * Adding a member which all nodes have, like a type symbol, would produce much stronger typing for this.
 * This is only a class to enable stronger typing in a future change,
 * but other future changes may want to replace it with a branded interface if the runtime oddities related to this are not cleaned up.
 *
 * Currently not all node implications include this in their prototype chain (some hide it with a proxy), and thus cause `instanceof` to fail.
 * This results in the runtime and compile time behavior of `instanceof` differing.
 * TypeScript 5.3 allows altering the compile time behavior of `instanceof`.
 * The runtime behavior can be changed by implementing `Symbol.hasInstance`.
 * One of those approaches could be used to resolve this inconsistency if TreeNode is kept as a class.
 * @beta
 */
export declare abstract class TreeNode {
}

/**
 * A collection of events that can be raised by a {@link TreeNode}.
 * @beta
 */
export declare interface TreeNodeEvents {
    /**
     * Raised on a node right after a change is applied to one of its fields or the fields of a descendant node.
     */
    afterChange(): void;
}

/**
 * Type of of tree node for a field of the given schema.
 * @beta
 */
declare type TreeNodeFromImplicitAllowedTypes<TSchema extends ImplicitAllowedTypes_2 = TreeNodeSchema> = TSchema extends TreeNodeSchema ? NodeFromSchema<TSchema> : TSchema extends AllowedTypes_2 ? NodeFromSchema<FlexListToUnion<TSchema>> : unknown;

/**
 * Schema for a tree node.
 * @typeParam Name - The full (including scope) name/identifier for the schema.
 * @typeParam Kind - Which kind of node this schema is for.
 * @typeParam TNode - API for nodes that use this schema.
 * @typeParam TBuild - Data which can be used to construct an unhydrated node of this type.
 * @remarks
 * Captures the schema both as runtime data and compile time type information.
 * @beta
 */
export declare type TreeNodeSchema<Name extends string = string, Kind extends NodeKind = NodeKind, TNode = unknown, TBuild = never, ImplicitlyConstructable extends boolean = boolean> = TreeNodeSchemaClass<Name, Kind, TNode, TBuild, ImplicitlyConstructable> | TreeNodeSchemaNonClass<Name, Kind, TNode, TBuild, ImplicitlyConstructable>;

/**
 * T must extend TreeSchemaSpecification.
 * This can not be enforced using TypeScript since doing so breaks recursive type support.
 * See note on SchemaBuilder.fieldRecursive.
 * @internal
 */
export declare abstract class TreeNodeSchemaBase<const out Name extends string = string, const out Specification = unknown> {
    readonly builder: Named<string>;
    readonly name: TreeNodeSchemaIdentifier<Name>;
    readonly info: Specification;
    readonly stored: TreeNodeStoredSchema;
    protected _typeCheck?: MakeNominal;
    protected constructor(builder: Named<string>, name: TreeNodeSchemaIdentifier<Name>, info: Specification, stored: TreeNodeStoredSchema);
    abstract getFieldSchema(field: FieldKey): TreeFieldSchema;
}

/**
 * Tree node schema which is implemented using a class.
 * @remarks
 * Instances of this class are nodes in the tree.
 * This is also a constructor so that it can be subclassed.
 *
 * Using classes in this way allows introducing a named type and a named value at the same time, helping keep the runtime and compile time information together and easy to refer to un a uniform way.
 * Additionally, this works around https://github.com/microsoft/TypeScript/issues/55832 which causes similar patterns with less explicit types to infer "any" in the d.ts file.
 * @beta
 */
declare interface TreeNodeSchemaClass<out Name extends string = string, out Kind extends NodeKind = NodeKind, out TNode = unknown, in TInsertable = never, out ImplicitlyConstructable extends boolean = boolean> extends TreeNodeSchemaCore<Name, Kind, ImplicitlyConstructable> {
    /**
     * Constructs an {@link Unhydrated} node with this schema.
     * @remarks
     * This constructor is also used internally to construct hydrated nodes with a different parameter type.
     * Therefor overriding this constructor is not type-safe and is not supported.
     * @sealed
     */
    new (data: TInsertable): Unhydrated<TNode>;
}

/**
 * Data common to all tree node schema.
 * @beta
 */
declare interface TreeNodeSchemaCore<out Name extends string, out Kind extends NodeKind, out ImplicitlyConstructable extends boolean> {
    readonly identifier: Name;
    readonly kind: Kind;
    readonly info: unknown;
    /**
     * When constructing insertable content,
     * data that could be passed to the node's constructor can be used instead of an {@link Unhydrated} node
     * iff implicitlyConstructable is true.
     * @privateRemarks
     * Currently the logic for traversing insertable content,
     * both to build trees and to hydrate them does not defer to the schema classes to handle the policy,
     * so if their constructors differ from what is supported, some cases will not work.
     * Setting this to false adjusts the insertable types to disallow cases which could be impacted by these inconsistencies.
     */
    readonly implicitlyConstructable: ImplicitlyConstructable;
}

/**
 * Identifier for a TreeNode schema.
 * Also known as "Definition"
 *
 * Stable identifier, used when persisting data.
 * @internal
 */
export declare type TreeNodeSchemaIdentifier<TName extends string = string> = Brand<TName, "tree.TreeNodeSchemaIdentifier">;

/**
 * Schema which is not a class.
 * @remarks
 * This is used for schema which cannot have their instances constructed using constructors, like leaf schema.
 * @privateRemarks
 * Non-class based schema can have issues with recursive types due to https://github.com/microsoft/TypeScript/issues/55832.
 * @beta
 */
declare interface TreeNodeSchemaNonClass<out Name extends string = string, out Kind extends NodeKind = NodeKind, out TNode = unknown, in TInsertable = never, out ImplicitlyConstructable extends boolean = boolean> extends TreeNodeSchemaCore<Name, Kind, ImplicitlyConstructable> {
    create(data: TInsertable): TNode;
}

/**
 * @internal
 */
export declare abstract class TreeNodeStoredSchema {
    protected _typeCheck: MakeNominal;
    /**
     * @privateRemarks
     * Returns TreeNodeSchemaDataFormat.
     * This is uses an opaque type to avoid leaking these types out of the package,
     * and is runtime validated by the codec.
     */
    abstract encode(): ErasedTreeNodeSchemaDataFormat;
}

/**
 * Configuration to specialize a Tree DDS for a particular use.
 * @internal
 */
export declare interface TreeOptions extends SharedTreeOptions {
    /**
     * Name appended to {@link @fluidframework/datastore-definitions#IChannelFactory."type"} to identify this factory configuration.
     * @privateRemarks
     * TODO: evaluate if this design is a good idea, or if "subtype" should be removed.
     * TODO: evaluate if schematize should be separated from DDS construction.
     */
    readonly subtype?: string;
}

/**
 * Creates a new view schema using the stored schema.
 * @remarks
 * This is really only suitable for use with code that happens to need access to things which require a view schema, but isn't actually schema aware.
 * If the input schema came from a view schema, it will not return the same view schema, and will not be compatible:
 * the returned TreeSchema is simply one which schematize will not object to.
 * Assumes the schema uses the default field kinds.
 * @internal
 */
export declare function treeSchemaFromStoredSchema(schema: TreeStoredSchema): FlexTreeSchema;

/**
 * Status of the tree that a particular node belongs to.
 * @beta
 */
export declare enum TreeStatus {
    /**
     * Is parented under the root field.
     */
    InDocument = 0,
    /**
     * Is not parented under the root field, but can be added back to the original document tree.
     */
    Removed = 1,
    /**
     * Is removed and cannot be added back to the original document tree.
     */
    Deleted = 2
}

/**
 * Document schema data that can be stored in a document.
 *
 * @remarks
 * Note: the owner of this may modify it over time:
 * thus if needing to hand onto a specific version, make a copy.
 * @internal
 */
export declare interface TreeStoredSchema extends StoredSchemaCollection {
    /**
     * Schema for the root field which contains the whole tree.
     */
    readonly rootFieldSchema: TreeFieldStoredSchema;
}

/**
 * A collection of stored schema that fires events in response to changes.
 * @internal
 */
export declare interface TreeStoredSchemaSubscription extends ISubscribable<SchemaEvents>, TreeStoredSchema {
}

/**
 * @internal
 */
export declare type TreeType = TreeNodeSchemaIdentifier;

/**
 * Set of allowed tree types.
 * Providing multiple values here allows polymorphism, tagged union style.
 *
 * If not specified, types are unconstrained
 * (equivalent to the set containing every TreeNodeSchemaIdentifier defined in the document).
 *
 * Note that even when unconstrained, children must still be in-schema for their own type.
 *
 * In the future, this could be extended to allow inlining a TreeNodeStoredSchema here
 * (or some similar structural schema system).
 * For structural types which could go here, there are a few interesting options:
 *
 * - Allow replacing the whole set with a structural type for terminal / non-tree data,
 * and use this as a replacement for values on the tree nodes.
 *
 * - Allow expression structural constraints for child trees, for example requiring specific traits
 * (ex: via TreeNodeStoredSchema), instead of by type.
 *
 * There are two ways this could work:
 *
 * - Constrain the child nodes based on their shape:
 * this makes schema safe editing difficult because nodes would incur extra editing constraints to prevent them
 * from going out of schema based on their location in such a field.
 *
 * - Constrain the types allowed based on which types guarantee their data will always meet the constraints.
 *
 * Care would need to be taken to make sure this is sound for the schema updating mechanisms.
 * @internal
 */
export declare type TreeTypeSet = ReadonlySet<TreeNodeSchemaIdentifier> | undefined;

/**
 * Value that may be stored on a leaf node.
 * @internal
 */
export declare type TreeValue<TSchema extends ValueSchema = ValueSchema> = [
    {
    [ValueSchema.Number]: number;
    [ValueSchema.String]: string;
    [ValueSchema.Boolean]: boolean;
    [ValueSchema.FluidHandle]: IFluidHandle;
    [ValueSchema.Null]: null;
}[TSchema]
][_InlineTrick];

/**
 * An editable view of a (version control style) branch of a shared tree.
 * @privateRemarks
 * This is a wrapper around ITreeView that adjusted it for the public package API.
 * TODO:
 * Establish a naming conversion between these internal and wrapper types.
 * @beta
 */
export declare interface TreeView<in out TRoot> extends IDisposable {
    /**
     * The current root of the tree.
     */
    readonly root: TRoot;
    /**
     * Events for the tree.
     */
    readonly events: ISubscribable<CheckoutEvents>;
}

/**
 * A {@link JsonValidator} implementation which uses TypeBox's JSON schema validator.
 *
 * @privateRemarks Take care to not reference this validator directly in SharedTree code:
 * the intent of factoring JSON validation into an interface is to make validation more pay-to-play
 * (i.e. a JSON validator is only included in an application's bundle if that application references it).
 *
 * Defining this validator in its own file also helps to ensure it is tree-shakeable.
 *
 * @internal
 */
export declare const typeboxValidator: JsonValidator;

/**
 * `{ [key: string]: FieldSchemaTypeInfo }` to `{ [key: string]: TypedTree }`
 *
 * In Editable mode, unwraps the fields.
 * @internal
 */
export declare type TypedFields<TFields extends undefined | {
    readonly [key: string]: TreeFieldSchema;
}> = [
TFields extends {
    [key: string]: TreeFieldSchema;
} ? {
    -readonly [key in keyof TFields]: InsertableFlexField<TFields[key]>;
} : EmptyObject
][_InlineTrick];

/**
 * `Object.values`, but with more specific types.
 * @internal
 */
export declare function typedObjectValues<TKey extends string, TValues>(object: Record<TKey, TValues>): TValues[];

/**
 * A symbol for the name of the type of a tree in contexts where string keys are already in use for fields.
 * See {@link TreeNodeSchemaIdentifier}.
 * @internal
 */
export declare const typeNameSymbol: unique symbol;

/**
 * Remove type brand from name.
 * @internal
 */
export declare type UnbrandedName<TName> = [
TName extends TreeNodeSchemaIdentifier<infer S> ? S : string
][_InlineTrick];

/**
 * A placeholder to use in extends constraints when using the real type breaks compilation of some recursive types due to [a design limitation of TypeScript](https://github.com/microsoft/TypeScript/issues/55758).
 *
 * These extends constraints only serve as documentation:
 * to avoid breaking compilation, this type has to not actually enforce anything, and thus is just `unknown`.
 * Therefore the type safety is the responsibility of the user of the API.
 * @internal
 */
export declare type Unenforced<_DesiredExtendsConstraint> = unknown;

/**
 * Type alias to document which values are un-hydrated.
 *
 * Un-hydrated values are nodes produced from schema's create functions that haven't been inserted into a tree yet.
 *
 * Since un-hydrated nodes become hydrated when inserted, strong typing can't be used to distinguish them.
 * This no-op wrapper is used instead.
 * @beta
 */
export declare type Unhydrated<T> = T;

/**
 * Path from a location in the tree upward.
 * UpPaths can be used with deduplicated upper parts to allow
 * working with paths localized to part of the tree without incurring
 * costs related to the depth of the local subtree.
 *
 * UpPaths can be thought of as terminating at a special root node (that is `undefined`)
 * whose FieldKeys correspond to detached sequences.
 *
 * UpPaths can be mutated over time and should be considered to be invalidated when any edits occurs:
 * Use of an UpPath that was acquired before the most recent edit is undefined behavior.
 * @internal
 */
export declare interface UpPath<TParent = UpPathDefault> {
    /**
     * The parent, or undefined in the case where this path is a member of a detached sequence.
     */
    readonly parent: TParent | undefined;
    /**
     * The Field under which this path points.
     * Note that if `parent` returns `undefined`, this key corresponds to a detached sequence.
     */
    readonly parentField: FieldKey;
    /**
     * The index within `parentField` this path is pointing to.
     */
    readonly parentIndex: NodeIndex;
}

/**
 * Identical to {@link UpPath}, but a duplicate declaration is needed to make
 * the default type parameter compile.
 *
 * @internal
 */
export declare type UpPathDefault = UpPath;

/**
 * Value stored on a node.
 * @internal
 */
export declare type Value = undefined | TreeValue;

/**
 * @internal
 */
export declare interface ValueFieldEditBuilder {
    /**
     * Issues a change which replaces the current newContent of the field with `newContent`.
     * @param newContent - the new content for the field. Must be in Nodes mode.
     */
    set(newContent: ITreeCursor): void;
}

/**
 * Implementation detail of type branding. Should not be used directly outside this file,
 * but shows up as part of branded types so API-Extractor requires it to be exported.
 * @internal
 */
export declare type ValueFromBranded<T extends BrandedType<any, string>> = T extends BrandedType<infer ValueType, string> ? ValueType : never;

/**
 * Schema for what {@link TreeValue} is allowed on a Leaf node.
 * @internal
 */
export declare enum ValueSchema {
    Number = 0,
    String = 1,
    Boolean = 2,
    FluidHandle = 3,
    Null = 4
}

/**
 * A symbol for the value of a tree node in contexts where string keys are already in use for fields.
 * @internal
 */
export declare const valueSymbol: unique symbol;

export { }
